#!/bin/env python
#cluster_jobs.py
"""
Simple helper script to make pbs jobs and submit to the cluster 

Will only work if run on bmf or hemisphere

Can run from command line or import into python code.
"""

from os import makedirs, removedirs
from sys import argv, exit
from commands import getoutput
from string import strip

# module constants 

# qsub template
QSUB_TEXT = """# Walltime Limit: hh:nn:ss 
#PBS -l walltime=%s 

# Node Specification: Two CPUs on one node
#PBS -l nodes=%d:ppn=%d -l ncpus=%d

# Mail: options are (a) aborted, (b) begins execution, (e) ends execution
# use -M <email> for additional recipients
# Queue: Defaults to friendlyq 
#PBS -q %s 

# supress email notification
#PBS -m n

# Job Name:
#PBS -N %s 

######################################################################

# Executable commands go here. Make sure that mpirun -np and -npn match the
# number of processors you requested from PBS! If you are using ppn and want
# npn=2, then you can leave -np and -npn off and PBS will add it for you. If
# you want npn=1, then you have to request all of the processors on all of the
# nodes, and then force it to use less using -npn.

echo ------------------------------------------------------
echo PBS: qsub is running on $PBS_O_HOST
echo PBS: originating queue is $PBS_O_QUEUE
echo PBS: executing queue is $PBS_QUEUE
echo PBS: working directory is $PBS_O_WORKDIR
echo PBS: execution mode is $PBS_ENVIRONMENT
echo PBS: job identifier is $PBS_JOBID
echo PBS: job name is $PBS_JOBNAME
echo PBS: node file is $PBS_NODEFILE
echo PBS: current home directory is $PBS_O_HOME
echo PBS: PATH = $PBS_O_PATH
echo ------------------------------------------------------

cd $PBS_O_WORKDIR 
%s
""" 

# qsub command
QSUB_CMD = "qsub -v PATH,PYTHONPATH -cwd %s"

# command line usage string
USAGE_STR = """Usage: cluster_jobs.py <-m|-ms|-clean|-kill> [<name of commands file> <jobs prefix>]

Options:

  -m        just make jobs. this will create a "jobs" directory in the current 
              directory and write out the pbs job files there. 
  -ms       make and submit jobs.
  -clean    clean up temp and output files (std and stderrror) from previous 
              run(s).
  -kill     kill all of queued or running jobs for current user.

The following are required if -m or -ms specified (2nd and 3rd arguments):

  <name of commands file>   name of file with one command per line.
  <jobs_prefix>             prefix of jobs, 10 chars. e.g. micah_blat
"""


class ClusterJobs:
    """
    Wrapper to queue managament system. 

    Creates and/or sumbits jobs to cluster.
    """

    def __init__(self, commands_list, jobs_prefix, jobs_dir="jobs",
                 ppn=1, nnodes=1, num_jobs=16, wall_time="24:00:00",
                 queue_name="all", job_range=None,
                 use_mpirun = True):
        """
        Initialize jobs

        commands_list: list of commands to run on cluster

            **PLEASE NOTE** path to executable must be fully qualified.
            e.g. to run a python script, the job must look something like
            /usr/local/bin/python2.4 my_script.py [script params'

            **ALSO NOTE** If your python (et al) script takes parameters from
            the command line, and you are checking the number of parameters 
            you should be aware that mpirun will add parameters after the 
            parameters you have passed. 
            (e.g. if len(argv) != 4: <do something>, you will need to modify
            your code)

        jobs_prefix: job prefix. must be <= 10 characters. no spaces. first 
            character must be a-Z. this is what you will see in the job queue, 
            and will be the prifix of all of the standard out and standard 
            error output. required.

        jobs_dir: path to directory to store generated jobs. default = "jobs"

        ppn: number of processors per node to use, default = 1

        nnodes: number of nodes per job, default = 1

        num_jobs: number of jobs to make, default = 16

        wall_time: maximum time the job is allowed to run. friendlyq limit is
            72 hours. if debugging a program that might hang (e.g. an mpi
            program that blocks) it is a good idea to set this value very low, 
            e.g. 5 minutes. format is HH:MM:SS. default = 12 hours

        queue_name: name of queue to submit jobs to. default = friendlyq        

        job_range: only create/submit jobs in this range. not implemented 

        use_mpirun: will run jobs using mpi run. this will send your command
            through mpirun shell script. both mpi and non-mpi programs will
            run fine when this is True. default = True
        """
        if not commands_list:
            raise ValueError, "You must specify a list of commands."

        if not jobs_prefix or len(jobs_prefix) > 10:
            raise ValueError, "Jobs prefix must be a 1-10 characters long."

        if not jobs_prefix[0].isalpha():
            raise ValueError, "First letter of prefix must be a-Z."

        if ppn < 1 or ppn > 2:
            print "WARNING! Possible invalid ppn value. %d" % ppn

        if nnodes < 1 or nnodes > 8:
            print "WARNING! Possible invalid number of nodes. %d" % nnodes

        if num_jobs < 1: 
            raise ValueError, "You must specify > 1 jobs to create."

        if num_jobs > 256:
            print "WARNING! Large number of jobs specified! %d" % num_jobs

        # make sure that absolute path to executable is specified
        if len(filter(lambda x: x.startswith("/"), commands_list)) != \
            len(commands_list):
            raise ValueError, "One or more commands not fully qualified."

        # try to make directory for jobs
        try:
            makedirs(jobs_dir)
        except OSError:
            # if directory already exists, ignore
            pass

        # make sure jobs dir has training slash 
        if not jobs_dir.endswith("/"):
            jobs_dir += "/"

        self.JobsDir = jobs_dir

        # store values
        self.CommandList = commands_list
        self.JobsPrefix = jobs_prefix
        self.ProcsPerNode = ppn
        self.NumNodes = nnodes
        self.NumJobs = num_jobs
        self.WallTime = wall_time
        self.QueueName = queue_name
        self.JobRange = job_range
        self.UseMPIRun = use_mpirun
        self.JobFilenames = []
   
    def makeJobs(self):
        """
        Write out jobs
        """

        out = []
        out.append("Making %d jobs." %  self.NumJobs)
        out.append("Writing jobs to: " +  self.JobsDir)

        ct = 0
        job_lists = {}

        # for each command, add to growing list of jobs
        for cmd in map(strip, self.CommandList):
            cur_job = ct % self.NumJobs

            # check if new job
            if not cur_job in job_lists:
                job_lists[cur_job] = []

            # if UseMPIRun, prefix command with mpirun command
            mpirun = "mpirun -np %d " % self.ProcsPerNode
            if not self.UseMPIRun:
                mpirun = ""

            # append command to current job
            job_lists[cur_job].append(mpirun + cmd)

            ct += 1

        # write out pbs files, cache job filenames
        for jkey, jlist in job_lists.items():
            job_out = "%s%s%d.pbs" %(self.JobsDir,
                                      self.JobsPrefix, 
                                      jkey)

            self.JobFilenames.append(job_out)
            o_file = open(job_out, "w+")
            o_file.write(QSUB_TEXT % (self.WallTime,
                                      self.NumNodes, 
                                      self.ProcsPerNode, 
                                      self.ProcsPerNode,
                                      self.QueueName,
                                      "%s%d" % (self.JobsPrefix, jkey),
                                      '\n'.join(jlist)))
            o_file.close()

        return '\n'.join(out)


    def submitJobs(self):
        """
        Submit all jobs.
        """
        if not self.JobFilenames:
            raise ValueError, "No job filenames available. Have you run makeJobs yet?"
        out = []
        out.append("Submitting %d jobs to %s queue!" % (len(self.JobFilenames),
                                                         self.QueueName))
        
        for p_file in self.JobFilenames:
            # ct += 1
            # if ct < SKIP_NUM_JOBS:
            #     print "skipping: %d" % ct
            #     continue
            # print "Submitting: ", p_file
            out.append(getoutput(QSUB_CMD % p_file))
            
        out.append("done.\nType 'qstat' or 'qstat -a' to check status of your jobs.") 
        return '\n'.join(out)


if __name__ == "__main__":
    
    # check if right number of command line options
    if not(len(argv) ==4 or len(argv) == 2):
        print USAGE_STR 
        exit(1)

    run_type = argv[1]

    # check run type and perform appropriate actions 
    if run_type == "-m" or run_type == "-ms":
        
        # make jobs
        cj = ClusterJobs([x for x in open(argv[2])], argv[3])
        print cj.makeJobs()
        print "done."

        # check if need to submit jobs
        if run_type == "-ms":
            print cj.submitJobs()
            print getoutput("qstat")
            print "done."

    # check if cleaning up after a run
    elif run_type == "-clean":
        print "Removing temp and output files."
        getoutput("rm -f *.[oe][0-9]*[0-9]") 
        getoutput("rm -f PI[0-9]*[0-9]") 
        print "done."

    # check if killing running / queued jobs
    elif run_type == "-kill":
        print "Killing all of my queued / running jobs."

        print getoutput("for i in `qstat | grep $USER | sed -e s/\.bmf.*// " +\
                        "| sort -r`;  do echo 'killing' $i; qdel $i; done")

        print "done."

    # bad runtype specified
    else: 
        print "Invalid run type. %s" % run_type


