#!/usr/bin/env python
# classification.py
"""Provides classification object and some other classification functions

10/17/06 Sandra Smit: written file.
8/9/07 Sandra Smit: switched from Numeric to numpy
8/10/07 Sandra Smit: introduced alpha_all in _decompose_aln because '?' 
(IUPAC Missing) was not accepted when changing string into indices.
6/12/08 Sandra Smit: removed '?' from alphabet in _decompose_aln because
'?' was added to the alphabet.
"""

from __future__ import division

from numpy import array, flatnonzero, arange, take, zeros, put,\
    ones, sum
from cogent.core.moltype import RNA
from cogent.core.alphabet import CharAlphabet
from cogent.core.alignment import Alignment, DenseAlignment
from cogent.core.sequence import RnaSequence, ModelSequence


class Classification(str):
    """Represents classification string of an RNA secondary structure"""
 
    def _get_symbols(self):
        """Return a list of symbols present in the classification string"""
        return dict.fromkeys(self).keys()

    Symbols = property(_get_symbols)

    def positionsPerClass(self, classes=None, include_total=False):
        """Return dict of {class: [positions]}

        classes -- list of symbols, specifying the classes that should be 
            in the resulting dictionary. If None, all symbols in the 
            classifiaction string are included.
        include_total -- boolean, if True, an entry 'TOTAL':[all pos] is
            added to the result
        
        The return value is a dictionary of {class: [positions]} where
            each key is a structural class or 'TOTAL' and each key is a list
            of positions that are classified in a particular class.
        """
        symbols_to_use = classes or self.Symbols
        result = {}
        if symbols_to_use:
            if not self:
                for s in symbols_to_use:
                    result[s] = array([])
            else:
                a = array(list(self))
                for s in symbols_to_use:
                    result[s] = flatnonzero(a==s)
        if include_total:
            result['TOTAL'] = arange(len(self))
        return result

    def _decompose_rna_sequence(self, seq, classes=None, include_total=False):
        """Return array of fractions U, C, A, and G of sequence per class.

        seq: RnaSequence object
        classes: list of classes that should be included in the result. If
            classes is given, only those classes are inlcuded.
        include_total: boolean. If set to True, 'TOTAL' (sequence composition)
            is included in the result.

        Only U, C, A, and G are counted, all unknown characters are ignored.
        
        Result is a dictionary of {class: counts}. counts is an array of 4
            fractions: U C A G in this order.
        """
        if not isinstance(seq, RnaSequence):
            raise ValueError('Wrong sequence type: %s'%(type(seq)))
        if len(seq) != len(self):
            raise ValueError(\
                "Classification and matching sequence have different lengths")
        result = {}
        pos_per_class = self.positionsPerClass(classes, include_total)
        for cl, locations in pos_per_class.items():
            if locations.any():
                sub_seq = seq[locations.tolist()]
                result[cl] = array([sub_seq.count(x) for x in "UCAG"], float)
                if result[cl].any():
                    result[cl] /= sum(result[cl])
            else:
                result[cl] = zeros(4, float)
        return result

    def _decompose_model_sequence(self, seq, classes=None, include_total=False):
        """Return array of fractions U, C, A, and G of sequence.

        seq: RnaSequence object
        classes: list of classes that should be included in the result. If
            classes is given, only those classes are inlcuded.
        include_total: boolean. If set to True, 'TOTAL' (sequence composition)
            is included in the result.

        Only U, C, A, and G are counted, all unknown characters are ignored.
        
        Result is a dictionary of {class: counts}. counts is an array of 4
            fractions: U C A G in this order.
        """
        if not isinstance(seq, ModelSequence):
            raise ValueError('Wrong sequence type: %s'%(type(seq)))
        assert len(seq) == len(self)
        wanted_indices = seq.Alphabet.toIndices('UCAG')
        result = {}
        pos_per_class = self.positionsPerClass(classes, include_total)
        for cl, locations in pos_per_class.items():
            if locations.any():
                sub_seq = take(array(seq._data), locations)
                all_counts = seq.Alphabet.counts(sub_seq)
                seq_counts = array([all_counts[idx] for idx in wanted_indices])
                if seq_counts.any():
                    result[cl] = seq_counts/sum(seq_counts)
                else:
                    result[cl] = zeros(4, float)
            else:
                result[cl] = zeros(4, float)
        return result

    def decomposeSequence(self, seq, classes=None, include_total=False):
        """Return array of fractions U, C, A, and G of sequence.

        seq: RnaSequence or ModelSequence object
        classes: list of classes that should be included in the result. If
            classes is given, only those classes are inlcuded.
        include_total: boolean. If set to True, 'TOTAL' (sequence composition)
            is included in the result.

        This is the general method to get the UCAG fractions in a sequence.
        This method will delegate the work to the appropriate private method.

        Only U, C, A, and G are counted, all unknown characters are ignored.
        
        Result is a dictionary of {class: counts}. counts is an array of 4
            fractions: U C A G in this order.
        """
        if isinstance(seq, RnaSequence):
            return self._decompose_rna_sequence(seq, classes, include_total)
        elif isinstance(seq, ModelSequence):
            return self._decompose_model_sequence(seq, classes, include_total)
        else:
            raise ValueError('Unknown sequence type: %s'%(type(seq)))
        

    def _decompose_dense_aln(self, aln, classes=None, include_total=False):
        """Return dict of {seq_id: {class: [U,C,A,G]}}

        aln: DenseAlignment object. Sequences should be RnaSequences. 
            The sequences in the alignment should have unique names.
        classes: list of classes that should be included in the result. If
            classes is given, only those classes are inlcuded. If classes
            is None, it uses all the classes present in the classification.
        include_total: boolean. If set to True, 'TOTAL' (sequence composition)
            is included in the result.

        Only U, C, A, and G are counted, all unknown characters are ignored.
        
        Result is a dictionary of {seq_id: {class: counts}}. 
            counts is an array of 4 fractions: U C A G in this order.
        """
        if not isinstance(aln, DenseAlignment):
            raise ValueError('Alignment should be a DenseAlignment')
        if not aln.SeqLen == len(self):
            raise ValueError(\
            "Sequences in alignment and classification have different lengths")
        names = aln.Names
        result = {}
        pos_per_class = self.positionsPerClass(classes, include_total)
        for cl, locations in pos_per_class.items():
            if locations.any():
                sub_aln = aln.getSubAlignment(pos=locations)
                counts = sub_aln.getSeqFreqs().Data[:,:4] #only UCAG
            else:
                counts = zeros((len(aln.Seqs),4))
            for idx in range(len(counts)):
                seq_name = names[idx]
                seq_counts = counts[idx]
                if seq_counts.any():
                    fractions = seq_counts/sum(seq_counts)
                else:
                    fractions = array(seq_counts, float)
                if seq_name not in result:
                    result[seq_name] = {}
                result[seq_name][cl] = fractions
        return result

    def _decompose_aln(self, aln, classes=None, include_total=False):
        """Return dict of {seq_id: {class: [U,C,A,G]}}

        aln: DenseAlignment object. Sequences should be RnaSequences. 
            The sequences in the alignment should have unique names.
        classes: list of classes that should be included in the result. If
            classes is given, only those classes are inlcuded. If classes
            is None, it uses all the classes present in the classification.
        include_total: boolean. If set to True, 'TOTAL' (sequence composition)
            is included in the result.

        Only U, C, A, and G are counted, all unknown characters are ignored.
        
        Result is a dictionary of {seq_id: {class: counts}}. 
            counts is an array of 4 fractions: U C A G in this order.
        """
        if not isinstance(aln, Alignment):
            raise ValueError('Alignment should be an Alignment object')
        assert aln.SeqLen == len(self)
        alpha_degen = RNA.Alphabets.DegenGapped
        alpha_all = alpha_degen
        #alpha_all = CharAlphabet(list(RNA.Alphabets.DegenGapped)+['?'],\
        #    MolType='RNA')
        alpha_base = RNA.Alphabets.Base
        counts_f = alpha_base.counts
        result = {}
        pos_per_class = self.positionsPerClass(classes, include_total)
        for cl, locations in pos_per_class.items():
            if locations.any():
                sub_aln = aln.takePositions(cols=locations)
                for seq in sub_aln.iterSeqs():
                    seq_name = seq.Name
                    seq_counts = alpha_base.counts(alpha_all.toIndices(seq))
                    if seq_counts.any():
                        fractions = seq_counts/sum(seq_counts)
                    else:
                        fractions = zeros(4, float)
                    if seq_name not in result:
                        result[seq_name] = {}
                    result[seq_name][cl] = fractions
            else:
                for seq in aln.iterSeqs():
                    seq_name = seq.Name
                    if seq_name not in result:
                        result[seq_name] = {}
                    result[seq_name][cl] = zeros(4, float)
        return result

    def decomposeAlignment(self, aln, classes=None, include_total=False):
        """Return dict of {seq_id: {class: [U,C,A,G]}}

        aln: DenseAlignment object. Sequences should be RnaSequences. 
            The sequences in the alignment should have unique names.
        classes: list of classes that should be included in the result. If
            classes is given, only those classes are inlcuded. If classes
            is None, it uses all the classes present in the classification.
        include_total: boolean. If set to True, 'TOTAL' (sequence composition)
            is included in the result.

        This is the public method that should be used to decompose your
        alignment. It accepts either an Alignment object or an DenseAlignment
        object. An error will be raised on any other input. 

        Only U, C, A, and G are counted, all unknown characters are ignored.
        
        Result is a dictionary of {seq_id: {class: counts}}. 
            counts is an array of 4 fractions: U C A G in this order.
        """
        if isinstance(aln, DenseAlignment):
            return self._decompose_dense_aln(aln, classes, include_total)
        if isinstance(aln, Alignment):
            return self._decompose_aln(aln, classes, include_total)
        else:
            raise ValueError('Unknown alignment type: %s'%(type(aln)))

        
    def decomposeAlnOrder(self, aln, classes=None, include_total=False):
        """If taking a sub alignment would return the sequences in the same
        order, this method would work
        """
        if not isinstance(aln, Alignment):
            raise ValueError('Alignment should be an Alignment object')
        print aln.MolType.Sequence
        print 'in here'
        assert aln.SeqLen == len(self)
        alpha_degen = RNA.Alphabets.DegenGapped
        alpha_base = RNA.Alphabets.Base
        print alpha_base
        counts_f = alpha_base.counts
        result = {}
        pos_per_class = self.positionsPerClass(classes, include_total)
        for cl, locations in pos_per_class.items():
            if locations:
                sub_aln = aln.takePositions(cols=locations)
                print sub_aln
                print cl
                aln_indices = [alpha_degen.toIndices(s)\
                    for s in sub_aln.iterSeqs()]
                #print 'aln_indices', aln_indices
                counts = map(counts_f, aln_indices)
                print 'counts', counts
            else:
                counts = zeros((len(aln.Seqs),4))
                print 'counts', counts
            names = aln.Names
            for idx in range(len(counts)):
                seq_name = names[idx]
                seq_counts = counts[idx]
                print idx, seq_name, seq_counts
                if counts[idx]:
                    fractions = seq_counts/sum(seq_counts)
                else:
                    fractions = array(seq_counts, Float)
                print idx, seq_name, fractions
                if seq_name not in result:
                    result[seq_name] = {}
                result[seq_name][cl] = fractions
        return result

    def decompose_aln(self, aln, classes='SLBJEF', struct_id='struct'):
        """Old: used on old_cogent.alignment object"""
        result = []
        pos = self.positionsPerClass()
        for label, seq in aln.items():
            for cl in classes:
                #print cl
                temp = []
                temp.append(struct_id)
                temp.append(cl)
                temp.append(label)
                if cl in pos:
                    #print take(seq,pos[cl])
                    bu = BaseUsage(take(seq,pos[cl]))

                else:
                    bu = BaseUsage('')
                #print bu
                bu.normalize()
                #print bu
                temp.extend([bu['U'], bu['C'], bu['A'], bu['G'],\
                    bu['UC'], bu['UG'], bu['UA']])
                #gl = GroupList([bu], Groups=[cl,label])
                result.append(temp)
            temp = []
            temp.append(struct_id)
            temp.append('Total')
            temp.append(label)
            bu = BaseUsage(seq)
            bu.normalize()
            temp.extend([bu['U'], bu['C'], bu['A'], bu['G'],\
                bu['UC'], bu['UG'], bu['UA']])
            result.append(temp)

        return result
