#!/usr/bin/env python
# input.py

"""
Author: Sandra Smit (Sandra.Smit@colorado.edu)

Revision History:
File created on 08 May 2007.

"""
from __future__ import division
from cogent.parse.fasta import MinimalFastaParser
from cogent.core.moltype import RNA
from cogent.core.sequence import RnaSequence, ModelSequence
from cogent.core.alignment import Alignment, DenseAlignment
from cogent.struct.rna2d import ViennaStructure
from rnalib.core.classification import Classification
from cogent.parse.fasta import FastaFinder, is_fasta_label,\
    default_label_to_name
from rnalib.struct.pairs_util import vienna_mask

def plain_fasta_aln(lines, dense=False, moltype=RNA,\
    alpha=RNA.Alphabets.DegenGapped):
    """Return Alignment or DenseAlignment object from plain fasta file

    lines -- list of lines or anything that behaves like it
    dense -- boolean, if True, DenseAlignment is returned, if False,
        Alignment is returned.
    alpha -- Alphabet object, default is RNA.Alphabets.DegenGapped
    """
    seqs = []
    for label, seq in MinimalFastaParser(lines):
        if dense:
            s = ModelSequence(seq, Name=label, Alphabet=alpha)
        else:
            s = RnaSequence(seq, Name=label)
        seqs.append(s)
    if dense:
        aln = DenseAlignment(seqs, MolType=moltype, Alphabet=alpha)
    else:
        aln = Alignment(seqs, MolType=moltype)
    return aln

def get_aln_pairs(lines, dense=False, alpha=RNA.Alphabets.DegenGapped):
    """Return tuple of Alignment and Pairs object
    
    lines -- file stream or list of lines
    dense -- bookean, when False Alignment is returned, when True
        DenseAlignmentis returned. Default is False.

    Handles format:
    >consensus_vienna
    ...((..))...
    >consensus_mask
    ...11..11...
    >seq1
    ACCGGUACCUGC
    >seq2
    A--GCUUGC-GA
    ...
    """
    
    cons_vienna = None
    cons_mask = None
    cons_pairs = None    
    seqs = []
    aln = None

    for label, seq in MinimalFastaParser(lines):
        if label == 'consensus_vienna':
            cons_vienna = seq
            continue
        elif label == 'consensus_mask':
            cons_mask = seq
            continue
        else:
            if dense:
                s = ModelSequence(seq, Name=label, Alphabet=alpha)
            else:
                s = RnaSequence(seq, Name=label)
            seqs.append(s)

    if dense:
        aln = DenseAlignment(seqs, MolType=RNA, Alphabet=alpha)
    else:
        aln = Alignment(seqs, MolType=RNA)
    cons_pairs = vienna_mask(cons_vienna, cons_mask)

    return aln, cons_pairs

def triple_parser(lines, finder=FastaFinder, strict=True,\
    label_to_name=default_label_to_name, is_label=is_fasta_label):
    """Return tuple of (label, vienna, mask, cl)

    Handles the following format:
    >label
    ...((...)).
    ...11...11.
    OOOSSLLLSSO
    >label2
    .(((...))).
    etc.
    """
    for rec in finder(lines):
        #first line must be a label line
        if not is_label(rec[0]):
            if strict:
                raise RecordError("Found Fasta record without label line: %s"%\
                    rec)
            else:
                continue
        #record must have at least one sequence
        if len(rec) != 4:
            if strict:
                raise RecordError("Record should contain 4 lines")
            else:
                continue
            
        label = label_to_name(rec[0])
        vienna = ViennaStructure(rec[1].strip())
        mask = rec[2].strip()
        cl = Classification(rec[3].strip())

        yield label, vienna, mask, cl


def id_to_vmc(lines):
    """Return dict of {id: (vienna, mask, classification)}
    """

    result = {}
    for l, v, m, c in triple_parser(lines):
        result[l] = (v,m,c)
    return result


def remove_whitespace(label):
    return '_'.join(label.split())

def aln_from_label_seq(data, moltype=RNA, alpha=RNA.Alphabets.DegenGapped,\
    dense=False, label_f=None, to_upper=False):
    """Return (Dense)Alignment alphabet from data
    
    data -- list of [(seq_label, seq)] where seq_label and seq are normal
        strings
    moltype -- MolType object
    alpha -- Alphabet object
    dense -- boolean, if True DenseAlignment is returnes, otherwise normal 
        Alignment object is returned.
    label_f -- function to be applied to the sequence label. If None, 
        original label is used.
    to_upper -- boolean, if True all characters in the sequences are 
        converted to upper case before checking whether they occur in 
        the alphabet.
   
    Checklist:
    - sequence length should be the same for all items
    - all characters should be in the specified alphabet
    - all labels should be unique
    - sequence order should be maintained
    - sequences should be of correct input for type of alignment
    """
    # GROUP SEQS BY LABEL
    by_label = {}
    label_order = []
    for label, seq in data:
        label_order.append(label)
        if label not in by_label:
            by_label[label] = []
        by_label[label].append(seq)
    
    # PROCESS UNIQUE LABELS ONLY, SKIP DUPLICATES
    unique_seqs = dict([(k,v[0]) for (k,v) in by_label.items() if len(v) == 1])
    unique_label_order = [i for i in label_order if i in unique_seqs]
    assert len(unique_label_order) == len(unique_seqs)
    
    # FIND MAJORITY LENGTH
    lengths = {}
    for label, seq in unique_seqs.items():
        if len(seq) not in lengths:
            lengths[len(seq)] = 0
        lengths[len(seq)] += 1
    majority_len = -1
    majority_count = -1
    for k,v in lengths.items():
        if v == majority_count:
            raise ValueError("Double majority count")
        if v > majority_count:
            majority_count = v
            majority_len = k

    # COLLECT SEQS OF MAJORITY LENGTH AND WITH CHARS IN ALPHA
    quality_seqs = []
    for label in unique_label_order:
        seq = unique_seqs[label]
        if to_upper:
            seq = seq.upper() # CONVERT TO UPPER CASE
        if len(seq) != majority_len: # SKIP SEQS OF BAD LENGTH
            continue
        else:
            if label_f:
                label = label_f(label)
            all_ok = True
            for rna_char in seq:
                if rna_char not in alpha:
                    all_ok = False
                    break
            if all_ok: # add to quality sequences
                if dense:
                    s = ModelSequence(seq, Name=label, Alphabet=alpha)
                else:
                    s = moltype.Sequence(seq, Name=label)
                quality_seqs.append(s)

    if dense:
        aln = DenseAlignment(quality_seqs, MolType=moltype, Alphabet=alpha)
    else:
        aln = Alignment(quality_seqs, MolType=moltype)

    return aln


if __name__ == "__main__":
    pass
