#!/usr/bin/env python

from __future__ import division
from string import strip
from Numeric import sum
from MLab import average, std
from cogent.struct.rna2d import classify, Vienna
from cogent.util.organizer import Filter, Organizer, GroupList
from cogent.maths.stats.special import fix_rounding_error
from cogent.format.mage import MagePoint, MageList, MageGroup,\
    RainbowColors, MageColors, Kinemage, CartesianSimplexHeader, is_essentially
from basecomp.analyzers import MultipleSeqAnalyzer, SingleSeqAnalyzer
from basecomp.simplex import Simplex
from basecomp.structseq import StructSeq, six_classifier, StructSeqFromString

def ucag_to_mage_point(vector, label=None, radius=None, cartesian=True):
    """Return a MagePoint from a vector of UCAG fractions.

    vector -- list or Numeric array, fractions of U, C, A, and G,
        in this ORDER!
    label -- str, Label of the MagePoint.
    radius -- float, radius of MagePoint.
    cartesian -- boolean, indicates whether the point should be in Cartesian
        coordinates (i.e. X=U+C, Y=U+G, Z=U+A)
        or not (i.e. X=A, Y=C, Z=G, U is implicit).
    """
    # check input
    if is_essentially(0, sum(vector)):
        return None
    else:
        if not is_essentially(1, sum(vector)):
            raise ValueError("U, C, A, and G values should add up to one")

    vector = map(fix_rounding_error, vector)
    # calculate mage coordinates
    if cartesian:
        x = vector[0]+vector[1]
        y = vector[0]+vector[3]
        z = vector[0]+vector[2]
    else:
        x = vector[2]
        y = vector[1]
        z = vector[3]
    # create and return MagePoint
    m = MagePoint(Coordinates=[x,y,z], Label=label, Radius=radius)
    return m

def get_colors(classes, colors=None):
    """Return a dictionary of {class: color} for each class in classes.

    if colors is specified, it'll add a color for every class not in the
        dictionary. If colors is None and the number of classes is smaller
        than seven, the RainbowColors will be used, otherwise the normal
        MageColors will be used. If the number of classes is larger than
        25, several colors will be used more than once.
    """
    result = {}
    if colors:
        for c in classes:
            if c in colors:
                result[c] = colors[c]
            else:
                used_colors = result.values()
                found = False
                idx = 0
                while found is False:
                    color = MageColors[idx]
                    if color not in used_colors:
                        found = True
                        result[c] = color
                    else:
                        idx += 1
    else:
        if len(classes) <= 7:
            result = dict(zip(classes, RainbowColors))
        else:
            result = dict(zip(classes,\
                (divmod(len(classes),20)[0] + 1)*MageColors))
    return result

def create_mage_group(decomposition, label=None, style='dot',\
    radius=None, colors=None, cartesian=True):
    """Return a MageGroup object from decomposition results.

    decomposition: dict of {seq_id: {class: [counts]}} where counts
        is a vector of 4 fractions: U, C, A, G (in this order!).
        They should add up to 1 if there was at least one base in that
        class. If they add up to zero, the class was empty.
    label: str, label of MageGroup
    style: str, style of the MageLists that will be created. See MageList
        documentation for options.
    radius: float, radius of all mage points that will be created.
    colors: dict of {class: color}. See Mage documentation for allowed colors.
    cartesian: boolean, indicates whether the point should be in Cartesian
        coordinates (i.e. X=U+C, Y=U+G, Z=U+A)
        or not (i.e. X=A, Y=C, Z=G, U is implicit).
    """
    mage_lists = {}
    # taking the keys of the first decomposition is not safe
    # need to see them all
    categories = {} 
    for i in decomposition.values():
        categories.update(dict.fromkeys(i.keys()))
    categories = categories.keys()
    categories.sort() #make sure colors in same order in case of same keys
    list_colors = get_colors(categories, colors)
    for seq_id, classes in decomposition.items():
        for cl, ucag_vector in classes.items():
            if cl not in mage_lists:
                mage_lists[cl] = MageList(Label=cl, Style=style,\
                    Color=list_colors[cl])
            mp = ucag_to_mage_point(ucag_vector,\
                label=seq_id, radius=radius, cartesian=cartesian)
            if mp:
                mage_lists[cl].append(mp)
    result = MageGroup(mage_lists.values(), Label=label)
    return result

def kin_single_aln(aln, pairs, classification_f=None, include_total=False,\
    name=None, style='dot',\
    radius=None, colors=None, cartesian=True, count=1):
    """Return Kinemage object of composition of alignment.

    aln: Alignment or DenseAlignment object
    pairs: Pairs object
    classification_f: classification function (should take pairs and length
        as input).
    name: name of the mage group that is created. If None, it'll try to 
        extract the alignment name, else it will use 'RNA' as default.
    """
    
    # group_name: prefer input, otherwise try to extract alignment
    # Name, fall back to default name
    default_group_name = 'RNA'
    group_name = name or aln.Name or default_group_name
    
    if cartesian:
        mage_header = CartesianSimplexHeader()
    else:
        mage_header = SimplexHeader()

    cl = classification_f(pairs, aln.SeqLen)
    decomposition = cl.decomposeAlignment(aln, include_total=include_total)
    kinemage = Kinemage(Count=count, Header=mage_header)
    kinemage.Groups.append(create_mage_group(decomposition,\
        label=group_name, style=style, radius=radius,\
        colors=colors, cartesian=cartesian))
    return kinemage

def kin_indiv_seqs_structs(seqs, structures, classification_f=None,\
    include_total=False, name=None, style='dot',\
    radius=None, colors=None, cartesian=True, count=1):
    """Return Kinemage object of composition of alignment.

    seqs: dictionary of {seq_id: RnaSequence}
    structures: dictionary of {seq_id: Pairs}
    classification_f: classification function (should take pairs and length
        as input).
    name: name of the mage group that is created. If None, it'll try to 
        extract the alignment name, else it will use 'RNA' as default.
    """
    # group_name: prefer input, otherwise fall back to default name
    default_group_name = 'RNA'
    group_name = name or default_group_name
    
    if cartesian:
        mage_header = CartesianSimplexHeader()
    else:
        mage_header = SimplexHeader()

    decomposition = {}
    for k in seqs:
        s = seqs[k]
        p = structures[k]
        c = classification_f(p, len(s))
        decomposition[k] = c.decomposeSequence(s, include_total=include_total)

    kinemage = Kinemage(Count=count, Header=mage_header)
    kinemage.Groups.append(create_mage_group(decomposition,\
        label=group_name, style=style, radius=radius,\
        colors=colors, cartesian=cartesian))
    return kinemage



def old_to_new_kinemage(s, change_size=False):
    new_simplex = Kinemage(Count=1, Header=CartesianSimplexHeader())
    for gr in s.iterGroups():
        new_magegroup = MageGroup(Label=gr.Label,\
            Subgroup=gr.Subgroup, Off=gr.Off, NoButton=gr.NoButton,\
            RecessiveOn=gr.RecessiveOn)
        for ml in gr.iterLists():
            if ml.Label in ['Stem','Total','Loop','Bulge','Junction',\
                'End','Flexible','Knot']:
                cart_ml = ml.toCartesian()
                if change_size:
                    cart_ml.Style='sphere'
                    for mp in cart_ml.iterPoints():
                        mp.Radius=0.01
                new_magegroup.append(cart_ml)
            else:
                new_magegroup.append(ml)
        new_simplex.Groups.append(new_magegroup)
            
    #print s
    return new_simplex

def kin_single_aln_from_cl(aln, cl, include_total=False,\
    name=None, style='dot',\
    radius=None, colors=None, cartesian=True, count=1):
    """Return Kinemage object of composition of alignment.

    aln: Alignment or DenseAlignment object
    pairs: Pairs object
    classification_f: classification function (should take pairs and length
        as input).
    name: name of the mage group that is created. If None, it'll try to 
        extract the alignment name, else it will use 'RNA' as default.
    """
    
    # group_name: prefer input, otherwise try to extract alignment
    # Name, fall back to default name
    default_group_name = 'RNA'
    group_name = name or aln.Name or default_group_name
    
    if cartesian:
        mage_header = CartesianSimplexHeader()
    else:
        mage_header = SimplexHeader()

    #cl = classification_f(pairs, aln.SeqLen)
    decomposition = cl.decomposeAlignment(aln, include_total=include_total)
    kinemage = Kinemage(Count=count, Header=mage_header)
    kinemage.Groups.append(create_mage_group(decomposition,\
        label=group_name, style=style, radius=radius,\
        colors=colors, cartesian=cartesian))
    return kinemage

# ============================================================================
# FUNCTIONS ON KINEMAGE
# ============================================================================

LABELS_TO_SKIP = ['Tetrahedron','Lines','Cube']
KIN_AXES = {0:'UC',1:'UG',2:'UA'}

def get_particular_value(kinemage, ss, axis_label, func=std,\
    kin_axes=KIN_AXES, labels_to_skip=LABELS_TO_SKIP):
    """Return requested value from Kinemage object.
    
    kinemage -- Kinemage object
    ss -- string, label of MageList of interest, e.g. "S", "L", "TOTAL"
    axis_label -- string, axis of interest, "UA", "UC", or "UG"
    func -- function that can be applied to array and returns some value

    Assumes standard kinemage object in Cartesian coordinates.
    """
    result = None
    for gr in kinemage.iterGroups():
        if gr.Label in labels_to_skip:
            continue
        else:
            for l in gr.iterLists():
                if l.Label != ss:
                    continue
                data = l.toArray(include_radius=False)
                for ax_col, ax_str in kin_axes.items():
                    if ax_str != axis_label:
                        continue
                    result = func(data[:,ax_col])
    return result

def kin_stats(kinemage, functions=[std, average, min, max],\
    kin_axes=KIN_AXES, labels_to_skip=LABELS_TO_SKIP):
    """Return dict of {ss_type: axis_label: [std, avg, min, max]}

    functions -- list of functions that should be applied to the data
    kin_axes -- standard axis columns and axis labels in Cartesian kinemage
    """
    result = {}
    for gr in kinemage.iterGroups():
        if gr.Label in labels_to_skip:
            continue
        else:
            for l in gr.iterLists():
                if not l:
                    continue
                if l.Label not in result:
                    result[l.Label] = {}
                data = l.toArray(include_radius=False)
                for ax_col, ax_str in kin_axes.items():
                    res = []
                    for fn in functions:
                        res.append(fn(data[:,ax_col]))
                    result[l.Label][ax_str] = res
    return result

def get_unpaired_avg_kin(kinemage, ss_types="LBO",\
    kin_axes=KIN_AXES, func=std):
    """Return AVG value of function on ss_type and axes.
    """
    stats = kin_stats(kinemage, functions=[func],\
        kin_axes=kin_axes, labels_to_skip=LABELS_TO_SKIP)
    value = get_unpaired_avg_stats(stats, ss_types=ss_types,\
        kin_axes=kin_axes, col_idx=0)
    return value
    
def get_unpaired_avg_stats(stats_data, ss_types="LBO",\
    kin_axes=KIN_AXES, col_idx=0):
    """Return unpaired AVG from stats

    Iterate along all structural elements and axes, collect the value
    at col_idx. Remove Nones. Take the average.
    """
    all_numbers = []
    for ss_elem in ss_types:
        for ax_str in kin_axes.values():
            try:
                all_numbers.append(stats_data[ss_elem][ax_str][col_idx])
            except KeyError:
                continue
    filtered = filter(None, all_numbers)
    if filtered:
        rec_value = average(filtered)
    else:
        rec_value = 0
    return rec_value

# ==================================================================
# OLD FUNCTIONS
# ==================================================================
def aln_to_kin(aln, structure):
    
    #seqs = [RnaSequence(s.strip()) for s in list(open(seq_file))]
    #seqs = aln.values()

    #for struct_id, line in enumerate(open(struct_file)):
        for k, v in aln.items():
            aln[k] = RnaSequence(v)
        #struct = Vienna(line.strip().replace('-','.'))
        classification = classify(structure)
        #print classification

        structseqs = []
        for label, seq in aln.items():
            structseqs.append(StructSeq(str(label), seq, structure,\
            classification))

        group_name = 'RNA'
        seqs_group = GroupList(structseqs, [group_name])

        an = SingleSeqAnalyzer()
        res = an(six_classifier, seqs_group)

        colors = {'Total':'purple','Stem':'yellow','Loop':'blue','Bulge':'red',\
        'Junction':'green','End':'cyan','Flexible':'orange','Knot':'white'} 
        simplex = Simplex(Colors=colors,Ignore_none=True)
        #out = open(out_dir + '/' + str(struct_id) + '.kin','w')
        #out.write(str(simplex(res)))
        #out.close()
        return simplex(res)

def aln_to_kin_individual(aln, structures):
    """
        aln: Alignment object
        structures: dict of structures, keys should match sequence
            keys in the alignment
    """
    
    #seqs = [RnaSequence(s.strip()) for s in list(open(seq_file))]
    #seqs = aln.values()

    #for struct_id, line in enumerate(open(struct_file)):
    for k, v in aln.items():
        aln[k] = RnaSequence(v)
    #struct = Vienna(line.strip().replace('-','.'))
    #classification = classify(structure)
    #print classification

    structseqs = []
    for label, seq in aln.items():
        classification = classify(structures[label])
        structseqs.append(StructSeq(str(label), seq, structures[label],\
        classification))

    group_name = 'RNA'
    seqs_group = GroupList(structseqs, [group_name])

    an = SingleSeqAnalyzer()
    res = an(six_classifier, seqs_group)

    colors = {'Total':'purple','Stem':'yellow','Loop':'blue','Bulge':'red',\
    'Junction':'green','End':'cyan','Flexible':'orange','Knot':'white'} 
    simplex = Simplex(Colors=colors,Ignore_none=True)
    #out = open(out_dir + '/' + str(struct_id) + '.kin','w')
    #out.write(str(simplex(res)))
    #out.close()
    return simplex(res)
# ==================================================================
# END OLD FUNCTIONS
# ==================================================================



