#!/usr/bin/env python
# test_mage_util.py

"""
Author: Sandra Smit (Sandra.Smit@colorado.edu)

Revision History:
File created on 17 Oct 2006.

"""
from __future__ import division
from numpy import array
from cogent.util.unit_test import TestCase, main
from rnalib.format.mage import MageColors, MagePoint
from rnalib.format.mage_util import ucag_to_mage_point, create_mage_group,\
    get_colors


class MageUtilTests(TestCase):

    def test_ucag_to_mage_point(self):
        """ucag_to_mage_point: should return None or proper MagePoint"""
        # empty class
        v = array([0,0,0,0])
        self.assertEqual(ucag_to_mage_point(v), None)
        # full class standard
        v = array([.1, .2, .3, .4])
        self.assertEqual(str(ucag_to_mage_point(v)), "0.3 0.5 0.4")
        v = array([1/3, 1/3, 1/3, 0])
        self.assertEqual(str(ucag_to_mage_point(v)),\
            ' '.join(map(str, array([2/3,1/3,2/3]))))
        v = array([.1, .2, .3, .4])
        # with label
        self.assertEqual(str(ucag_to_mage_point(v, label='joe')),\
        "{joe} 0.3 0.5 0.4")
        # with label and radius
        self.assertEqual(str(ucag_to_mage_point(v, label='joe', radius=0.01)),\
        "{joe} r=0.01 0.3 0.5 0.4")
        # not cartesian
        self.assertEqual(str(ucag_to_mage_point(v, cartesian=False)),\
        "0.3 0.2 0.4")
        # input as list
        v = [1/3, 1/3, 1/3, 0]
        self.assertEqual(str(ucag_to_mage_point(v)),\
            ' '.join(map(str, array([2/3,1/3,2/3]))))
    
    def test_get_colors(self):
        """get_colors: should work with fully or partially specified colors"""
        self.assertEqual(get_colors('SL', {'S':'red', 'L':'green'}),
            {'S':'red', 'L':'green'})
        self.assertEqual(get_colors('SL', {'S':'blue'}),
            {'S':'blue', 'L':'red'})
        self.assertEqual(get_colors('ABCDEFG'),\
            {'A':'red', 'B':'orange', 'C':'yellow', 'D':'green',\
            'E':'cyan', 'F':'blue', 'G':'purple'})
        self.assertEqual(get_colors('ABCDEFGHIJKLMNO'),
            dict(zip('ABCDEFGHIJKLMNO', MageColors)))
        self.assertEqual(get_colors('ABCDEFGHIJKLMNOPQRSTUVWXYZ'),
            dict(zip('ABCDEFGHIJKLMNOPQRSTUVWXYZ', MageColors*2)))
        self.assertEqual(get_colors(''),{})

    def test_create_mage_group(self):
        """create_mage_group: should work with different parameter settings"""
        data = {'seq_1':{'S':[.1,.2,.3,.4], 'L':[1/3,1/3,0,1/3]},\
                'seq_2':{'S':[1/8,0,7/8,0], 'L':[0,0,0,0]}}
        mp1 = MagePoint([.3,.5,.4], Label='seq_1')
        mp2 = MagePoint([2/3,2/3,1/3], Label='seq_1')
        mp3 = MagePoint([1/8,1/8,1.0], Label='seq_2')
        exp_lines = ['@group {RNA} recessiveon', '@dotlist {S} color=orange',\
            str(mp1), str(mp3), '@dotlist {L} color=red', str(mp2)]
        self.assertEqual(str(create_mage_group(data, label='RNA')),\
            '\n'.join(exp_lines))

        # colors specified
        exp_lines = ['@group {RNA} recessiveon', '@dotlist {S} color=yellow',\
            str(mp1), str(mp3), '@dotlist {L} color=blue', str(mp2)]
        self.assertEqual(str(create_mage_group(data, label='RNA',\
            colors={'S':'yellow','L':'blue'})),\
            '\n'.join(exp_lines))

        # other parameters on
        data = {'seq_1':{'S':[.1,.2,.3,.4], 'L':[1/3,1/3,0,1/3]},\
                'seq_2':{'S':[1/8,0,7/8,0], 'L':[0,0,0,0]}}
        mp1 = MagePoint([.3,.5,.4], Label='seq_1', Radius=0.01)
        mp2 = MagePoint([2/3,2/3,1/3], Label='seq_1', Radius=0.01)
        mp3 = MagePoint([1/8,1/8,1.0], Label='seq_2', Radius=0.01)
        exp_lines = ['@group {RNA} recessiveon',\
            '@spherelist {S} color=orange',\
            str(mp1), str(mp3), '@spherelist {L} color=red', str(mp2)]
        self.assertEqual(str(create_mage_group(data, label='RNA',\
            style='sphere', radius=0.01)),\
            '\n'.join(exp_lines))

        # not cartesian
        data = {'seq_1':{'S':[.1,.2,.3,.4], 'L':[1/3,1/3,0,1/3]},\
                'seq_2':{'S':[1/8,0,7/8,0], 'L':[0,0,0,0]}}
        mp1 = MagePoint([.3,.2,.4], Label='seq_1', Radius=0.01)
        mp2 = MagePoint([0,1/3,1/3], Label='seq_1', Radius=0.01)
        mp3 = MagePoint([7/8,0,0], Label='seq_2', Radius=0.01)
        exp_lines = ['@group {RNA} recessiveon',\
            '@spherelist {S} color=orange',\
            str(mp1), str(mp3), '@spherelist {L} color=red', str(mp2)]
        self.assertEqual(str(create_mage_group(data, label='RNA',\
            style='sphere', radius=0.01, cartesian=False)),\
            '\n'.join(exp_lines))

    #def test_kin_single_aln(self):


if __name__ == "__main__":
    main()
