#!/usr/bin/env python

from sys import argv
from string import strip
from cogent.parse.genbank import GbFinder, indent_splitter,\
    block_consolidator, parse_organism, all_chars
from cogent.core.moltype import RNA
#from cogent.core.sequence import Rna
from cogent.core.alignment import Alignment
#from cogent.core.alphabet import RnaAlphabet
from cogent.util.transform import make_trans
from cogent.parse.record_finder import LabeledRecordFinder

def definition_parser(lines):
    line = ' '.join(map(strip,lines))
    without_label = line.lstrip('DEFINITION')
    return without_label.strip(': ')

def definition_adaptor(lines, curr):
    curr['definition'] = definition_parser(lines)

def accession_parser(line):
    without_label = line.lstrip('ACCESSION')
    return without_label.strip(': ')

def accession_adaptor(lines, curr):
    curr['accession'] = accession_parser(lines[0])

source_label_marker = ' ' * 6
source_field_finder = LabeledRecordFinder(lambda x: \
    not x.startswith(source_label_marker), constructor=None)

def source_parser(lines):
    """Simple parser for source fields."""
    result = {}
    all_lines = list(lines)
    source_field = source_field_finder(all_lines).next()
    label, data = block_consolidator(source_field)
    result[label.lower()] = ' '.join(map(strip, data))
    source_length = len(source_field)
    species, taxonomy = parse_organism(lines[source_length:])
    result['species'] = species
    result['taxonomy'] = taxonomy
    return result

def source_adaptor(lines, curr):
    curr.update(source_parser(lines))

def sequence_parser(lines, constructor=''.join):
    """Parses a GenBank sequence block. Doesn't care about ORIGIN line."""
    result = []
    for i in lines:
        if i.startswith('ORIGIN'):
            continue
        result.append(i.translate(all_chars, '0123456789 \t\n\r'))
    return constructor(result)

def sequence_adaptor(lines, curr):
    curr['sequence'] = sequence_parser(lines)

def CrwGenbankParser(lines):
    for rec in GbFinder(lines):
        curr = {}
        for field in indent_splitter(rec):
            #print "FIELD:\n", field
            first_word = field[0].split(None, 1)[0]
            #print "FIRST WORD:", first_word
            if first_word.startswith('ORIGIN'):
                sequence_adaptor(field, curr)
            elif first_word.startswith('ACCESSION'):
                accession_adaptor(field, curr)
            elif first_word.startswith('DEFINITION'):
                definition_adaptor(field, curr)
            elif first_word.startswith('SOURCE'):
                source_adaptor(field, curr)
            else:
                continue
        yield curr

#GUTELL_SYM contains all the symbols that are found in the alignments
#the last three (+*/) are not documented in the format description, but
#occurr regularly. The + and * always occur in places where the other
#sequences have gaps, the slash occurs in other places as well, but I 
#decided to translate all three symbols into gaps.
#In addition, a single ' and T were found, but they
#won't be translated
GUTELL_SYM = dict.fromkeys(".~|()<>[]+*/")
aln_trans = make_trans(".~|()<>[]+*/","??----------")

def crw_seq_to_rna(seq, trans_table=aln_trans, to_upper=True,\
    constructor=RNA.Sequence):
    """Returns Rna sequence constructed from CRW input
    
    Because the constructor is Rna characters that aren't in the RnaAlphabet
    will silently be stripped out or be converted. If you want to prevent this,
    use another constructor or check for non Rna characters beforehand.
    """
    if to_upper:
        return constructor(seq.translate(trans_table).upper()) 
    else:
        return constructor(seq.translate(trans_table))

def crw_to_alignment(lines, trans_table=aln_trans, to_upper=True,\
    seq_constructor=RNA.Sequence, strict=True):
    
    alpha = RNA.Alphabets.DegenGapped
    res = {} #result dict, will become alignment
    acc = {} #dict of accession numbers
    
    for rec in CrwGenbankParser(lines):
        defi = '_'.join(map(strip,rec['definition'].split()))
        access = '_'.join(map(strip,rec['accession'].split()))
        if (defi,access) in acc: #duplicate entry
            continue
        else:
            acc[(defi,access)] = None
        key = '_'.join([defi, access])
        if to_upper:
            seq = rec['sequence'].upper()
        else:
            seq = rec['sequence']
        bad_symbol_found=False
        for i in seq:
            if not i in alpha and not i in GUTELL_SYM:
                if strict:
                    raise ValueError, "Symbol %s not allowed"%(i)
                else:
                    bad_symbol_found=True
                    break
        if bad_symbol_found:
            continue
        if key in res:
            raise ValueError, "Key already exists %s and seq not identical"\
                    %(key)
        if trans_table is None:
            res[key] = rec['sequence']
        else:
            res[key] = crw_seq_to_rna(rec['sequence'], trans_table,\
                to_upper, seq_constructor)
    return Alignment(res)

def crw_to_fasta(lines, trans_table=aln_trans, to_upper=True,\
    seq_constructor=RNA.Sequence, strict=True):

    return crw_to_alignment(lines, trans_table, to_upper, seq_constructor,\
        strict).toFasta()

def aln_statistics(lines):
    symbols = {}
    for rec in CrwGenbankParser(lines):
        defi = rec['definition']
        seq = rec['sequence']
        for sym in seq:
            if sym not in symbols:
                symbols[sym] = 0
            symbols[sym] += 1
    #for k, v in symbols.items():
    #    print k, v
    return symbols
        


if __name__ == "__main__":

    count = 0
    for r in GutellAlnGenbankParser(open(argv[1])):
        count += 1
        #print r
        #continue
        #print r['definition']
        #print r['accession']
        #print r['sequence']
    print count
