#!/usr/bin/env python
# ct.py

"""
Author: Sandra Smit (Sandra.Smit@colorado.edu)

Revision History:
File created on 11 Dec 2006.

"""
from __future__ import division
from cogent.parse.record_finder import LabeledRecordFinder
from cogent.struct.rna2d import Pairs
from cogent.core.moltype import RNA
from cogent.core.info import Info
from cogent.struct.pairs_util import adjust_base

class CtError(Exception):
    pass

def is_ct_line(line):
    """Return True if line is header line
    
    line -- str, single line

    Header line is recognized by one of the following words:
    dG, ENERGY, or Energy
    """
    if 'dG' in line:
        return True
    if 'ENERGY' in line or 'Energy' in line:
        return True
    return False

# Record finder
ct_finder = LabeledRecordFinder(is_ct_line)


def ct_head_body(lines):
    """Separate first line (header) from the rest (seq/struct info)

    lines -- list of lines or anything that behaves like it, e.g.
        filestream.
    """
    head = lines[0]
    body = lines[1:]
    return head, body

# =============================================================================
# PARSE FUNCTIONS FOR HEADER LINES
# =============================================================================

def ct_parse_header(line):
    """Return simple string of header"""
    return line.strip()

def bralibase_parse_header(line):
    """Return tuple of (sequence_length, structure_energy, sequence_label)

    line -- str, single header line or structure record in CT format

    Parse header from BRALIBASE data
    """
    seq_len = None
    energy = None
    seq_label = None

    seq_len, rest = line.split(None,1)
    seq_len = int(seq_len)
    energy_label, rest = rest.split('=',1)
    energy_label = energy_label.strip()
    if energy_label == 'ENERGY' or energy_label == 'Energy':
        # assume Energy = xx.xx <species name>
        parts = rest.split(None,1)
        try:
            energy = float(parts[0].strip())
        except ValueError:
            raise CtError(\
                "Energy is not a floating point: %s"%(parts[0].strip()))
        if len(parts) >= 2:
            seq_label = parts[1].strip()
        else: # no sequence label present
            pass
    else:
        raise CtError("Unknown BRALIBASE header line: %s"%(line))

    return seq_len, energy, seq_label
    

def unafold_parse_header(line):
    """Parse UNAfold formatted header

    UNAfold format is: sequence lenth, followed by dG = -x.xxx, followed by
    sequence name. 
    Can handle tab delimited or space delimited.
    """
    try:
        parts = line.split('\t',2)
        print 'tab parts', parts
        assert len(parts) == 3
        assert '=' in parts[1]
        seq_len = int(parts[0])
        seq_label = parts[2]
        energy = float(parts[1].split('=')[-1])
    except AssertionError:
        try:
            parts = line.split()
            print 'space parts', parts
            assert parts[2] == "="
            seq_len = int(parts[0])
            energy = float(parts[3])
            seq_label = ' '.join(parts[4:])
        except AssertionError:
            raise CtError("Can't handle label")
            
    print 'header', seq_len, energy, seq_label
    return None, None, None, None

# =============================================================================
# PARSE FUNCTIONS FOR CONTENT (STRUCTURE) LINES
# =============================================================================

def ct_parse_content(lines):
    """Return tuple of (raw_sequence, Pairs object)
   
    lines -- list of lines or anything behaving like it

    This function parses seq/struct lines of files in .ct format.
    Return value is tuple of sequence and Pairs object
        raw_sequence is simple string of characters
        Pairs object will be zero-based. 

    This method should be universal for .ct files from different sources.
    In general they only differ in their header.
    
    Assumes 1-based, and returns 0-based structure object, corresponding
        to the sequence.
    Not yet: (Parses standard 6-column format.)
    """
    # check assumption (1-based numbering)
    try:
        first_char = int(lines[0].strip().split()[0])
        if first_char != 1:
            raise CtError("First content line starts with %s, should be 1"\
                %(first_char))
    except ValueError:
        raise CtError("First content line doesn't start with an int: %s"\
            %(lines[0].strip()))

    seq = []
    struct = Pairs()
    last_line = None
    for line in lines:
        parts = line.strip().split()
        if len(parts) != 6:
            raise CtError("Number of parts in line is %s, expected 6"\
                %(len(parts)))
        seq.append(parts[1])
        up, down = int(parts[0]), int(parts[4])

        # check file consistency
        if last_line is not None and up != last_line +1:
            raise CtError("Numbers out of order: %s after %s"%(up, last_line))
        if int(parts[2]) != up-1 or int(parts[3]) != up+1 or\
            int(parts[5]) != up:
            raise CtError("Inconsistency in line: %s"%(line.strip()))
        # if everything is ok, append
        if down == 0:
            struct.append((up, None))
        if down != 0:
            struct.append((up, down))
        last_line = up
    if struct.hasConflicts():
        raise CtError("Stucture has conflicts")
    struct = adjust_base(struct.directed(), -1)
    seq = ''.join(seq)

    return seq, struct

def unafold_parse_content(lines):
    """Return tuple of (raw_sequence, Pairs object)
   
    lines -- list of lines or anything behaving like it

    This function parses seq/struct lines of files in .ct format.
    Return value is tuple of sequence and Pairs object
        raw_sequence is simple string of characters
        Pairs object will be zero-based. 

    This method should be universal for .ct files from different sources.
    In general they only differ in their header.
    
    Assumes 1-based, and returns 0-based structure object, corresponding
        to the sequence.
    Not yet: (Parses standard 6-column format.)
    """
    num_lines = len(lines)
    print num_lines
    # check assumption (1-based numbering)
    try:
        first_char = int(lines[0].strip().split()[0])
        if first_char != 1:
            raise CtError("First content line starts with %s, should be 1"\
                %(first_char))
    except ValueError:
        raise CtError("First content line doesn't start with an int: %s"\
            %(lines[0].strip()))

    seq = []
    struct = Pairs()
    last_line = None
    for line in lines:
        parts = line.strip().split()
        if len(parts) != 8:
            raise CtError("Number of parts in line is %s, expected 8"\
                %(len(parts)))
        seq.append(parts[1])
        up, down = int(parts[0]), int(parts[4])

        # check file consistency
        if last_line is not None and up != last_line +1:
            raise CtError("Numbers out of order: %s after %s"%(up, last_line))
        if up < num_lines: 
            if int(parts[2]) != up-1 or int(parts[3]) != up+1 or\
                int(parts[5]) != up:
                raise CtError("Inconsistency in line: %s"%(line.strip()))
        if up == num_lines:
            print 'up', up
            if int(parts[2]) != up-1 or int(parts[3]) != 0 or\
                int(parts[5]) != up:
                raise CtError("Inconsistency in line: %s"%(line.strip()))
        # if everything is ok, append
        if down == 0:
            struct.append((up, None))
        if down != 0:
            struct.append((up, down))
        last_line = up
    if struct.hasConflicts():
        raise CtError("Stucture has conflicts")
    struct = adjust_base(struct.directed(), -1)
    seq = ''.join(seq)

    return seq, struct


# =============================================================================
# HEADER AND BODY INTEGRATION
# =============================================================================

def ct_integrate(header_info, body_info):
    try:
        header_line = header_info
        raw_seq, pairs = body_info
    except ValueError:
        raise CtError("Unexpected data")
    info = Info({'SEQ_STRUCT': header_line})
    seq = RNA.Sequence(raw_seq, Info=info)
    return seq, pairs

def integrate_bralibase(header_info, body_info):
    """
    header_info -- tuple of (seq_len, energy, seq_label), result
        of bralibase_parse_header function
    body_info -- tuple of (raw_sequence, pairs)
    """
    try:
        raw_seq, pairs = body_info
        seq_len, energy, seq_label = header_info
    except ValueError:
        raise CtError("Unexpected data")

    seq = RNA.Sequence(raw_seq, Name=seq_label)
    if len(seq) != seq_len:
        raise CtError(\
            "Inconsisten lengths: reported length=%s, sequence length=%s"\
            %(seq_len, len(seq)))
    if energy is not None:
        pairs.Energy=energy
    pairs.sort()
    return seq, pairs

# =============================================================================
# MINIMAL AND FULL PARSER AND SPECIFIC APPLICATION SUPPORT
# =============================================================================


supported =\
    {'generic':{'REC_FINDER':ct_finder,\
                'HEAD_BODY':ct_head_body,\
                'HEADER_PARSER': ct_parse_header,\
                'CONTENT_PARSER': ct_parse_content,\
                'INTEGRATION': ct_integrate},
    'unafold':{'REC_FINDER':ct_finder,\
                'HEAD_BODY':ct_head_body,\
                'HEADER_PARSER':unafold_parse_header,\
                'CONTENT_PARSER': unafold_parse_content,\
                'INTEGRATION': ct_integrate},
    'naview':{'REC_FINDER':ct_finder,\
                'HEAD_BODY':ct_head_body,\
                'HEADER_PARSER':unafold_parse_header},
    'sfold':{'REC_FINDER':ct_finder,\
                'HEAD_BODY':ct_head_body,\
                'HEADER_PARSER':unafold_parse_header},
    'RNaseP_db':{'REC_FINDER':ct_finder,\
                'HEAD_BODY':ct_head_body,\
                'HEADER_PARSER':unafold_parse_header},
    'bralibase':{'REC_FINDER':ct_finder,\
                'HEAD_BODY':ct_head_body,\
                'HEADER_PARSER': bralibase_parse_header,\
                'CONTENT_PARSER': ct_parse_content,\
                'INTEGRATION': integrate_bralibase}}

def MinimalCtParser(lines, program='generic', strict=True):
    """Return pure data as is"""
    for rec in supported[program]['REC_FINDER'](lines):
        try:
            head, body = supported[program]['HEAD_BODY'](rec)
            header_info = supported[program]['HEADER_PARSER'](head)
            content_info = supported[program]['CONTENT_PARSER'](body)
            yield header_info, content_info
        except CtError, e:
            if strict:
                raise CtError(str(e))
            else:
                continue

def CtParser(lines, program='generic', strict=True):
    """Return RNA sequence object etc.
    """
    for header_info, content_info in MinimalCtParser(lines, program=program,\
        strict=strict):
        try:
            sequence, pairs = supported[program]['INTEGRATION']\
                (header_info, content_info)
            yield sequence, pairs
        except CtError, e:
            if strict:
                raise CtError(str(e))
            else:
                continue


if __name__ == "__main__":
    from sys import argv
    #for lines in CTRecordFinder(open(argv[1],'U')):
    #    print lines

    for res in MinimalCtParser(open(argv[1],'U')):
        print res
