#!/usr/bin/env python
"""Provides parser for PDB formatted files

Owner: Sandra Smit (Sandra.Smit@colorado.edu)

Status: Development

Revision History:
Feb 06: Sandra Smit: basic parser written
3/29/06 Sandra Smit: rewritten parsers for coordinate section.

A PDB file is build-up of several sections. At the moment the following
parts are parsed:
Title section (some):
    HEADER
    TITLE
Primary structure section (all):
    DBREF
    SEQADV
    SEQRES
    MODRES
Heterogen section (none)
Secondary structure section (none)
Connectivity annotation section (none)
Miscellaneous features section (none)
Crystallographic and coordinate transformation section (none)
Coordinate section (all):
    MODEL
    ATOM
    SIGATM
    ANISOU
    SIGUIJ
    TER
    HETATM
    ENDMDL
Connectivity section (none)
Bookkeeping section (none)
"""
from string import strip
from Numeric import array
from cogent.parse.record import RecordError
from cogent.parse.record_finder import LabeledRecordFinder, TailedRecordFinder
from rnalib.parse.pdb_model import STANDARD_RES, DbRef, Atom, Residue,\
    Chain, Model, Models

#TITLE SECTION
def header_parser(lines):
    """Returns information on the PDB file: classification, date, PDB ID
    
    Doesn't check whether spacing is correct. If they mess up the spacing,
    and e.g PDB_ID is '', this function just returns '' in the dict.
    
    Should we catch possible errors and extend the PDB_ID search to [59:66]?
    """
    if not len(lines) == 1:
        raise RecordError,\
        'header_parser: PDB file should contain only one header line'
    l = lines[0]
    classification, date, pdb_id = map(strip,[l[10:50], l[50:59], l[62:66]])
    return {'CLASSIFICATION':classification,'DATE':date,'PDB_ID':pdb_id}

def title_parser(lines):
    """Parses TITLE line: returns title string"""
    return ''.join([l[10:79] for l in lines])

def expdta_parser(lines):
    """Parses the EXP line: returns string describing experimental techniques
    """
    return ''.join([l[10:70] for l in lines])

def remark_parser(lines):
    """Creates line_dict from REMARK lines and parses each remark individually
    """
    line_dict = {}
    for line in lines:
        remark_num = line[7:10].strip()
        if remark_num not in line_dict:
            line_dict[remark_num] = []
        line_dict[remark_num].append(line)
    return parse_line_dict(line_dict, strict=False)

def remark_210_parser(lines):
    """Parses REMARK 210 lines: contains NMR experimental information

    Only parses the BEST REPRESENTATIVE CONFORMER IN THIS ENSEMBLE.
    Could be a single integer, several integers separated by '/', or NULL.
    When the description contains several integers separated by slashes,
    the best representative is set to the first model (same behavior as
    RnaView).
    """
    result = {}
    for line in lines:
        descr = line[11:70]
        if descr.startswith('BEST REPRESENTATIVE'):
            model_id = descr.split(':')[-1].strip()
            if '/' in model_id:
                model_id = model_id.split('/')[0].strip()
            if model_id == 'NULL':
                model_id = None
            result['BEST_REPRESENTATIVE'] = model_id
    return result
    
#END TITLE SECTION

#PRIMARY STRUCTURE SECTION
def dbref_parser(lines):
    """Returns Dbref object containing all database reference information
    """
    result = []
    for l in lines:
        pdb_id, chain_id, pdb_from, pdb_to, db_name, db_accession,\
            db_id_code, db_from, db_to = map(strip,[l[7:11],l[12],l[14:19],\
            l[20:25],l[26:32],l[33:41],l[42:54],l[55:61],l[62:68]])
        result.append(DbRef(pdb_id, chain_id, pdb_from, pdb_to, db_name,\
            db_accession, db_id_code, db_from, db_to))
    return result

def seqadv_parser(lines):
    """Returns dict of {(chain_id, res_id, res_name): db_res}
    """
    result = {}
    for l in lines:
        res_name = l[12:15].strip()
        chain_id = l[16].strip() or None
        res_id = l[18:23].strip()
        if not res_id:
            continue
        db_res = l[39:42].strip() or None
        result[(chain_id, res_id, res_name)] = db_res    
    return result

def modres_parser(lines):
    """Returns dict of {(chain_id, res_id, res_name): (standard_res, meaning)}
    """
    result = {}
    for line in lines:
        res_name = line[12:15].strip()
        chain_id = line[16].strip() or None
        res_id = line[18:23].strip()
        standard_res = line[24:27].strip() or None
        meaning = line[29:70].strip() or None
        if standard_res:
            result[(chain_id, res_id, res_name)] = (standard_res, meaning)
    return result

def seqres_parser(lines):
    """Returns dict of {chainID:[residues]}

    Checks wether the number of residues is the same as reported.
    'Residues' is a list of residue strings. if there are no residues the value
    in the dict will be None.
    """
    per_chain = {}
    lengths = {}
    for l in lines:
        ser_num = int(l[8:10].strip())
        chain_id = l[11].strip() or None
        num_res = int(l[13:17].strip())
        res = l[19:70]
        if ser_num == 0: #no seqres available
            if not res[0:3] == 'UNK':
                raise RecordError,\
                    "First res in list should be 'UNK' for unknown residues"
            per_chain[chain_id] = None
        else:
            if chain_id not in lengths:
                lengths[chain_id] = num_res
            res_list = map(strip,res.split())
            try:
                per_chain[chain_id].extend(res_list)
            except KeyError:
                per_chain[chain_id] = res_list
    for chain in per_chain:
        if per_chain[chain]:
            if not len(per_chain[chain]) == lengths[chain]:
                raise RecordError,\
                    "Actual number of residues doesn't match the numRes'+\
                    ' number reported"
    return per_chain

#END PRIMARY STRUCTURE SECTION


#COORDINATE SECTION

def is_chain_terminator(line):
    """Returns True if line terminates a chain (starts with TER)"""
    if line.startswith('TER'):
        return True
    return False

def is_model_label(line):
    """Returns True if line is the start of a model (starts with MODEL)
    """
    if line.startswith('MODEL'):
        return True
    return False
    
ModelFinder = LabeledRecordFinder(is_model_label)
ChainFinder = TailedRecordFinder(is_chain_terminator, strict=False)

def anisou_parser(line):
    """Parses ANISOU line: returns dict of features."""
    l = line
    features = map(int,map(strip,[l[28:35], l[35:42], l[42:49], l[49:56],\
        l[56:63], l[63:70]]))
    return dict(zip(['U00','U11','U22','U01','U02','U12'],features))

def siguij_parser(line):
    """Parses SIGUIJ line: returns dict of features."""
    l = line
    features = map(int,map(strip,[l[28:35], l[35:42], l[42:49], l[49:56],\
        l[56:63], l[63:70]]))
    return dict(zip(['S00','S11','S22','S01','S02','S12'],features))

def sigatm_parser(line):
    """Parses SIGATM line: returns dict of features."""
    l = line
    features = map(float,map(strip,\
        [l[30:38],l[38:46],l[46:54], l[54:60], l[60:66]]))
    return dict(zip(['SigX','SigY','SigZ','SigOcc','SigTemp'],features))

def atom_parser(atom_line, model_id):
    """Returns Atom object created from ATOM or HETATM line in PDB record.

    atom_line: atom line from PDB file (string) starting with ATOM or HETATM.
    model_id: string containing the identifier of the current model
    """
    l = atom_line
    chain_id = l[21].strip() or None
    res_seq = int(l[22:26].strip())
    icode = l[26].strip()
    res_id = str(res_seq) + icode
    atom_id = l[6:11].strip()
    atom_name = l[12:16].strip()
    alt_loc = l[16].strip() or None
    res_name = l[17:20].strip() 
    
    #coordinates and occupancy, temp_factor should always be there
    coords = array(map(float,map(strip,[l[30:38],l[38:46],l[46:54]])))
    occupancy, temp_factor = map(float,[l[54:60], l[60:66]])
    
    seg_id = l[72:76].strip() or None
    element = l[76:78].strip() or None
    charge = l[78:80].strip() or None
    
    return Atom(ModelId=model_id, ChainId=chain_id, ResId=res_id,\
        AtomId=atom_id, AtomName=atom_name, AltLoc=alt_loc, ResName=res_name,\
        ResSeq=res_seq, Icode=icode, Coordinates=coords, Occupancy=occupancy,\
        TempFactor=temp_factor, SegId=seg_id, Element=element, Charge=charge)

def parse_atom_lines(lines, model_id):
    """Parses list of lines belonging to the same atom: returns Atom object.

    lines: list of lines containing to a singe atom. Has to start with a 
        ATOM or HETATM line, can be followed by SIGATM, ANISOU, or SIGUIJ lines
    model_id: string containing the identifier of the current model.
    """
    atom_attr = {'SIGATM':'ParamStdev','ANISOU':'AnisoTemp',\
        'SIGUIJ':'AnisoTempStdev'}
    atom = atom_parser(lines[0], model_id)
    for line in lines[1:]:
        label = line[0:6].strip()
        setattr(atom, atom_attr[label], _parsers[label](line))
    return atom
     
def group_by_atom_serial(lines, model_id):
    """Returns dict of {atom_serial (int): Atom object}

    lines: list of lines belonging to the same chain
    model_id: string containing identifier of the current model
    """
    result = {}
    for line in lines:
        atom_serial = int(line[6:11].strip())
        if atom_serial not in result:
            result[atom_serial] = []
        result[atom_serial].append(line)
    for k, v in result.items():
        result[k] = parse_atom_lines(v, model_id)
    return result
 
def group_by_residue(atom_dict):
    """Takes {atom_serial: Atom} and returns a list of Residue objects.

    atom_dict: dict of {atom_serial: Atom object}
    """
    seen_residues = {}
    residues = []
    keys = atom_dict.keys()
    keys.sort()
    res_pos = -1
    prev_id, prev_name = None, None
    for k in keys:
        a = atom_dict[k]
        if (a.ResId, a.ResName) != (prev_id, prev_name):
            if residues:
                seen_residues[residues[-1].ResId] = None
            res_pos += 1
            r = Residue(a.ModelId, a.ChainId, a.ResId, a.ResName,\
                ResSeq=a.ResSeq, Icode=a.Icode, SeqPos=res_pos)
            residues.append(r)
        if residues[-1].ResId == a.ResId and a.ResId not in seen_residues:
            residues[-1].Atoms.append(a)
        else:
            raise RecordError, "Corrupt chain, failed on atom: %s"%(a.AtomId)
        prev_id, prev_name = a.ResId, a.ResName
    return residues
        

def parse_chain(lines, model_id):
    """Returns Chain object from lines belonging to the same chain
    
    lines: list of lines belonging to the same chain
    model_id: string containing the identifier of the current Model
    """
    atom_dict = group_by_atom_serial(lines, model_id)
    residues = group_by_residue(atom_dict)
    chain_id = residues[0].ChainId
    return Chain(model_id, chain_id, residues)

def parse_model(lines):
    """Returns Model object from lines belonging to the same Model
    
    lines: list of lines belonging to the same PDB Model
    """
    if lines[0].startswith('MODEL') and lines[-1].startswith('ENDMDL'):
        model_id = lines[0][10:14].strip()
        lines = lines[1:-1]
    else:
        model_id = None
    
    chains = []
    chain_order = []
    model_atoms = None
    
    for rec in ChainFinder(lines):
        if rec[-1].startswith('TER'):
            c = parse_chain(rec[:-1], model_id)
            chains.append(c)
            chain_order.append(c.ChainId)
        else:
            #Model Atoms 
            model_atom_dict = group_by_atom_serial(rec, model_id)
            atom_tuples = model_atom_dict.items()
            atom_tuples.sort()
            model_atoms = [v for k,v in atom_tuples]
    result = Model(model_id, ChainOrder=chain_order,\
        ModelAtoms=model_atoms, chains=chains)
    return result

def coordinate_parser(lines):
    """Parses the Coordinate section of a PDB file, returns Models object

    lines: list of lines, containing the entire coordinate section of
        a PDB file
    """
    result = Models()
    for model_lines in ModelFinder(lines):
        m = parse_model(model_lines)
        result[m.ModelId] = m
    return result

#END COORDINATE SECTION
        
_parsers = {'HEADER': header_parser,
            'TITLE' : title_parser,
            'EXPDTA': expdta_parser,
            'REMARK': remark_parser,
            'DBREF' : dbref_parser,
            'SEQRES': seqres_parser,
            'SEQADV': seqadv_parser,
            'MODRES': modres_parser,
            'MODELS': coordinate_parser,
            'SIGATM': sigatm_parser,
            'ANISOU': anisou_parser,
            'SIGUIJ': siguij_parser,
            '210'   : remark_210_parser,
            }

def parse_line_dict(line_dict, strict=True):
    """Parses a dictionary of header lines
    
    line_dict: dictionary of lines grouped by their key. For example:
        {'TITLE':['abc'],'SEQRES':['13 A U G C','13 U C C A']}
    strict: boolean indicating whether an error should be raised on an
        exception or whether unparsable lines should be returned in their
        original state.

    The parser is determined by the _parsers dictionary above. If the key
    is not in the parser dictionary, the lines are added to the result unparsed
    """
    result = {}
    for key in line_dict:
        try:
            parser = _parsers[key]
        except KeyError: #no parser yet
            #just throw in the data as is...
            result[key] = line_dict[key]
            continue
        try:
            result[key] = parser(line_dict[key])
        except RecordError, e:
            if strict:
                raise RecordError, str(e)
            else:
                #unparsable, return the original lines
                result[key] = line_dict[key]
    return result

def MinimalPdbParser(infile):
    """Returns a dictionary of lines grouped by their key

    infile: something that behaves as a list of PDB lines.

    Most lines are grouped by their actual key. All lines in the Coordinate
    section of the PDB file are grouped under the key 'MODELS'.
    """
    line_dict = {}
    for line in infile:
        label = line[0:6].rstrip()
        if label in ['MODEL','ATOM','SIGATM','ANISOU','SIGUIJ','TER',\
            'HETATM', 'ENDMDL']:
            label = 'MODELS'
        if not label: #assume it's empty line and continue
            continue
        else:
            line = line.strip()
            try:
                line_dict[label].append(line)
            except KeyError:
                line_dict[label] = [line]
    return line_dict


def PdbParser(lines, result_constructor=parse_line_dict, strict=True,\
    set_standard=False, standard_res=STANDARD_RES, specific_changes=None,\
    general_changes=None):
    """Parser for the PDB data format

    lines: input data (list of lines or file stream)
    result_constructor: function to process the header information. Default
        is 'parse_header'
    strict: whether an exception should be raised in case of a problem
        (strict=True) or whether the bad record should be skipped
        (strict=False).


    If set_standard=True:
        
        It'll try to set the DbResidues if SEQADV is in the result
    """   
    line_dict = MinimalPdbParser(lines)
    try:
        result = result_constructor(line_dict, strict=strict)
    except RecordError, e:
        if strict:
            raise RecordError, str(e)
        else:
            return {}
    
    if set_standard:
        m = result['MODELS']

        if specific_changes is None: #see what's in MODRES
            if 'MODRES' in result:
                modres = result['MODRES']
                specific_changes = dict([(k,v[0]) for (k,v) in modres.items()])
        else:
            pass #user preference has priority

        m.setStandardResidues(standard_res=standard_res,\
            specific_changes=specific_changes, general_changes=general_changes)
            
        if 'SEQADV' in result:
            m.setDbResidues(result['SEQADV'])
        
    return result

#=============================================================================
#CODE TO FIX RESIDUE INSERTIONS
#=============================================================================
def is_insertion_free(lines):
    """Returns True if it found insertion codes in Coordinate section of PDB
    """
    line_dict = MinimalPdbParser(lines)
    for line in line_dict['MODELS']:
        try:
            icode = line[26].strip()    
            if icode:
                return False
        except IndexError:
            continue
    return True

def _get_normal_and_inserted(lines):
    """Returns dictionaries of normal and inserted residues in PDB file.
    
    normal is dict of {ChainId:[ResSeq]}
    inserted is dict of {ChainId: {(ResSeq, Icode): new res_seq}}

    If multiple models are present in the PDB file, the residues in each
    model should be the same. If not, an AssertionError will be raised.
    """
    m = PdbParser(lines)['MODELS']
    
    normal = {} #{ChainId:[ResSeq]}
    inserted = {} #{ChainId: {(ResSeq, Icode): new res_seq}}
    
    res_dict = {}
    
    prev_normal = None
    prev_inserted = None
    for model in m.values():
        normal = {} #{ChainId:[ResSeq]}
        inserted = {} #{ChainId: {(ResSeq, Icode): new res_seq}}
        res = {}
        for r in model.iterResidues():
            if not r.Icode:
                try:
                    normal[r.ChainId].append(r.ResSeq)
                except KeyError:
                    normal[r.ChainId] = [r.ResSeq]
            else:
                if r.ChainId not in inserted:
                    inserted[r.ChainId] = {}
                inserted[r.ChainId][(r.ResSeq, r.Icode)] = None
        if prev_normal:
            assert normal == prev_normal
        prev_normal = normal
        if prev_inserted:
            assert prev_inserted == inserted
        prev_inserted = inserted
    for c in inserted:
        start = 9999
        for item in inserted[c]:
            if normal:
                while start in normal[c]:
                    start -= 1
            if start < 0:
                raise ValueError("Sorry, can't fix this file")
            inserted[c][item] = start
            start -= 1
            
    return prev_normal, prev_inserted

def fix_insertions(lines):
    """Returns PDB string (that is PDB file in string format).
    
    Will fix insertion codes in PDB file (replace ResId by high integer).
    Just write the output to a file and you're done.

    Purpose: RNAVIEW can't handle residue insertion codes, it just ignores
        them. This means that if 131, 131L, 131M, and 131N are in your file
        and some or all of these are involved in base pairing, the RNAVIEW
        output will be ambiguous.

    This method will map each residue with an insertion code onto a high
    integer not used in that chain yet. Starting at 9999 and working down.

    If multiple models are present in the PDB file, the residues in each
    model should be the same. If not, an AssertionError will be raised.

    It fixes the insertion codes in the following fields:
    ['ATOM','HETATM','SIGATM','ANISOU','SIGUIJ','MODRES','SEQADV']
    If you need them corrected somewhere else, add them to this function.
    """
    result_lines = []
    normal, inserted = _get_normal_and_inserted(lines)
    for line in lines:
        label = line[0:6].strip()
        if label in ['ATOM','HETATM','SIGATM','ANISOU','SIGUIJ']:
            chain_id = line[21].strip() or None
            res_seq = int(line[22:26].strip())
            icode = line[26].strip()
            if not icode:
                result_lines.append(line.strip())
                continue
            first, rs, ic, second = line[0:22], line[22:26], line[26],\
                    line[27:].rstrip()
        elif label in ['MODRES','SEQADV']:
            chain_id = line[16].strip() or None
            try:
                res_seq = int(line[18:22].strip())
            except ValueError:
                result_lines.append(line.strip())
                continue
            icode = line[22].strip()
            if not icode:
                result_lines.append(line.strip())
                continue
            first, rs, ic, second = line[0:18], line[18:22], line[22],\
                    line[23:].rstrip()    
        else:
            result_lines.append(line.strip())
            continue
        new_rs = str(inserted[chain_id][(res_seq, icode)])
        new_rs = ' '*(4-len(new_rs)) + new_rs
        new_ic = ' '
        result_lines.append(''.join([first, new_rs, new_ic, second]))
    return '\n'.join(result_lines)



if __name__ == "__main__":
    from sys import argv
    #f = open('new_pdbs/1EHZ.pdb')
    #f = open('new_pdbs/2LDZ_leadzyme.pdb')
    f = open(argv[1])
    res = PdbParser(f, set_standard=True)
    f.close()
    
    m = res['MODELS']
     
    for model in m:
        print 'MODEL', model
        for r in m[model].iterResidues():
            print '\t'.join(map(str,[r.ChainId, r.ResId, r.ResName,\
                r.ResType, r.BaseType, r.StandardRes, r.DbRes]))
            #for a in r.Atoms:
            #    if a.AltLoc is not None:
            #        print a

        #for a in m[model].ModelAtoms:
        #    print a, a.AnisoTemp, a.AltLoc
        for chain in m[model].values():
            print chain.seq(attr_name='StandardRes',one_letter=False)
    
