#!/usr/bin/env python
# rnaview.py

"""
Provides RnaView objects and RnaViewParser

Original author :Greg Caporaso
Rewritten by Sandra Smit

Status: development

Revision History
5/18-19/04 Greg Caporaso: File creation, creation of all objects used to store 
data, ie.Base, BasePair, BasePairs
5/20/04 Greg Caporaso: Added capabilites for detecting Wobble Pairs to 
support objects, changed BpList object name to BasePairs; added 
toPairsFromWobble, toPairsFromWC to BasePairs object; some parser 
methods written
5/21/04 Greg Caporaso: Completed many private methods of parser
5/24/04 Greg Caporaso: Completed public methods of parser, added functionality
of parser to translate modified bases based on translation table in pdb file
5/25/04 Greg Caporaso: Optimizaton and organization of code
5/26/04 Greg Caporaso: Modified _parse_pair_seqs() to be a little neater; 
changed toPairs* methods to raise NotImplementedError rather than BasePairsError
when handed a molecule with more than one sequence
6/3/04 Greg Caporaso: changed some variable names as suggested by Rob
6/8/04 Greg Caporaso: minor modifications based on suggestions
3/25/05 Sandra Smit: changed BasePair interface to parse all annotation fields.
Changed _get_sequences to parse numbers as chain IDs correctly.
9/06 Sandra Smit: changed objects, functionality, and parser.
9/20/06 Sandra Smit: added parser for PDB file name.
"""

from string import upper
from Numeric import add
from old_cogent.base.sequence import Rna, RnaSequence
from cogent.struct.rna2d import Pairs, PseudoknotRemover
from rnalib.parse.pdb_format import PdbParser
from rnalib.parse.pdb_model import STANDARD_RES

RNAVIEW_ACCEPTED = dict.fromkeys('AGUCTIPaguct')
STANDARD_RES = dict.fromkeys('UCAG')
WC_PAIRS = dict.fromkeys(['GC','CG','AU','UA'])
WOBBLE_PAIRS = dict.fromkeys(['UG','GU'])

class RnaViewObjectError(ValueError):
    pass

class RnaViewParseError(SyntaxError):
    pass


class Base(object):

    def __init__(self, ChainId, ResId, ResName):
        """Initializes Base object.

        ChainId = Chain identifier, string. Could be None
        ResId = Residue identifier, string. Cannot be None
        ResName = Name of the residue, string. Accepted are AGUCTIPaguct.
            Cannot be None.
        """
        if not ResId or not ResName:
            raise RnaViewObjectError(\
            "Found missing attribute: ResId=%s, ResName=%s"\
                %(ResId, ResName))
        self.ChainId = ChainId
        self.ResId = ResId
        self.ResName = ResName

    def __str__(self):
        """String representation of Base object"""
        return ' '.join(map(str,[self.ChainId, self.ResId, self.ResName]))

    def __eq__(self, other):
        """Overwrite the == operator"""
        if self.ChainId != other.ChainId:
            return False
        if self.ResId != other.ResId:
            return False
        if self.ResName != other.ResName:
            return False
        return True

    def __ne__(self,other):
        """Overwrite the != operator"""
        return not self == other

class BasePair(object):
    """Object for storing paired RNA bases
    
    This object is made to store two Base objects which are involved in a
    base pairing interaction with each other.
    """

    def __init__(self, Up, Down, Edges=None, Orientation=None,\
        Conformation=None, Saenger=None):
        """ Initialize the object 
            
        Up: the upstream Base object
        Down: the downstream Base object
        Edges: the pairing edges of the base pair or 'stacked'
        Orientation: the orientation of the bases: 'cis' or 'tran'
        Conformation: 'syn' or 'syn syn'
        Saenger: the Saenger classification of the base pair, either
            roman numeral, 'n/a', or something starting with '!'
        """
        self.Up = Up
        self.Down = Down
        self.Edges = Edges
        self.Orientation = Orientation
        self.Conformation = Conformation
        self.Saenger = Saenger

    def __str__(self):
        """String representation of BasePair object"""
        return "U: %s; D: %s; Edges: %s; Orient: %s; Conf: %s; Saenger: %s"\
            %(self.Up, self.Down, self.Edges, self.Orientation,\
            self.Conformation,self.Saenger)

    def __eq__(self,other):
        """Overwrite the == operator"""
        if self.Up != other.Up:
            return False
        if self.Down != other.Down:
            return False
        if self.Edges != other.Edges:
            return False
        if self.Orientation != other.Orientation:
            return False
        if self.Conformation != other.Conformation:
            return False
        if self.Saenger != other.Saenger:
            return False

        return True

    def __ne__(self,other):
        """Overwrite the != operator"""
        return not self == other

    def isWC(self):
        """Returns True if base pair is a Watson-Crick pair.

        WARNING: this method returns True for GC and AU pairs independent
        of the rest of the annotation. It does not check the specific edges
        in the base pair etc.
        """
        # The complicated looking one-liner just makes an upper-case string
        # out of the two Base identities and checks to see if it exists in
        # WC_PAIRS
        return ''.join([self.Up.ResName, self.Down.ResName]).upper()\
            in WC_PAIRS

    def isWobble(self):
        """Returns True if base pair is a wobble pair.

        WARNING: this method returns True for GU pairs independent
        of the rest of the annotation. It does not check the specific edges
        in the base pair etc.
        """
        # The complicated looking one-liner just makes an upper-case string
        # out of the two Base identities and checks to see if it exists in
        # WOBBLE_PAIRS
        return ''.join([self.Up.ResName, self.Down.ResName]).upper()\
            in WOBBLE_PAIRS


class BasePairs(list):

    def __init__(self, base_pairs=None, Model=None):
        """Initializes BasePairs object.

        base_pairs = list or tuple of BasePair objects.
        Model = Model object (extracted from PDB file).
        """
        if base_pairs is None:
            base_pairs = []
        try:
            self[:] = list(base_pairs)
        except TypeError:
            raise RnaViewObjectError(\
                'base_pairs must be convertable to a list')

        self.Model = Model

    def __str__(self):
        """String representation of BasePairs object"""
        return '\n'.join([str(i) for i in self])

    def select(self, selection_function, rest=False):
        """Returns a new BasePairs object of pairs that pass the selection.

        selection_function = should be function that works on a BasePair
            and returns True or False.

        The Model of the new BasePairs object is set to the original Model.
        """
        sel = []
        if rest:
            not_sel = []
        for bp in self:
            if selection_function(bp):
                sel.append(bp)
            else:
                if rest:
                    not_sel.append(bp)
                else:
                    continue
        if rest:
            return BasePairs(sel, Model=self.Model),\
                BasePairs(not_sel, Model=self.Model)
        else:
            return BasePairs(sel, Model=self.Model)

    def _get_present_chains(self):
        """Return the chains present in the BasePairs object (in ORDER)"""
        up_chains = set([i.Up.ChainId for i in self])
        down_chains = set([i.Down.ChainId for i in self])
        result = list(up_chains | down_chains)
        result.sort()
        return result

    PresentChains = property(_get_present_chains)

    def cliques(self):
        """Yields all cliques in the base pairs as new BasePairs objects.

        A clique is a set of interacting chains. All base pairs involved
        in a clique are returned together in a new BasePairs object.
        """
        if len(self.PresentChains) == 1:
            yield self
        else:
            ia = {} #interactions
            for i in self:
                chain_up = i.Up.ChainId
                chain_down = i.Down.ChainId
                if not ia:
                    bp_list = [i]
                    ia[chain_up] = bp_list
                    ia[chain_down] = bp_list
                elif chain_up in ia and not chain_down in ia:
                    ia[chain_up].append(i)
                    ia[chain_down] = ia[chain_up]
                elif chain_down in ia and not chain_up in ia:
                    ia[chain_down].append(i)
                    ia[chain_up] = ia[chain_down]
                elif chain_up not in ia and chain_down not in ia:
                    bp_list = [i]
                    ia[chain_up] = bp_list
                    ia[chain_down] = bp_list
                elif chain_up in ia and chain_down in ia and\
                    ia[chain_up] is ia[chain_down]:
                    ia[chain_up].append(i)
                elif chain_up in ia and chain_down in ia and not\
                    ia[chain_up] is ia[chain_down]:
                    ia[chain_up].append(i)
                    ia[chain_up].extend(ia[chain_down])
                    for k,v in ia.items():
                        if k != chain_down and v is ia[chain_down]:
                            ia[k] = ia[chain_up]
                    ia[chain_down] = ia[chain_up]
                else:
                    continue
            
            uniques = []
            for k,v in ia.items():
                if v not in uniques:
                    uniques.append(v)
            for bps in uniques:
                yield BasePairs(bps, Model=self.Model)

    def base_pair_cliques(self):
        ia = {} #interactions
        for i in self:
            chain_up = (i.Up.ChainId, i.Up.ResId)
            chain_down = (i.Down.ChainId, i.Down.ResId)
            if not ia:
                bp_list = [i]
                ia[chain_up] = bp_list
                ia[chain_down] = bp_list
            elif chain_up in ia and not chain_down in ia:
                ia[chain_up].append(i)
                ia[chain_down] = ia[chain_up]
            elif chain_down in ia and not chain_up in ia:
                ia[chain_down].append(i)
                ia[chain_up] = ia[chain_down]
            elif chain_up not in ia and chain_down not in ia:
                bp_list = [i]
                ia[chain_up] = bp_list
                ia[chain_down] = bp_list
            elif chain_up in ia and chain_down in ia and\
                ia[chain_up] is ia[chain_down]:
                ia[chain_up].append(i)
            elif chain_up in ia and chain_down in ia and not\
                ia[chain_up] is ia[chain_down]:
                ia[chain_up].append(i)
                ia[chain_up].extend(ia[chain_down])
                for k,v in ia.items():
                    if k != chain_down and v is ia[chain_down]:
                        ia[k] = ia[chain_up]
                ia[chain_down] = ia[chain_up]
            else:
                continue
        
        uniques = []
        for k,v in ia.items():
            if v not in uniques:
                uniques.append(v)
        for bps in uniques:
            yield BasePairs(bps, Model=self.Model)
        
    
    def seq(self, requested_chains=None):
        """Returns sequence of standard residues for requested chains
        
        This just passes the request to the associated model. Chains
        are concatenated in the given order.
        """
        m = self.Model
        if m is None:
            raise RnaViewObjectError(\
                "Need PDB Model to extract sequence information")
        if requested_chains is None:
            requested_chains = self.PresentChains
        for c in requested_chains:
            assert c in m
        
        return m.seq(chains=requested_chains, attr_name='StandardRes',\
            one_letter=True)

    def seqAndPairs(self, requested_chains=None, allow_conflicts=True):
        """Returns seq, Pairs.

        requested_chains should be a list of chain_IDs IN ORDER!
        """
        # Necessary information. Cross-check.
        m = self.Model
        if m is None:
            raise RnaViewObjectError(\
                "Need PDB Model to extract sequence information")
        if requested_chains is None:
            requested_chains = self.PresentChains
        for i in requested_chains:
            assert i in m

        lengths = add.accumulate([0]+[len(m[chain]) for chain in\
            requested_chains])

        seq = self.seq(requested_chains)
        
        pairs = []
        seen = {}
        for i in self:
            up = i.Up
            down = i.Down
            up_chain_id = up.ChainId
            down_chain_id = down.ChainId
            if up_chain_id in requested_chains and down_chain_id in\
                requested_chains:
                add_to_up = lengths[requested_chains.index(up_chain_id)]
                add_to_down = lengths[requested_chains.index(down_chain_id)]
                up_seq_pos = m[up_chain_id][up.ResId].SeqPos + add_to_up
                down_seq_pos = m[down_chain_id][down.ResId].SeqPos + add_to_down
                if not allow_conflicts:
                    if up_seq_pos not in seen and down_seq_pos not in seen:
                        seen[up_seq_pos] = None
                        seen[down_seq_pos] = None
                        pairs.append((up_seq_pos, down_seq_pos))
                    else:
                        print 'Base pair conflict: %s'%(str(i))
                        continue
                        #raise RnaViewObjectError(\
                        #    'Base pair conflict: %s'%(str(i)))
                else:
                    pairs.append((up_seq_pos, down_seq_pos))
            else:
                continue
        p = Pairs(pairs)
        p = p.directed()
        return seq, p
   
    def positions(self, requested_chains=None):
        """Returns dict {(up_pos, down_pos): bp}

        requested_chains should be a list of chain_IDs IN ORDER!
        """
        # Necessary information. Cross-check.
        m = self.Model
        if m is None:
            raise RnaViewObjectError(\
                "Need PDB Model to extract sequence information")
        if requested_chains is None:
            requested_chains = self.PresentChains
        for i in requested_chains:
            assert i in m

        lengths = add.accumulate([0]+[len(m[chain]) for chain in\
            requested_chains])

        result = {}
        pairs = []
        for i in self:
            up = i.Up
            down = i.Down
            up_chain_id = up.ChainId
            down_chain_id = down.ChainId
            if up_chain_id in requested_chains and down_chain_id in\
                requested_chains:
                add_to_up = lengths[requested_chains.index(up_chain_id)]
                add_to_down = lengths[requested_chains.index(down_chain_id)]
                up_seq_pos = m[up_chain_id][up.ResId].SeqPos + add_to_up
                down_seq_pos = m[down_chain_id][down.ResId].SeqPos + add_to_down
                if (up_seq_pos, down_seq_pos) in result:
                    raise ValueError("Impossible")
                result[(up_seq_pos, down_seq_pos)] = i
            else:
                continue
        return result

    def seqAndVienna(self, requested_chains=None, pseudoknot_rule='majority'):
        """Returns seq, vienna_structure for requested chains.
        
        requested_chains: list of chain IDs IN ORDER. The order is the order
            in which the different chains will be concatenated.
        pseudoknot_rule: rule that you want to use to remove the pseudoknots.
            'majority' keeps the majority of the base pairs, 'first' keeps the
            first base pairs that it runs into and removes the second helix
            in the pseudoknot.
        
        This is meant as a convenience method, it removes the pseudoknots
        for you and changes the list of base pairs into a Vienna structure.
        """
        seq, pairs = self.seqAndPairs(requested_chains)
        if pairs.hasPseudoknots():
            p = PseudoknotRemover()
            pairs = p(pairs, rule=pseudoknot_rule)
        vienna = pairs.toVienna(len(seq))
        return seq, vienna
            
    def nonConflictingPairs(self):
        nc = {}
        counts = {}
        for bp in self:
            bp_key = (bp.Up.ChainId, bp.Up.ResId,\
                bp.Down.ChainId, bp.Down.ResId)
            nc[bp_key] = bp
            if (bp.Up.ChainId, bp.Up.ResId) not in counts:
                counts[(bp.Up.ChainId, bp.Up.ResId)] = 1
            else:
                counts[(bp.Up.ChainId, bp.Up.ResId)] += 1
            if (bp.Down.ChainId, bp.Down.ResId) not in counts:
                counts[(bp.Down.ChainId, bp.Down.ResId)] = 1
            else:
                counts[(bp.Down.ChainId, bp.Down.ResId)] += 1
        non_conflicting = []
        rest = []
        for k,v in nc.items():
            if counts[k[:2]] == 1 and counts[k[2:]] == 1:
                non_conflicting.append(v)
            else:
                rest.append(v)
        return BasePairs(non_conflicting, Model=self.Model),\
            BasePairs(rest, Model=self.Model)

    def conflicts(self):
        seen = {}
        for bp in self:
            if bp.Up.ResId not in seen:
                seen[bp.Up.ResId] = [bp]
            else:
                seen[bp.Up.ResId].append(bp)
            if bp.Down.ResId not in seen:
                seen[bp.Down.ResId] = [bp]
            else:
                seen[bp.Down.ResId].append(bp)
        
        for k, v in seen.items():
            if len(v) > 1:
                yield v



class BaseMultiplet(list):
    
    def __init__(self, bases=None):
        """Initializes a BaseMultiplet object.

        bases = a list or tuple of Base objects.
        """
        if bases is None:
            bases = []
        try:
            self[:] = list(bases)
        except TypeError:
            raise RnaViewObjectError(\
                'bases must be convertable to a list')
                
    def __str__(self):
        """String representation of a BaseMultiplet object"""
        return ';\t'.join([str(i) for i in self])
       
       
class BaseMultiplets(list):

    def __init__(self, multiplets=None, Model=None):
        """Initializes a BaseMultiplets object.

        multiplets = list or tuple of BaseMultiplet objects.
        Model = Model object (derived from PDB file.
        """
        self.Model = Model
        if multiplets is None:
            multiplets = []
        try:
            self[:] = list(multiplets)
        except TypeError:
            raise RnaViewObjectError(\
                'multiplets must be convertable to a list')

    def __str__(self):
        """String representation of BaseMultiplets object"""
        return '\n'.join([str(i) for i in self])

class PairCounts(dict):
    """A dictionary of base pair counts"""
    pass 


def is_canonical(bp):
    """Returns True if bp is standard canonical GC or AU pair, or wobble GU
    """
    if bp.Edges == '+/+': #GC pair
        return True
    if bp.Edges == '-/-': #AU pair
        return True
    if bp.Edges == 'W/W' and bp.Orientation == 'cis' and\
        bp.Saenger == 'XXVIII': #GU pair
        return True
    return False

def is_not_canonical(bp):
    return not is_canonical(bp)

def is_stacked(bp):
    if bp.Edges == 'stacked':
        return True
    return False

def is_not_stacked(bp):
    if bp.Edges == 'stacked':
        return False
    return True

def is_tertiary(bp):
    if bp.Saenger.startswith('!'):
        return True
    return False

def is_not_stacked_or_tertiary(bp):
    if bp.Edges == 'stacked':
        return False
    if bp.Saenger.startswith('!'):
        return False
    return True

def is_tertiary_base_base(bp):
    if bp.Saenger and bp.Saenger.startswith('!1H(b_b)'):
        return True
    return False

def is_part_of_clique(chains):
    c = dict.fromkeys(chains)
    def apply_to(bp):
        if bp.Up.ChainId in c and bp.Down.ChainId in c:
            return True
        return False
    return apply_to

#==============================================================================
#RNAVIEW PARSER
#==============================================================================

def is_roman_numeral(s):
    """Returns True if s is a roman numeral"""
    #comma in there, because some categories are combined, split by comma
    if not s:
        return False
    alphabet = dict.fromkeys("IVXDCM,")
    for i in s:
        if i not in alphabet:
            return False
    return True

def is_edge(s):
    """Returns True if s is a description of the base pair edges"""
    chars = dict.fromkeys("WHsS+-?.")
    try:
        first, second = s.split('/')
        assert first in chars and second in chars
        return True
    except:
        return False

def is_orientation(s):
    """Returns True if s is a description of the base pair orientation"""
    if s == 'cis' or s == 'tran':
        return True
    return False

def parse_annotation(data):
    """Parses the annotation of a base pair, returns tuple of 4 elements.
    
    data is a single string containing the annotation of a single base pair.
    For example: '-/- cis         XX' or 'stacked'
    Edges: could be 'stacked' or two chars separated by '/', e.g. 'W/H'
    Orientation: could be 'cis' or 'tran'
    Conformation: could be 'syn' or 'syn syn'
    Saenger: could be some roman numeral (or two separated by a comma) or
        'n/a' or anything starting with an exclamation mark.
    """
    edges = None
    orientation = None
    conformation = None
    saenger = None
    
    for field in data:
        if field == 'stacked' or is_edge(field):
            edges = field
        elif is_orientation(field):
            orientation = field
        elif is_roman_numeral(field) or field == 'n/a' or\
            field.startswith('!'):
            saenger = field
        elif field == 'syn':
            if not conformation:
                conformation = field
            else:
                conformation += ' '+field
        else:
            raise RnaViewParseError("Unknown annotation field: %s"%(field))

    return edges, orientation, conformation, saenger

def parse_filename(lines):
    """Return PDB filename from the rnaview output file."""
    assert len(lines) == 1
    return lines[0].split(':')[1].strip()

def parse_uncommon_residues(lines):
    """Returns dictionary of {(chain_id, res_id, res_name): res_assigned}

    Parses the uncommon residue lines from an rnaview file.
    The chain_id could be None, all other field should have some value. If
    any of these is missing, an error will be raised.
    """
    result = {}
    for line in lines:
        res_name = line[17:20].strip()
        res_id = line[20:26].strip()
        chain_id = line[36].strip() or None
        res_assigned = line.split(':')[1].strip()
        if not res_name or not res_id or not res_assigned:
            raise RnaViewParseError(\
            "Found missing field in uncommon residue line.\n"+\
            "res_id: %s, res_name: %s, res_assigned: %s"\
            %(res_id, res_name, res_assigned))
        result[(chain_id, res_id, res_name)] = res_assigned
    return result

def parse_base_pairs(lines, model=None, verify=False):
    """Returns BasePairs object. Parses BASE_PAIR section of rnaview output.

    model: pdb model corresponding to the rnaview output.
    verify: boolean indicating whether the identity of the bases in the
        base pair list should be verified against the bases in the pdb model.
    
    When model=None and verify=True an error will be raised.
    If verify is True and a model is provided, all bases will be checked
        against the pdb model. If the base is a standard residue, it should
        be the same as in the pdb model, if the residue is 'P', the 
        StandardRes in the pdb model should be 'U', in any other case
        the upper case version of the rnaview residue should be equal to
        the StandardRes in the pdb model.
    """
    pairs = []
    for line in lines:
        data = line.split()
        up_chain_id = data[1].strip(': ') or None
        down_chain_id = data[5].strip(': ') or None
        up_res_id = data[2].strip()
        down_res_id = data[4].strip()
        up_res_name, down_res_name = data[3].split('-')
        
        #verify found residues are standard in Rnaview
        try:
            assert up_res_name in RNAVIEW_ACCEPTED and\
                down_res_name in RNAVIEW_ACCEPTED
        except AssertionError:
            raise RnaViewParseError(\
                "Base found that is not generally accepted by Rnaview:"+\
                " %s or %s"%(up_res_name, down_res_name))
        
        #Build upstream and downstream Base objects
        up = Base(up_chain_id, up_res_id, up_res_name)
        down = Base(down_chain_id, down_res_id, down_res_name)
        
        if verify: #verify base identities against corresponding
            if model is None:
                raise RnaViewParseError(\
                    "Can't verify base identities if model is not set")
            else:
                try:
                    for b in [up, down]:
                        res = model[b.ChainId][b.ResId]
                        if b.ResName in STANDARD_RES:
                            assert b.ResName == res.ResName
                        elif b.ResName == 'P':
                            assert res.StandardRes == 'U'
                        else:
                            assert b.ResName.upper() == res.StandardRes
                except AssertionError:
                    raise RnaViewParseError(\
                    "Can't verify base identity of rnaview base"+\
                    " %s and pdb residue %s"%(str(b), str(res)))
        
        e, o, c, s = parse_annotation(data[6:])
        
        pairs.append(BasePair(up, down, Edges=e, Orientation=o,\
            Conformation=c, Saenger=s))
    return BasePairs(pairs, Model=model)



def parse_base_multiplets(lines, model=None, verify=False):
    """Returns BaseMultiplets object. Parses BASE_MULTIPLETS section.

    model: pdb model corresponding to the rnaview output.
    verify: boolean indicating whether the identity of the bases in the
        base pair list should be verified against the bases in the pdb model.

    See parse_base_pairs for comments on base verification.
    """
    residues = []
    for line in lines:
        multi = []
        bases_info = line.strip().split(']',1)[1]
        bases = bases_info.split('+')
        for b in bases:
            data = b.split()
            chain_id = data[0].strip(': ') or None
            res_id = data[1].strip()
            res_name = data[2].strip()
            
            #verify found residue is standard in Rnaview
            try:
                assert res_name in RNAVIEW_ACCEPTED
            except AssertionError:
                raise RnaViewParseError(\
                "Base found that is not generally accepted by RNAVIEW: %s"\
                    %(str(res_name)))
            
            if verify: #verify base identities against corresponding
                if model is None:
                    raise RnaViewParseError(\
                        "Can't verify base identities if model is not set")
                else:
                    try:
                        res = model[chain_id][res_id]
                        if res_name in STANDARD_RES:
                            assert res_name == res.ResName
                        elif res_name == 'P':
                            assert res.StandardRes == 'U'
                        else:
                            assert res_name.upper() == res.StandardRes
                    except AssertionError:
                        raise RnaViewParseError(\
                        "Can't verify base identity of base %s %s %s"\
                        %(chain_id, res_id, res_name))
            
            multi.append(Base(chain_id, res_id, res_name))
        residues.append(BaseMultiplet(multi))
    return BaseMultiplets(residues, Model=model)

def parse_pair_counts(lines):
    """Parses summary of base pairs at the end of the rnaview output file.

    Returns PairCounts object which is a dictionary of {Label: count}
    where label is a string (name from rnaview output), e.g. 'WW--cis',
    and count is an integer representing the base pair count of that
    type of base pair. 
    
    Will raise an error if the number of lines is odd (it expects (multiple)
    label lines followed by the line with the counts.

    If lines is empty it'll return an empty PairCounts object (=empty dict).
    """
    if not len(lines)%2 == 0:
        raise RnaViewParseError("Weird base pair counts format:\n%s"\
            %('\n'.join(lines)))
    res = PairCounts()
    for x in range(0,len(lines),2):
        res.update(zip(lines[x].split(),map(int,lines[x+1].split())))
    return res

def MinimalRnaviewParser(lines):
    """Returns line groups for uncommon res, base pairs, base multiples, counts
    """
    result = {'UC':[],'BP':[],'BM':[],'PC':[], 'FN':[]}
    in_bp = False
    in_bm = False
    in_pc = False
    for line in lines:
        line = line.strip()
        if line:
            if line.startswith('PDB data file name'):
                result['FN'].append(line)
            if line.startswith('uncommon'):
                result['UC'].append(line)
            elif line.startswith('BEGIN_base-pair'):
                in_bp = True
            elif line.startswith('END_base-pair'):
                in_bp = False
            elif line.startswith('BEGIN_multiplets'):
                in_bm = True
            elif line.startswith('END_multiplets'):
                in_bm = False
            elif line.startswith('-------'):
                in_pc = True
            else:
                if in_bp:
                    result['BP'].append(line)
                elif in_bm:
                    result['BM'].append(line)
                elif in_pc:
                    result['PC'].append(line)
                else:
                    continue
    return result

def RnaviewParser(rnaview_lines, pdb_lines=None, verify=False, strict=True,\
    standard_res=STANDARD_RES, specific_changes=None, general_changes=None):
    """Parses output from the Rnaview program.

    rnaview_lines: rnaview output
    pdb_lines: the corresponding pdb lines. If these are not provided 
        the base pairs and base multiples won't be provided with a model.
    verify: boolean indicating whether the rnaview output should be verified
        against the corresponding pdb file.
    strict: error raised or not
    standard_res, specific_changes, general_changes: see PdbParser doc.
    
    Verification means the base identity is checked. In case of a standard
        Rna or Dna residue, it should be the same as in the pdb file. When
        it's an uncommon residue, it should be the same as the StandardRes
        in the pdb model (in case of P (pseudouridine) it should be 'U').

    NMR files can contain multiple models. Rnaview picks either
        the best model or the first model. The algorithm determines what
        model is used in the same way as RnaView.
    """
    if pdb_lines: #get PDB information

        #set standard residues
        pdb_info = PdbParser(pdb_lines, strict=strict, set_standard=True,\
            standard_res=standard_res, specific_changes=specific_changes,\
            general_changes=general_changes)
        
        # determine used model
        models = pdb_info['MODELS']
        if len(models) > 1:
            try:
                model_id = pdb_info['REMARK']['210']['BEST_REPRESENTATIVE']
            except KeyError:
                model_id = '1'
            if model_id is None:
                model_id = '1'
            model = models[model_id]
        else: #take the first and only model
            model = models.values()[0]
    else:
        model = None
    
    #parse RNAVIEW lines
    grouped_lines = MinimalRnaviewParser(rnaview_lines)
    result = {}
    result['FN'] = parse_filename(grouped_lines['FN'])
    result['UC'] = parse_uncommon_residues(grouped_lines['UC'])
    result['BP'] = parse_base_pairs(grouped_lines['BP'], model=model,\
        verify=verify)
    result['BM'] = parse_base_multiplets(grouped_lines['BM'], model=model,\
        verify=verify)
    result['PC'] = parse_pair_counts(grouped_lines['PC'])
    
    return result


if __name__ == "__main__":
    from old_cogent.struct.rna2d import PseudoknotRemover
    from sys import argv
    from controller import is_canonical
    #f = open('new_pdbs/1MME.pdb.out')
    #g = open('new_pdbs/1MME.pdb')
    f = open(argv[1])
    g = open(argv[2])
    res = RnaviewParser(f,g)
    #for k,v in res.items():
    #    print '==========', k, '=========='
    #    print v
    p = res['BP'].select(is_canonical)
    print len(p)
    for x in p.cliques():
        print 'CLIQUE', len(x)
        print x
        print x.PresentChains
        seq, pairs = x.seqAndVienna()
        print seq
        print pairs
        #print pairs.toVienna(len(seq))
    
    """
    def sel(bp):
        if (bp.Edges == '+/+' or bp.Edges == '-/-') and\
            bp.Conformation is None:
            return True
        elif bp.Saenger == 'XXVIII':
            return True
        #if bp.Edges != 'stacked' and bp.isWC() and\
        #    not bp.Saenger.startswith('!'):
        #    return True
        else:
            return False
            
    selection = res['BP'].select(sel)
    selection = res['BP']
    #print selection
    seq, pairs = selection.toPairs(chain_order='ABCD')
    print pairs
    if not pairs.hasPseudoknots():
        print seq
        pairs = pairs.directed()
        print pairs.toVienna(length=len(seq))
    else:
        print 'Pairs has pseudoknots'
        p = PseudoknotRemover()
        new = p(pairs, rule='first', return_removed=False)
        print seq
        print new.toVienna(length=len(seq))
    #print pairs.toVienna(length=len(seq))
    """
    f.close()
    g.close()
