#!/usr/bin/env python
#pdb_model.py
"""
Provides objects to store PDB information

Revision History:
Apr 2006: Written by Sandra Smit
9/18/06 Sandra Smit: cleaned up code. Added tests. Commented out all untested
    code.
12/13/06 Sandra Smit: if there's no chain_id, it's kept as ' ' rather than None
"""
from cogent.util.array import euclidean_distance

class PdbError(Exception):
    pass

STANDARD_RES = {'RNA': {'A':'A', 'G':'G', 'U':'U', 'C':'C'},
    'DNA': {'A':'A','G':'G','T':'T', 'C':'C','U':'U'},
    'AA': {'ALA':'A','CYS':'C','ASP':'D','GLU':'E', 'PHE':'F', 'GLY':'G',
    'HIS':'H','ILE':'I','LYS':'K','LEU':'L','MET':'M','ASN':'N','PRO':'P',
    'GLN':'Q','ARG':'R','SER':'S','THR':'T','SEC':'U','VAL':'V','TRP':'W',
    'TYR':'Y', 'UNK':'X'}}
TRANS_RES = {'RNA': {'T':'U', 'I':'A'}, 'DNA': {'I':'A'}}

#STANDARD_RES = {'ALA':'A','CYS':'C','ASP':'D','GLU':'E', 'PHE':'F', 'GLY':'G',
#    'HIS':'H','ILE':'I','LYS':'K','LEU':'L','MET':'M','ASN':'N','PRO':'P',
#    'GLN':'Q','ARG':'R','SER':'S','THR':'T','SEC':'U','VAL':'V','TRP':'W',
#    'TYR':'Y', 'A':'A','G':'G','U':'U','T':'T', 'C':'C'}

class Models(dict):
    
    #_standard_res = STANDARD_RES
    #_trans_res = TRANS_RES
    
    def __init__(self, *args, **kwargs):
        dict.__init__(self, *args, **kwargs)
        #self.StandardRes = self._standard_res
        #self.Trans = self._trans_res
        
    def __str__(self):
        result = []
        for k, v in self.items():
            result.append(str(v))
        return '\n'.join(result)

    def setStandardResidues(self, standard_res=STANDARD_RES,\
        specific_changes=None, general_changes=None):
        """Sets StandardResidue for all residues in all models.

        First it sets the ResidueType to 'RNA', 'DNA', or 'AA'.
        If the residue is in the list of standard residues for that type,
        the standard residue is the same as the original residue.
        If the ResName is not in the list of standard residues, it tries to
        translate the residue based on the trans_res dict. 
        If that doesn't work either, it tries to figure out the residue
        identity by examining the atoms of the residue.
        If that doesn't work, the StandardRes is None.
        """
        for model in self.values():
            for c in model.iterChains():
                c.setStandardResidues(standard_res, specific_changes,\
                general_changes)

    
    def setDbResidues(self, seqadv_dict):
        """Sets _db_res of all residues in seqadv_dict.

        seqadv_dict: dictionary of {(chain_id, res_id, res_name): db_res_name}

        The SEQADV record identifies conflicts between sequence information
        in the ATOM records of the PDB entry and the sequence database
        entry given on DBREF.
        """
        for model in self.values():
            for res_tuple in seqadv_dict:
                chain_id, res_id, res_name = res_tuple
                if not res_id or not res_name:
                    continue
                res = model[chain_id][res_id]
                if res.ResName != res_name:
                    continue
                db_res = seqadv_dict[res_tuple]
                if db_res != res.ResName and db_res != res.StandardRes:
                    res._db_res = db_res
    


class Model(dict):

    def __init__(self, ModelId, ChainOrder=None, chains=None):
        """Initialize new Model object.
        
        ModelId: Identifier of Model, string
        ...
        """
        self.ModelId = ModelId
        chains = chains or {}
        if isinstance(chains, list) and isinstance(chains[0], Chain):
            self.update(dict([(c.ChainId, c) for c in chains]))
        else:
            self.update(dict(chains))
        if ChainOrder is None:
            self.setChainOrder()
        else:
            self.ChainOrder = ChainOrder
        
    def __str__(self):
        """Return string representation of Model object"""
        result = []
        result.append('Model: ' + str(self.ModelId))
        for k, v in self.items():
            result.append(str(v))
        return '\n'.join(result)

    def setChainOrder(self):
        """Set ChainOrder of Model to the list of keys in self"""
        self.ChainOrder = self.keys()
    
    def iterChains(self):
        """Yield all chains in the model"""
        for chain in self.values():
            yield chain

    def iterResidues(self):
        """Yield all residues in the Sequences of all Chains in the Model"""
        for chain in self.iterChains():
            for res in chain.iterResidues():
                yield res
   
    def seq_info(self, chains=None):
        if chains is None:
            chains = self.ChainOrder
        else:
            for c in chains:
                if c not in self:
                    raise PdbError(\
                        "Chain %s doens't exist in Model"%(c))
        
        result = []
        for chain in chains:
            result.extend(self[chain].seq_info())
        return result


    def seq(self, chains=None, attr_name='ResName', one_letter=True,\
        standard_res=STANDARD_RES, accepted_res_types=['DNA','RNA','AA',None]):
        """Returns concatenated sequences of requested chains.

        chains: list of chain IDs, e.g. ['A','B',None]
        """
        if chains is None:
            chains = self.ChainOrder
        else:
            for c in chains:
                if c not in self:
                    raise PdbError(\
                        "Chain %s doens't exist in Model"%(c))

        if one_letter:
            return ''.join([self[chain].seq(attr_name,\
                one_letter, standard_res, accepted_res_types)\
                for chain in chains])
        else:
            res = []
            for chain in chains:
                res.extend(self[chain].seq(attr_name, one_letter,\
                    standard_res, accepted_res_types))
            return res
    

class Chain(dict):
    """Hold a Chain object.

    A Chain is a dictionary of residues {residue ID: Residue object}
    """

    def __init__(self, ModelId, ChainId, residues=None):
        """Initializes Chain object.
        
        ModelId: Identifier of the model, string
        ChainId: Identifier of the chain, string (even if '0')
        residues: list of Residue objects or dict of {ResId: Residue}

        Chain is a dict of {ResId: Residue}
        Sequence will be constructed as a list of [Residue] in order 
            of SeqPos, residues without SeqPos will be excluded from this list.
        """
        self.ModelId = ModelId
        self.ChainId = ChainId
        residues = residues or {}
        if isinstance(residues, list) and isinstance(residues[0], Residue):
            self.update(dict([(res.ResId, res) for res in residues]))
        else:
            self.update(dict(residues))
        self.updateSequence()

    def __str__(self):
        """Return string representation of Chain object"""
        result = []
        result.append('\t' + ' '.join(map(str,[self.ModelId, self.ChainId])))
        for res_id, res in self.items():
            result.append('\t\t'+str(res))
        return '\n'.join(result)
 
    def updateSequence(self):
        """Put all Residues in a list ordered by SeqPos (set self.Sequence)
        
        Only includes Residues whose SeqPos is not None.
        """
        seq_res = [r for r in self.values() if r.SeqPos is not None]
        if not seq_res:
            self.Sequence = []
        else:
            max_index = max([r.SeqPos for r in seq_res])
            self.Sequence = [None]*(max_index+1)
            for res in seq_res:
                self.Sequence[res.SeqPos] = res
                
    def addResidue(self, residue, update=True):
        """Add a Residue after the object is already initialized
        
        Update self (the dictionary) and self.Sequence.
        """
        self[residue.ResId] = residue
        if update:
            self.updateSequence()
       
    def seq_info(self):
        result = []
        for r in self.Sequence:
            curr = []
            curr.append(r.ChainId) 
            curr.append(r.ResId) 
            curr.append(r.ResName) 
            curr.append(r.StandardRes)
            curr.append(r.ResType)
            curr.append(r.SeqPos)
            result.append(curr)
        return result
    
    def seq(self, attr_name='ResName', one_letter=True,\
        standard_res=STANDARD_RES,\
        accepted_res_types=['DNA','RNA','AA',None]):
        """Returns list or string representation of self.Sequence.

        attr_name: name of the Residue attribute that should be returned,
            could be 'ResName','StandardRes','DbRes'
        one_letter=True: indicates whether three letter residue names
            will be translated in one letter names
        """
        if one_letter:
            try:
                result = []
                for r in self.Sequence:
                    #print r.ResName, r.StandardRes, r.ResType
                    if r.ResType not in accepted_res_types:
                        #print "res not accepted", r.ResName, r.ResType
                        continue
                    if r.ResType is None:
                        raise PdbError(\
                        "Residue type unknown for residue (%s, %s, %s, %s)"\
                        %(r.ModelId, r.ChainId, r.ResId, r.ResName))
                    residue = getattr(r,attr_name)
                    if residue is None:
                        raise PdbError("No one-letter code for None")
                    
                    result.append(standard_res[r.ResType][residue])
                return ''.join(result)
            except KeyError, e:
                raise PdbError, "No one-letter symbol for %s"%(str(e))
        else:
            return [getattr(r,attr_name) for r in self.Sequence]

    def iterResidues(self):
        """Iterates over all Residue objects in self.Sequence"""
        for res in self.Sequence:
            yield res

    def iterAllResidues(self):
        """Iterates over all residues in the chain (independent of Sequence)
        """
        for res in self.values():
            yield res

    def setStandardResidues(self, standard_res=STANDARD_RES,\
        specific_changes=None, general_changes=None):
        """Determines and sets standard residue for each res in the chain.

        standard_res: 2D dictionary where there's a dictionary of
            accepted residues for each Residue Type. For example:
            {'DNA':{'A':'A','T':'T'},'RNA':{'U':'U'},
            'AA':{'ALA':'A','CYS':'C'}}
        specific_changes: dictionary of required changes to specific residues,
            for example: a dictionary of modified residues. Keys should be
            (chain_id, res_id, res_name). Values should be the standard
            residue.
        general_changes: 2D dictionary, same format as standard_res. Can
            be used to make user-specified translations.

        The standard residue is determined as follows:
            First the Residue Type is set: 'RNA', 'DNA' or 'AA'
            If the residue type is not None and the residue name is
                in the list of accepted residues for that type, the standard
                residue is set to that.
            If the combination of chain ID, residue ID, and residue Name is
                in the dictionary of specific changes, the standard residue
                is set to that.
            If the residue type is not None and their is a user-specified
                change for this residue of this residue type, the standard
                residue is set to that.
            If the residue type is 'AA', there's nothing else we can do. If
                the type is 'DNA' or 'RNA', we identify the residue by its
                atoms and set the standard residue to that.
        """
        #print 'standard', standard_res
        for res in self.iterAllResidues():
            if res.ResType is None:
                res.setResType()
            if res.ResType and res.ResType in standard_res and\
                res.ResName in standard_res[res.ResType]:
                res._standard_res = res.ResName
            elif specific_changes and (res.ChainId, res.ResId, res.ResName)\
                in specific_changes:
                res._standard_res =\
                    specific_changes[(res.ChainId, res.ResId, res.ResName)]
                #if res._standard_res in general_changes[res.ResType]:
                #    res._standard_res =\
                #        general_changes[res.ResType][res._standard_res]
            elif general_changes and res.ResType and\
                res.ResType in general_changes and res.ResName in\
                general_changes[res.ResType]:
                res._standard_res = general_changes[res.ResType][res.ResName]
            else:
                if res.ResType == 'DNA' or res.ResType == 'RNA':
                    res._standard_res = res.identifyUnknown()
                else:
                    continue
    

class Residue(object):
    """Residue is a list of Atom objects"""

    def __init__(self, ModelId, ChainId, ResId, ResName,\
        ResSeq=None, Icode=None, SeqPos=None, Atoms=None):
        """Initialize a new Residue object.

        Required arguments:
        ModelId: string describing the model or None
        ChainId: string, chain identifier 
        ResId:
        ResName:
        Together these parameters uniquely identify the Residue.

        Optional arguments:


        """
        self.ModelId = ModelId
        self.ChainId = ChainId
        self.ResId = ResId #ResSeq+Icode
        self.ResName = ResName #Original residue name
        self._standard_res = None
        self._db_res = None
        self.ResSeq = ResSeq
        self.Icode = Icode
        self.SeqPos = SeqPos
        if Atoms is None:
            self.Atoms = []
        else:
            self.Atoms = Atoms
        self._res_type = None
        self._base_type = None

    def setResType(self):
        """Set ResType and possibly BaseType (if nucl) of the Residue
        
        If the Residue has no atoms, ResType stays None.
        Amino Acids are recognized by the presence of CA in the list.
        RNA and DNA are distinguished by the O2*.
        Purines and pyrimidines are distinguished by particular distances
            between atoms.

        This methods is the same as in the RnaView package.
        """
        self._res_type = None
        self._base_type = None
        if len(self.Atoms) == 1:
            self._res_type = 'SA' # single atom
        if self.Atoms: #if not Atoms, res_type stays None
            d_small = 2.0
            d_large = 3.0
            names = {}
            for a in self.Atoms:
                if a.AtomName not in names:
                    names[a.AtomName] = a
            if 'N1'in names and 'C2' in names and 'C6' in names: #base
                d1 = names['N1'].distance(names['C2'])
                d2 = names['N1'].distance(names['C6'])
                d3 = names['C2'].distance(names['C6'])
                if d1 < d_small and d2 < d_small and d3 < d_large:
                    if 'N9' in names and\
                        3.5 <= names['N9'].distance(names['N1']) <= 4.5:
                        self._base_type = 'R' #purine
                    else:
                        self._base_type = 'Y' #pyrimidine
                    if 'O2*' in names or "O2'" in names:
                        self._res_type = 'RNA'
                    else:
                        self._res_type = 'DNA'
            #elif: #AA
            elif 'CA' in names and ('C' in names or 'N' in names): #AA
                try:
                    d = names['CA'].distance(names['C'])
                except KeyError:
                    d = names['CA'].distance(names['N'])
                if d < d_small:
                    self._res_type = 'AA'
            else: # assume missing atoms in RNA
                #print 'here', self
                self._res_type = 'RNA'
                self._base_type = self.ResSeq
                
    
    def _get_res_type(self):
        """Return ResType of the molecule"""
        return self._res_type

    ResType = property(_get_res_type)
    
    def _get_base_type(self):
        """Return BaseType of the molecule"""
        return self._base_type

    BaseType = property(_get_base_type)
            
    def identifyUnknown(self):
        """Return the identity of the nucleotide, based on the list of Atoms
        
        Only works for NUCLEOTIDES. If ResType is AA, None is returned.
        If base can't be identified, None is returned.
        """
        
        if self.ResType == 'DNA' or self.ResType == 'RNA':
            names = dict.fromkeys([a.AtomName for a in self.Atoms])
            if self.BaseType == 'R':
                if 'N2' in names:
                    return 'G'
                else:
                    return 'A'
            else:
                if self.ResType == 'DNA' and 'C5M' in names:
                    return 'T'
                elif 'N4' in names:
                    return 'C'
                else:
                    return 'U'
        return None


    def __str__(self):
        """Return string representation of Residue"""
        result = map(str,[self.ModelId, self.ChainId, self.ResId,\
            self. ResName, self.SeqPos])
        result = [' '.join(result)]
        for a in self.Atoms:
            result.append('\t\t\t'+str(a))
        return '\n'.join(result)

    def _get_standard_res(self):
        """Return the standard residue"""
        return self._standard_res
        
    StandardRes = property(_get_standard_res)

    def _get_db_res(self):
        """Return Residue as it occurs in the database entry

        If the database residue is not set, it returns the standard
        residue. If that one isn't set either, it means that the original
        residue is a standard residue or the standard residue just isn't set, 
        it returns the original residue name.

        Assumption: always standard residue in the database.
        """
        if self._db_res is not None:
            return self._db_res
        elif self.StandardRes is not None:
            return self.StandardRes
        else:
            return self.ResName
        
    DbRes = property(_get_db_res)


class Atom(object):
    """Hold information on an ATOM or HETATM from a PDB file"""
    
    def __init__(self, ModelId, ChainId, ResId, AtomId, AtomName,\
        AltLoc=None, ResName=None, ResSeq=None, Icode=None, Coordinates=None,\
        Occupancy=None, TempFactor=None, SegId=None, Element=None,\
        Charge=None, ParamStdev=None, AnisoTemp=None, AnisoTempStdev=None):
        """Initialize a new Atom object
        
        ModelId: string of Model name, e.g. '1'
        ChainId: string containing ID of chain, e.g. 'A' or '0' or None
        ...
        """
        self.ModelId = ModelId
        self.ChainId = ChainId
        self.ResId = ResId
        self.AtomId = AtomId
        self.AtomName = AtomName
        self.AltLoc = AltLoc
        self.ResName = ResName
        self.ResSeq = ResSeq
        self.Icode = Icode
        self.Coordinates = Coordinates
        self.Occupancy = Occupancy
        self.TempFactor = TempFactor
        self.SegId = SegId
        self.Element = Element
        self.Charge = Charge
        self.ParamStdev = ParamStdev
        self.AnisoTemp = AnisoTemp
        self.AnisoTempStdev = AnisoTempStdev

    def __str__(self):
        """Return essential Atom information"""
        return ' '.join(map(str,[self.ModelId, self.ChainId, self.ResId,\
            self.AtomId, self.AtomName]))

    def distance(self, other):
        """Calculate the distance between two Atoms"""
        return euclidean_distance(self.Coordinates, other.Coordinates)


class DbRef(object):
    """Hold database reference information

    The DBREF record provides cross-reference links between PDB sequences
    and the corresponding database entry or entries.
    """
    
    def __init__(self, PdbId, ChainId=None, PdbFrom=None, PdbTo=None,\
        DbName=None, DbAccession=None, DbIdCode=None, DbFrom=None, DbTo=None):
        """Initialize a new DbRef object"""
        self.PdbId = PdbId
        self.ChainId = ChainId or None
        self.PdbFrom = PdbFrom
        self.PdbTo = PdbTo
        self.DbName = DbName
        self.DbAccession = DbAccession
        self.DbIdCode = DbIdCode
        self.DbFrom = DbFrom
        self.DbTo = DbTo

    def __str__(self):
        """Return string representation"""
        return ' '.join(map(str,[self.PdbId, self.ChainId]))
