#!/usr/bin/env python
"""Provides an algorithm that parses sequences (and their helix numbering)
from the european rRNA database in DCSE alignment format.

The main function "process_DCSE_aln" yields StructureAnnotation objects,
which contain all the necessary information about the sequence and its
structure. The most important properties are GappedSeq and Pairs. The list
of base pairs corresponds to this GappedSeq. 

Further processing, such as degapping the sequences and structure, or
converting them to other objects should be done in downstream analyses.

This module has been tested on the whole rRNA database, downloaded
in August 2006 (plastid and mitochondria files downloaded in January 2007).
It can handle all sequences without BasicStructureErrors, except for 2:
Chl. Picea mariana L07059 (a single chloroplast in the SSU plastid file)
Kin. Herpetomonas muscarum U01015 (a single kinetoplast in the SSU
    mitochondria file)
"""
from __future__ import division
from sys import argv
from string import strip

# =============================================================================
# CLASSES
# =============================================================================

class BasicStructureError(Exception):
    pass

class HelixError(Exception):
    pass

class FinalCountError(Exception):
    pass

class AnnotationError(Exception):
    pass

class DcseParseError(Exception):
    pass

class StructureAnnotation(object):

    def __init__(self, SeqId, RdbSeq, HelixNum):
        """Initialize StructureAnnotation object
        
        SeqId -- string, sequence ID, e.g. Pla. Palmaria palmata Z18289
        RdbSeq -- string, the original sequence as extracted from
            the DCSE format.
        HelixNum -- string, the original helix numbering as extracted
            from the DCSE format.

        This methods initialized a few variables, performs basic checks, 
        and pairs up all the helices.
        """
        self.SeqId = SeqId
        self.RdbSeq = RdbSeq
        self.HelixNum = HelixNum
        # Check length sequence and annotation 
        if len(self.RdbSeq) != len(self.HelixNum):
            raise AnnotationError(\
                "Length sequence different than length helix numbering")
        self.basicSeqCheck() # check bracket counts
        # collect pairing information
        id_to_lmr, id_to_content, pos_to_base, unpaired =\
            extract_struct_info(self.RdbSeq, self.HelixNum)
        self.IdToLmr = id_to_lmr
        self.IdToContent = id_to_content
        self.PosToBase = pos_to_base
        self.Unpaired = unpaired
        self.basicHelixCheck() # check pairing potential helices
        
        self.GappedSeq = self.reconstructSequence()
        orig_len_paired = sum(map(len,self.IdToContent.values()))
        self.IdToPoss = self.getIdToPoss()

        # pair all helices together
        self.Unprocessed = self.IdToLmr.keys()
        self.Processed = []
        self.Pairs = []
        self.UsedAnnotations = {}
        self.pairUp()

        # check final counts
        if self.Unprocessed:
            raise FinalCountError("Unprocessed helices left")
        if not self.Unprocessed:
            if len(self.Processed) == len(self.IdToLmr):
                if orig_len_paired != len(self.Pairs)*2:
                    raise FinalCountError(\
                    "Resulting number of pairs doesn't match original info")
            else:
                raise FinalCountError("RESULT: LENGTHS DON'T MATCH")
        if len(self.IdToContent) != len(self.UsedAnnotations):
            raise FinalCountError(\
                "Length UsedAnnotations should match length IdToContent")
        

    def basicSeqCheck(self):
        """Raise error if basic counts are bad.

        The basic counts are bad if:
        1. the number of opening and closing brackets of a particular
        type differs.
        2. the number of helices (calculated from the brackets) is odd.
        """
        counts = {}
        for sym in "[](){}^":
            counts[sym] = self.RdbSeq.count(sym)
        if counts['['] != counts[']']:
            raise BasicStructureError(\
                "Stem brackes don't match: %s opening, %s closing"\
                %(counts['['], counts[']']))
        if counts['{'] != counts['}']:
            raise BasicStructureError(\
                "Bulge brackes don't match: %s opening, %s closing"\
                %(counts['{'], counts['}']))
        if counts['('] != counts[')']:
            raise BasicStructureError(\
            "Uncommon base pair brackes don't match: %s opening, %s closing"\
                %(counts['('], counts[')']))
        if (counts['['] + counts['^'])%2 == 1:
            raise BasicStructureError("Odd number of helixes (should be even)") 
        
    def basicHelixCheck(self):
        """Raise error if something is wrong with the found helices.
        
        Some things that could be wrong:
        1. IdToContent and IdToLmr could have different keys
        2. IdToContent could contain empty helices
        3. There could be an odd number of helices of a particular length
        (if all helices would have a partner, the number of helices for each
        length should be even)
        """
        lmr_keys = self.IdToLmr.keys()
        content_keys = self.IdToContent.keys()
        lmr_keys.sort()
        content_keys.sort()
        # check keys in both dictionaries
        if lmr_keys != content_keys:
            raise BasicStructureError("Dictionaries contain different keys")
        # check for empty helices
        for k, v in self.IdToContent.items():
            if not v:
                raise BasicStructureError("Empty helix found (LMR info %s)"\
                    %(str(self.IdToLmr[k])))
        # check matching potential
        length_to_count, length_to_id = self.getLengthCounts()
        for k, v in length_to_count.items():
            if v%2 == 1:
                raise BasicStructureError(\
                    "Odd number of helices of length %s found"%(k))

    def getIdToPoss(self):
        """Return {helix_id: [helix_annotations]} from IdToLmr
        
        Basically the three different lists in IdToLmr (left, middle, right)
        are merged into a single list. 
        """
        id_to_poss = {}
        for k, (l,m,r) in self.IdToLmr.items():
            id_to_poss[k] = []
            for i in l+m+r:
                id_to_poss[k].append(i)
        return id_to_poss

    def getLengthCounts(self):
        """Return {length: count} and {length:[helix_ids]} for all helices.
        
        Returns the results for all helices in self.IdToContent
        """
        length_to_count = {}
        length_to_id = {}
        for k, v in self.IdToContent.items():
            l = len(v)
            if l not in length_to_count:
                length_to_count[l] = 0
                length_to_id[l] = []
            length_to_count[l] += 1
            length_to_id[l].append(k)
        return length_to_count, length_to_id
    
    def groupByBasicAnn(self, ids):
        """Return list of sets of helix_ids with the same basic annotation

        For example, given:
        {3: ["1","2"], 4: ["2","2'"], 5: ["3"]}
        The intermediate is:
        {"1":[3], "2":[3,4], "3": [5]}
        The return value is (after merging the sets):
        [[3,4],[5]]
        """
        by_upstream = {}
        for id in ids:
            options = self.IdToPoss[id]
            if not options:
                if None not in by_upstream:
                    by_upstream[None] = set([])
                by_upstream[None].update([id])
            for o in options:
                if basic_ann(o) not in by_upstream:
                    by_upstream[basic_ann(o)] = set([])
                by_upstream[basic_ann(o)].update([id])
        s = set_merger(by_upstream.values())
        return s

    def unprocessedByLength(self):
        """Return unprocessed helix IDs grouped by length"""
        length_to_id = {}
        for i in self.Unprocessed:
            length = len(self.IdToContent[i])
            if length not in length_to_id:
                length_to_id[length] = []
            length_to_id[length].append(i)
        return length_to_id        

    def unprocessedByLengthAnn(self):
        """Return unprocessed helix_ids grouped by length and annotation"""
        length_to_count = {}
        for i in self.Unprocessed:
            length = len(self.IdToContent[i])
            if length not in length_to_count:
                length_to_count[length] = []
            length_to_count[length].append(i)
        result = []
        for k, v in length_to_count.items():
            result.extend(self.groupByBasicAnn(v))
        if len(self.Unprocessed) != sum(map(len, result)):
            raise HelixError("Forming cliques: lost IDs on the way")
        unp = self.Unprocessed
        unp.sort()
        cl = []
        for i in result:
            cl.extend(i)
        cl.sort()
        if unp != cl:
            raise HelixError("Forming cliques: lost IDs on the way")
        return result

    def processMatches(self, matches):
        """Match two helices and update all information in the object

        matches -- dictionary of:
            {upstream_helix_id: (downstream_helix_id, basic annotation)}
            For example, {(1:(4, "1"), 15:(27, "E23_2"}

        This method takes care of internal consistency in the object. It
        moved IDs from unprocessed to processed, it records the used
        annotations, and it renews the possibilities for all left-over
        helices.
        """
        for up_helix, (down_helix, helix_ann) in matches.items():
            if helix_ann in self.Processed or\
                opposite_helix(helix_ann) in self.Processed:
                raise AnnotationError("Helix already in results")
            up = self.IdToContent[up_helix]
            down = self.IdToContent[down_helix]
            self.Pairs.extend(helix_to_pairs(up, down))
            self.Unprocessed.remove(up_helix)
            self.Processed.append(up_helix)
            self.Unprocessed.remove(down_helix)
            self.Processed.append(down_helix)
            self.UsedAnnotations[helix_ann] = None
            self.UsedAnnotations[opposite_helix(helix_ann)] = None
            new_id_to_poss = {}
            for k, v in self.IdToPoss.items():
                new_poss = []
                for i in v:
                    if i not in self.UsedAnnotations:
                        new_poss.append(i)
                new_id_to_poss[k] = new_poss
            self.IdToPoss = new_id_to_poss

    def findIndices(self, helix_annotations):
        """Return (index_in_ann, helix_ann) for helix_ann in helix_annotations.

        helix_annotations -- list of helix annotations, such as
            "1", "2'", "E23_7'"

        It searches for every annotation in the HelixNumbering. Searches
        are done with specific prefixes and suffixes to avoid bad matches.

        Results are sorted!
        """
        opt_idx = []
        for h in helix_annotations:
            idx = None
            for pre in ["-", " "]:
                if idx:
                    break
                for suf in ["-"," "]:
                    if idx:
                        break
                    try:
                        idx = self.HelixNum.index(pre+h+suf)
                        break
                    except ValueError:
                        continue
            if idx is None:
                raise HelixError("Can't find helix in annotation")
            else:
                opt_idx.append((idx, h))
        opt_idx.sort()
        return opt_idx

    def scratchImpossible(self, ids):
        """Removes helix annotations that not have a match in the groups of IDs
        
        Removes annotations from which the partner annotation (pairing helix) 
        is not in any of the helix possibilities of the other helices.
        """
        for i in ids:
            all_other_poss = []
            for j in ids:
                if j == i:
                    continue
                all_other_poss.extend(self.IdToPoss[j])
            new_poss = []
            for opt in self.IdToPoss[i]:
                if opposite_helix(opt) in all_other_poss:
                    new_poss.append(opt)
            self.IdToPoss[i] = new_poss
    
    def pairUp(self):
        """Pair all helices in the object"""
        
        # process the majority --> unique annotation in middel of helix
        res = self.matchSingleMiddle()
        self.processMatches(res)
        
        # process all groups of helices (size 2) that are grouped based
        # on length and annotation, and that have support by at least on
        # helix annotation in the middle of the helix
        # do two round to process the new options after the first round
        for x in range(2):
            if self.Unprocessed:
                cliques = self.unprocessedByLengthAnn()
                for ids in cliques:
                    if len(ids) == 2:
                        t = tuple(ids)
                        if (self.IdToLmr[t[0]][1] or self.IdToLmr[t[1]][1]):
                            self.processMatches(self.matchUp(ids))

        for x in range(2):
            if self.Unprocessed:
                # match set of two helices, any support (not only from middle)
                cliques = self.unprocessedByLengthAnn()
                for ids in cliques:
                    if len(ids) == 2:
                        self.processMatches(self.matchUp(ids))
           
            if self.Unprocessed:
                # match len 4 and 6 by single options for annotation
                cliques = self.unprocessedByLengthAnn()
                for ids in cliques:
                    if len(ids) == 4 or len(ids) == 6 or len(ids) == 8:
                        self.scratchImpossible(ids)
                        matches = self.pairSingleOption(ids)
                        self.processMatches(matches)

            if self.Unprocessed:
                # match len 4 and 6 by annotation index 
                cliques = self.unprocessedByLengthAnn()
                for ids in cliques:
                    if len(ids) == 4 or len(ids) == 6 or len(ids) == 8:
                        matches = self.pairByAnnIndex(ids, force_ann=True)
                        self.processMatches(matches)

        if self.Unprocessed:
            # match cliques of any length by single option
            cliques = self.unprocessedByLengthAnn()
            for ids in cliques:
                if len(ids) > 1: # can't match a single helix
                    matches = self.pairSingleOption(ids)
                    self.processMatches(matches)

        if self.Unprocessed:
            # make groups of IDs by length (not annotation)
            # match len 2, multiple viable options, or no options
            length_to_id = self.unprocessedByLength()
            groups = length_to_id.values()
            for ids in groups:
                matches = self.pairMultiOrNone(ids)
                self.processMatches(matches)

        if self.Unprocessed:
            # group IDs just by length (not annotation)
            # try to match up by annotation index in a last attempt
            length_to_id = self.unprocessedByLength()
            groups = length_to_id.values()
            for ids in groups:
                matches = self.pairByAnnIndex(ids, force_ann=False)
                self.processMatches(matches)
        
    def matchSingleMiddle(self):
        """Match helices that have a single annotation in their middle

        1. Look at the annotation in the middle of a helix and to its left
        and right (LMR). Select all helices that have a single annotation
        in their middle.
        2. Select the helices that are in the middle of just 1 helix and
        that have one downstream partner, which is in the middle of a
        single other helix.

        The return value is a dict of:
        {upstream_helix_id: (downstream_helix_id, helix_ann)}
        E.g. {3:(7, "4"), 25:(54, "48")}

        This method matches up about 70 to 80% of all helices.
        """
        # find all the helices that have a single middle key, and that key
        # is not the middle key in another helix
        unique_middle_keys = {}
        for k, v in self.IdToLmr.items():
            if len(v[1]) == 1: # length middle annotation is 1
                helix_id = basic_ann(v[1][0])
                if helix_id not in unique_middle_keys:
                    unique_middle_keys[helix_id] = {'U':[], 'D':[]}
                if not v[1][0].endswith("'"): #upstream
                    unique_middle_keys[helix_id]['U'].append(k)
                else: # downstream
                    unique_middle_keys[helix_id]['D'].append(k)
        result = {}
        for k, v in unique_middle_keys.items():
            if len(v['U']) == 1 and len(v['D']) == 1:
                #helix length should match
                if len(self.IdToContent[v['U'][0]]) ==\
                    len(self.IdToContent[v['D'][0]]):
                    ann = basic_ann(self.IdToLmr[v['U'][0]][1][0])
                    result[v['U'][0]] = (v['D'][0], ann) 
        return result

    def matchUp(self, ids):
        """Match up TWO helices based on the annotation.
        
        ids -- list of length 2, containing helix IDs, e.g. [17,23]

        If the two helices have no possibile helix annotation (i.e. there is
        no annotation in the middle of the helix and the options left and 
        right are already processed, the annotation will be
        'XXX_helixID_helixID'.
        If it can't find a suitable match, the return value is {}. A good 
        match is found if there is a single viable option to match an upstream
        and downstream annotation.
        """
        matches = {}
        if len(ids) == 2:
            up_helix, down_helix = min(ids), max(ids)
            up_poss, down_poss = self.IdToPoss[up_helix],\
                self.IdToPoss[down_helix]
            if not up_poss and not down_poss:
                ann = 'XXX_'+'_'.join(map(str,[up_helix, down_helix]))
            else:
                ann = matching_ann(up_poss, down_poss, flexible=False)
            if ann is not None:
                matches[up_helix] = (down_helix, ann)
        return matches

    def pairSingleOption(self, ids):
        """Match single option for annotation -- unique matching helix

        ids -- list of helix IDs, e.g. [13,17,18,20] 
        """
        matches = {}
        counter = {}
        # collect the helices with a single option for annotation
        double = set([])
        for i in ids:
            if len(self.IdToPoss[i]) == 1:
                if self.IdToPoss[i][0] in matches:
                    double.update([self.IdToPoss[i][0]])
                else:
                    matches[self.IdToPoss[i][0]] = i
        # remove doubles from matches
        for i in double:
            del matches[i]
        # find single alternatives for the helices in matches
        for i in ids:
            temp = []
            for opt in self.IdToPoss[i]:
                if opposite_helix(opt) in matches:
                    temp.append((opt, i))
            if len(temp) == 1:
                res_opt, res_helix = temp[0]
                if res_opt not in counter:
                    counter[res_opt] = []
                counter[res_opt].append(i)
        # collect the good matches
        result = {}
        for helix_ann, helix_id in matches.items():
            if opposite_helix(helix_ann) in counter and\
                len(counter[opposite_helix(helix_ann)]) == 1:
                counter_helix = counter[opposite_helix(helix_ann)][0]
                up_helix, down_helix = min(helix_id, counter_helix),\
                    max(helix_id, counter_helix)
                result[up_helix] = (down_helix, basic_ann(helix_ann))
        return result

    def pairByAnnIndex(self, ids, force_ann=True):
        """Match helices by annotation index

        ids -- list of helix IDs, e.g. [34, 45, 87]
        force_ann -- boolean, indicating whether the assigned annotation
        to a helix_id should be found in the original options for that helix
        """
        ids = list(ids)
        ids.sort()
        # collect all annotation possibilities
        helix_keys = []
        for id in ids:
            helix_keys.extend(self.IdToPoss[id])
        helix_keys = set(helix_keys) # make a set to remove duplicates
        # keep only the viable options
        viable = []
        for i in helix_keys:
            if opposite_helix(i) in helix_keys:
                viable.append(i)
        viable.sort()
        # only continue if the number of viable options == the number of IDs
        if len(viable) == len(ids):
            # get the indices of the annotation possibilities in HelixNum
            opt_idx = self.findIndices(viable) # results are sorted
            matches = {}
            for helix_key, (idx, helix_ann) in zip(ids, opt_idx):
                if force_ann:
                    if helix_ann not in self.IdToPoss[helix_key]:
                        return {}
                helix_id = basic_ann(helix_ann)
                if helix_id not in matches:
                    matches[helix_id] = {'U': [], 'D':[]}
                if helix_ann.endswith("'"):
                    matches[helix_id]['D'].append(helix_key)
                else:
                    matches[helix_id]['U'].append(helix_key)
            reformatted_matches = {}
            for k, v in matches.items():
                reformatted_matches[v['U'][0]] = (v['D'][0], k)
            return reformatted_matches
        else:
            return {}

    def pairMultiOrNone(self, ids):
        """Match two helices no matter what the annotation is.
        
        If there are no possibilities, ann = 'XXX_helixID_helixID'
        If there is one possibility, that is returned.
        If there are two viable options, the first one is chosen.
        If there are no viable options, ann = 'NNN_helixID_helixID'
        """
        matches = {}
        if len(ids) == 2:
            up_helix, down_helix = min(ids), max(ids)
            up_poss = self.IdToPoss[up_helix]
            down_poss = self.IdToPoss[down_helix]
            if not up_poss and not down_poss:
                ann = 'XXX_'+'_'.join(map(str,[up_helix, down_helix]))
            else:
                ann = matching_ann(up_poss, down_poss, flexible=True)
            if ann is not None:
                matches[up_helix] = (down_helix, ann)
            if ann is None:
                ann = 'NNN_'+'_'.join(map(str,[up_helix, down_helix]))
                matches[up_helix] = (down_helix, ann)
        return matches

    def reconstructSequence(self):
        max_pos = max(self.PosToBase.keys())
        seq = []
        for idx in range(max_pos+1):
            seq.append(self.PosToBase[idx])
        return ''.join(seq)

# =============================================================================
# STRUCTURE INFORMATION
# =============================================================================

def get_complete(helix_part):
    """Returns list of complete helix names found in helix part.

    helix_part -- string that contains the helix numbering

    For example:
    helix_part = "- -G6 - -" -> ["G6"]
    helix_part = "14- - - - - - -G" -> []
    helix_part = "- -G6'- - - -G7 -" -> ["G6'","G7"]
    """
    results = []
    parts = map(strip,helix_part.split('-'))
    for idx, i in enumerate(parts):
        if not i:
            continue
        if idx == 0 or idx == len(parts)-1:
            continue
        elif i and not parts[idx-1] and not parts[idx+1]:
            results.append(i)
        elif i and not parts[idx-1] and parts[idx+1]:
            results.append(i)
        elif i and parts[idx-1] and not parts[idx+1]:
            results.append(i)
    return results

def get_left_middle_right(helix_numbering, helix_idx, rdb_seq):
    """Returns list of helices matching left/right or middle of helix.

    helix_numbering -- string, full helix numbering as extracted
        from the DCSE file
    helix_idx -- int, index of start of helix
    rdb_seq -- string, original sequence as extracted from the file
        in DCSE format

    Return value is a list of 3 lists. The first one contains the 
    helix annotations found to the left of the helix, the second
    contains the helix annotations found in the middle of the helix
    between [ and ], the third contains the annotations to the 
    right of the helix.
    """
    seq_length = len(rdb_seq)
    #find index of closing bracket ] or carrot ^
    try:
        closing_bracket = min(rdb_seq[helix_idx:].index("]"),\
            rdb_seq[helix_idx+1:].index("^"))
    except ValueError: #substring not found, no carrot present
        closing_bracket = rdb_seq[helix_idx:].index("]")
    
    #determine initial start and end
    start = helix_idx
    end = helix_idx + closing_bracket + 1
    #if start and end too close together, widen range
    if helix_numbering[start] != '-' and helix_numbering[start] != ' ':
        start -= 3
    if helix_numbering[start+1] != '-' and helix_numbering[start+1] != ' ':
        start -= 3
    if helix_numbering[end-1] != '-' and helix_numbering[end-1] != ' ':
        end += 3
    if helix_numbering[end-2] != '-' and helix_numbering[end-2] != ' ':
        end += 3
    if end-start <= 6:
        end += 2
        start -= 2

    #get helix part and rdb part    
    helix_part = helix_numbering[start:end]
    rdb_part = rdb_seq[start:end]
    
    #get matching part(s) in the middle
    middle = get_complete(helix_part)
    middle_length = len(middle)
    
    #extend to the right
    new_end = end
    right = get_complete(helix_part)
    if middle_length > 0:
        while len(right) == middle_length and new_end < seq_length and\
            new_end < end + 100:
            right = get_complete(helix_numbering[start:new_end])
            new_end += 3
    else:
        while len(right) == middle_length and new_end < seq_length:
            right = get_complete(helix_numbering[start:new_end])
            new_end += 3
    
    #extend to the left
    new_start = start
    left = get_complete(helix_part)
    if middle_length > 0:
        while len(left) == middle_length and new_start >= 0 and\
            new_start > start -100:
            left = get_complete(helix_numbering[new_start:end])
            new_start -= 3
    else:
        while len(left) == middle_length and new_start >= 0:
            left = get_complete(helix_numbering[new_start:end])
            new_start -= 3
    
    #remove duplicates
    for i in middle:
        try:
            right.remove(i)
            left.remove(i)
        except ValueError:
            continue
    
    return left, middle, right

def extract_struct_info(seq, num):
    """Return helix IDs, annotations, content, and positions

    seq -- string, the original sequence as extracted from the DCSE file
    num -- string, the orignial helix numbering as extracted from the DCSE file

    Return value is a tuple of four:
    id_to_lmr: dict of {helix_id: [[L],[M],[R]]} where each list contains
        helix annotations, e.g. ["3'"]
    id_to_content: dict of {helix_id: [(seq_pos, base)]}
    pos_to_base: dict of {seq_pos: base}, e.g. {1:'A',2:'G'}
    unpaired: list of [(seq_pos, base)], all unpaired positions in the seq
    """
    
    id_to_lmr = {} # {temp_id: (L, M, R)}
    id_to_content = {} # {temp_id: [(seq_pos, sym)]}
    unpaired = [] # [(seq_pos, sym)]
    pos_to_base = {}

    curr_id = 0
    in_stem = False
    in_bulge = False
    seq_pos = -1
    for idx, sym in enumerate(seq):
        if idx%2==0: ## symbols
            if sym == ' ':
                continue
            elif sym == '[' or sym == '^':
                in_stem = True
                curr_id += 1
                LMR = get_left_middle_right(num, idx, seq)
                id_to_lmr[curr_id] = LMR
                id_to_content[curr_id] = []
            elif sym == ']':
                if not in_stem:
                    raise BasicStructureError("Closing bracket without opening")
                if in_bulge:
                    raise BasicStructureError(\
                        "Closing stem without closing bulge")
                in_stem = False
            elif sym == '{':
                in_bulge = True
            elif sym == '}':
                if not in_bulge:
                    raise BasicStructureError, "Closing bulge without opening"
                in_bulge = False
            elif sym == '(' or sym == ')': #non-standard base pair
                continue
            else:
                raise BasicStructureError, "Unknown symbol: %s"%(sym)
        else: #bases
            seq_pos += 1
            pos_to_base[seq_pos] = sym
            if sym == '-':
                unpaired.append((seq_pos, sym))
                continue
            else:
                if not in_stem or in_bulge:
                    unpaired.append((seq_pos, sym))
                    continue
                if in_stem and not in_bulge:
                    id_to_content[curr_id].append((seq_pos, sym))
                    continue

    
    if len(id_to_lmr) != len(id_to_content):
        raise BasicStructureError("Failed to extract structure information")
    if len(pos_to_base) !=\
        len(unpaired) + sum(map(len, id_to_content.values())):
        raise BasicStructureError("Some positions missing")
    return id_to_lmr, id_to_content, pos_to_base, unpaired

# =============================================================================
# BASIC CHECKS AND UTILITY FUNCTIONS
# =============================================================================

def basic_ann(helix_ann):
    """Return the basic annotation of a helix annotation (prime stripped)
    
    helix_ann -- string, formatted as in HelixNumbering, e.g. "E3'", "48"
    """
    return helix_ann.strip("'")

def opposite_helix(helix_ann):
    """Return the opposite annotation (in terms of the prime) of a helix
    
    helix_ann -- string, formatted as in HelixNumbering, e.g. "E3'", "48"

    Return the prime version if the original doesn't have a prime. Strip
    off the prime of the original does has a prime.
    """
    if helix_ann.endswith("'"):
        return helix_ann.strip("'")
    else:
        return helix_ann+"'"

def helix_to_pairs(upstream, downstream):
    """Returns list of paired positions from upstream and downstream content.

    upstream -- list of tuples (seq_pos, base), e.g. (1602, 'G')
    downstream -- list of tuples (seq_pos, base), e.g. (1602, 'G')
    """
    if len(upstream) != len(downstream):
        raise HelixError("Trying to pair helices with different lengths")
    pairs = []
    downstream.reverse()
    for up_base, down_base in zip(upstream, downstream):
        pairs.append((up_base[0], down_base[0]))
    return pairs

def set_merger(sets):
    """Merge sets into the smallest number of sets without overlap.
    
    sets -- a list of sets of elements, e.g. [[1,2,3],[3,4,5],[1,5,6]]

    It will merge sets that overlap (contain the same elements)
    For example:
    [1,2],[6,7],[2,3] --> [1,2,3],[6,7]
    [1,2],[4,5],[1,4],[6,7,8],[7,9,10] --> [1,2,4,5],[6,7,8,9,10]
    [1,3,5],[6,7,8] --> [1,3,5],[6,7,8]
    """
    a = sets
    new_res = []
    while 1:
        seen = {}
        for i in a:
            if tuple(i) in seen:
                continue
            temp = set(i)
            for j in a:
                if tuple(j) in seen:
                    continue
                if i == j:
                    seen[(tuple(j))] = None
                    continue
                if len(temp.union(j)) < len(temp) + len(j):
                    temp.update(j)
                    seen[tuple(i)] = None
                    seen[tuple(j)] = None
            new_res.append(temp)
        if len(new_res) < len(a):
            a = new_res
            new_res = []
        else:
            break
    return new_res

def matching_ann(up_poss, down_poss, flexible=False):
    """Return helix annotation of match given options for up and downstream
    """
    # collect upstream options
    options = {}
    for i in up_poss:
        if not i.endswith("'"):
            options[i] = None
    # add downstream options
    for i in down_poss:
        if i.endswith("'") and opposite_helix(i) in options:
            options[opposite_helix(i)] = i
    # pick out viable matches
    viable = []
    for k,v in options.items():
        if v is not None:
            viable.append(k)
    # if not flexible, return an answer only if there's ONE viable option
    if not flexible:
        if len(viable) == 1:
            return viable[0]
        else:
            return None
    # if flexible, return an answer if ONE or TWO viable options
    if flexible:
        if len(viable) == 1:
            return viable[0]
        elif len(viable) == 0:
            return None
        elif len(viable) == 2:
            return viable[0]
        else:
            return None

# =============================================================================
# CONTROL
# =============================================================================

def get_helix_numbering(filename):
    """Return helix numbering from filename
    
    Helix numbering is usually the last line in the file and you don't
    want to keep the possibly very large file in memory to reach the 
    last line.
    """
    f = open(filename)
    for line in f:
        if 'Helix numbering' in line:
            f.close()
            return line.rstrip()
    f.close()
    return None

def get_ann_seq_num(rdb_seq, helix_numbering, split_on=' '*5):
    """Return seq annotation, rdb_sequence, and helix numbering
    """
    try:
        rdb_seq, ann = rdb_seq.rsplit(split_on)
        rdb_seq = rdb_seq.rstrip()
        ann = ann.strip()
    except ValueError:
        ann = None
    try:
        helix_numbering, helix_ann = helix_numbering.rsplit(split_on)
        helix_numbering = helix_numbering.rstrip()
    except ValueError:
        helix_ann = None
    return ann, rdb_seq, helix_numbering 
         
def process_DCSE_aln(filename, strict=True):
    """Yield StructureAnnotation objects (or raise an error)
    """
    helix_numbering = get_helix_numbering(filename)
    f = open(filename)
    for line in f:
        if line.startswith('R') or 'Helix numbering' in line:
            continue
        if line.rstrip():
            try:
                ann, rdb_seq, helix_numbering =\
                    get_ann_seq_num(line.rstrip(), helix_numbering)
                #yield rdb_to_bpseq(line.rstrip(), helix_numbering)
                yield StructureAnnotation(ann, rdb_seq, helix_numbering)
            except (BasicStructureError, AnnotationError, HelixError), e:
                if strict:
                    raise DcseParseError(str(e))
                else:
                    continue
    f.close()

if __name__ == "__main__":
    pass
    #for i in process_DCSE_aln(argv[1]):
    #    continue
