#!/usr/bin/env python
# test_ct.py

"""
Author: Sandra Smit (Sandra.Smit@colorado.edu)

Revision History:
File created on 10 Aug 2007.

"""
from __future__ import division
from cogent.util.unit_test import TestCase, main
from cogent.core.moltype import RNA
from cogent.struct.rna2d import Pairs
from rnalib.parse.ct import CtError, is_ct_line, ct_finder, ct_head_body,\
    ct_parse_header, bralibase_parse_header,\
    ct_parse_content,\
    ct_integrate,\
    CtParser

class CtTests(TestCase):

    def test_is_ct_line(self):
        true_lines = [\
            ' 2904 ENERGY =     0.0    E.coli',\
            '  377 ENERGY =     0.0    1',\
            '2904 Energy = 0.0',\
            ]
        for tl in true_lines:
            self.assertEqual(is_ct_line(tl), True)

    def test_ct_finder(self):
        f = ct_finder
        lines = UNAFOLD.split('\n')
        obs = list(f(lines))
        self.assertEqual(len(obs), 2)
        self.assertEqual(len(obs[0]), 15)
        self.assertEqual(len(obs[1]), 12)
        self.assertEqual(obs[1][-1].strip(),\
            '11	A	10	0	0	11	10	0')

    def test_ct_head_body(self):
        f = ct_head_body
        lines = "a\nb\nc\nd".split('\n')
        self.assertEqual(f(lines), ('a',['b','c','d']))

class TestHeaderParsers(TestCase):

    def test_ct_parse_header(self):
        f = ct_parse_header
        line = ' 2904 ENERGY =     0.0    E.coli'
        self.assertEqual(f(line), '2904 ENERGY =     0.0    E.coli')

    def test_bralibase_parse_header(self):
        f = bralibase_parse_header
        line = ' 2904 ENERGY =     0.0    E.coli'
        self.assertEqual(f(line), (2904, 0.0, 'E.coli'))
        line = '2904 Energy = 0.0'
        self.assertEqual(f(line), (2904, 0.0, None))
        line = '  377 ENERGY =     0.0    1'
        self.assertEqual(f(line), (377, 0.0, '1'))
        line = '  377 ENERGY = 0 E.coli RNase P RNA'
        self.assertEqual(f(line), (377, 0.0, 'E.coli RNase P RNA'))

class TestContentParsers(TestCase):

    def test_ct_parse_content_errors(self):
        """ct_parse_content: error handling"""
        f = ct_parse_content
        lines = "3 G 2 4 0 3\n4 A 3 5 0 4".split('\n')
        self.assertRaises(CtError, f, lines)
        lines = "A G 2 4 0 3\n4 A 3 5 0 4".split('\n')
        self.assertRaises(CtError, f, lines)
        lines = "1 G 0 2 0 1 7\n2 A 1 3 0 2 8".split('\n')
        self.assertRaises(CtError, f, lines)
        lines = "1 G 0 2 0 1\n3 A 2 4 0 3".split('\n')
        self.assertRaises(CtError, f, lines) 
        lines = "1 G 0 2 0 1\n2 A 1 4 0 2".split('\n')
        self.assertRaises(CtError, f, lines)
        lines = "1 G 0 2 2 1\n2 A 1 3 0 2".split('\n')
        self.assertRaises(CtError, f, lines)

    def test_ct_parse_content(self):
        """ct_parse_content: valid input"""
        f = ct_parse_content
        lines = BODY_1.split('\n')
        self.assertEqual(f(lines),('UAC', []))
        lines = BODY_2.split('\n')
        self.assertEqual(f(lines),('GCGGAUUUAG', [(2,7),(3,6)]))

class TestIntegration(TestCase):

    def test_ct_integrate(self):
        f = ct_integrate
        header = "344 Energy=-3.4 E.coli"
        body = 'ACCUGGAUG', Pairs([(1,5), (2,4)])
        obs = f(header, body)
        self.assertEqual(obs[0], RNA.Sequence('ACCUGGAUG'))
        self.assertEqual(obs[0].Info, {'SEQ_STRUCT': header, 'Refs':{}})


class TestParsers(TestCase):

    def test_CtParser_bralibase(self):
        f = CtParser
        lines = BRALI_1.split('\n')
        exp_seq = 'GGUUCCGCGA'
        exp_pairs = [(0,5),(1,4)]
        obs = list(f(lines, program='bralibase'))
        self.assertEqual(len(obs), 1)
        self.assertEqual(obs[0], (exp_seq, exp_pairs))

    def test_CtParser_unafold(self):
        f = CtParser
        lines = UNAFOLD.split('\n')
        obs = list(f(lines, program='unafold'))
        for seq, struct in obs:
            print seq
            print seq.Info
            print struct
        



# =============================================================================
# TEST DATA
# =============================================================================

BODY_1=\
"""1 U 0 2 0 1
2 A 1 3 0 2
3 C 2 4 0 3"""

BODY_2=\
"""    1 G       0    2    0    1
    2 C       1    3    0    2
    3 G       2    4    8    3
    4 G       3    5    7    4
    5 A       4    6    0    5
    6 U       5    7    0    6
    7 U       6    8    4    7
    8 U       7    9    3    8
    9 A       8   10    0    9
   10 G       9   11    0   10"""

BRALI_1=\
"""10 Energy = 0.0
1  G  0  2  6  1
2  G  1  3  5  2
3  U  2  4  0  3
4  U  3  5  0  4
5  C  4  6  2  5
6  C  5  7  1  6
7  G  6  8  0  7
8  C  7  9  0  8
9  G  8  10  0  9
10  A  9  11  0  10
"""

UNAFOLD=\
"""14	dG = 0.9	 sequence one
1	A	0	2	0	1	0	0
2	A	1	3	0	2	0	0
3	A	2	4	0	3	0	0
4	C	3	5	0	4	0	5
5	G	4	6	14	5	4	6
6	U	5	7	13	6	5	0
7	G	6	8	0	7	0	0
8	A	7	9	0	8	0	0
9	C	8	10	0	9	0	0
10	U	9	11	0	10	0	0
11	U	10	12	0	11	0	0
12	U	11	13	0	12	0	0
13	G	12	14	6	13	0	14
14	C	13	0	5	14	13	0
11	dG = -1.2	 sequence two
1	C	0	2	10	1	0	2
2	G	1	3	9	2	1	3
3	U	2	4	8	3	2	0
4	A	3	5	0	4	0	0
5	G	4	6	0	5	0	0
6	C	5	7	0	6	0	0
7	U	6	8	0	7	0	0
8	A	7	9	3	8	0	9
9	C	8	10	2	9	8	10
10	G	9	11	1	10	9	11
11	A	10	0	0	11	10	0
"""


SFOLD_WEB=\
"""SFOLD of:  [initially 1.9]    MFE Structure    Check: 0 from: 1 to: 12
Length: 12 Energy: 1.9 ..
    1 U     0     2     0      1
    2 A     1     3     0      2
    3 C     2     4    10      3
    4 G     3     5     9      4
    5 A     4     6     0      5
    6 U     5     7     0      6
    7 G     6     8     0      7
    8 C     7     9     0      8
    9 U     8    10     4      9
   10 G     9    11     3     10
   11 G    10    12     0     11
   12 C    11     0     0     12
"""

if __name__ == "__main__":
    main()
