#!/usr/bin/env python
# dcse_to_vienna.py

"""Script to extract sequence and structure information from sequences in the
European rRNA database. It creates a single file in FASTA format:
>sequence ID
sequence
structure (Vienna format, no pseudoknots)

Author: Sandra Smit (Sandra.Smit@colorado.edu)

Revision History:
File created on 23 Jan 2007.
"""
from __future__ import division
import sys
from optparse import OptionParser
from cogent.struct.rna2d import Pairs
from rnalib.struct.pairs_util import gapped_to_ungapped
from rnalib.parse.rdb_dcse import process_DCSE_aln

def parse_command_line_parameters():
    """ Parses command line arguments """
    usage = "usage: %prog [options] dcse_aln_filename"
    result = OptionParser(usage=usage)

    #result.add_option('-v','--verbose',action='store_true',dest='verbose')
    result.add_option('-d','--degap',action='store_true',dest='degap',\
        help="degap sequences and structures")
    result.add_option('-r','--rule',action='store', type='string',\
        dest='rule',\
        help="rule for removing pseudoknots, options: first or majority; default is majority")
    result.add_option('-o','--out',action='store', type='string',\
        dest='out', help="name of output file; has to be specified")

    result.set_defaults(verbose=False)
    result.set_defaults(degap=False)
    result.set_defaults(rule='majority')
    result.set_defaults(out=None)
    return result.parse_args()


if __name__ == "__main__":
    opts,arg = parse_command_line_parameters()
    if opts.out is None:
        print "Specify a name for the output file."
        print "Run 'python dcse_to_vienna.py -h' for more help"
        sys.exit()
    if arg == []:
        print "Specify a filename."
        print "Run 'python dcse_to_vienna.py -h' for more help"
        sys.exit()
        
    aln_filename = arg[0]
    out_file = open(opts.out,'w')
    
    for struct_ann in process_DCSE_aln(aln_filename, strict=False):
        orig_seq = struct_ann.GappedSeq
        orig_pairs = Pairs(struct_ann.Pairs).directed()
        orig_pairs.sort()
        
        if orig_pairs.hasPseudoknots():
            nested = orig_pairs.nested(rule=opts.rule)
        else:
            nested = orig_pairs
        vienna_no_pseudo = nested.toVienna(len(orig_seq))

        if opts.degap:
            seq, pairs = gapped_to_ungapped(orig_seq, nested)
            vienna_no_pseudo = pairs.toVienna(len(seq))
        else:
            seq = orig_seq

        assert len(seq) == len(vienna_no_pseudo)

        out_file.write(">"+struct_ann.SeqId+'\n')
        out_file.write(seq+'\n')
        out_file.write(vienna_no_pseudo+'\n')
