#!/usr/bin/env python
# cl_util.py

"""
Author: Sandra Smit (S.Smit@few.vu.nl)

Revision History:
File created on 20 December 2007
2007/12/20 Sandra Smit: imported AlnToProfile
(2007/11/21 Changed to numpy. Checked against old Numeric version.)
"""

from __future__ import division
from cogent.core.profile import Profile, CharMeaningProfile 
from cogent.core.alignment import Alignment
from numpy import array, take, add, uint8, fromstring, average, flatnonzero,\
    argmax

def AlnToProfile(aln, alphabet=None, char_order=None, split_degenerates=False,\
    weights=None):
    """Generates a Profile object from an Alignment.

    aln: Alignment object
    alphabet (optional): an Alphabet object (or list of chars, but if you 
        want to split degenerate symbols, the alphabet must have a 
        Degenerates property. Default is the alphabet of the first seq in 
        the alignment.
    char_order (optional): order of the characters in the profile. Default
        is list(alphabet)
    split_degenerates (optional): Whether you want the counts for the 
        degenerate symbols to be divided over the non-degenerate symbols they
        code for.
    weights (optional): dictionary of seq_id: weight. If not entered all seqs
        are weighted equally

    A Profile is a position x character matrix describing which characters
    occur at each position of an alignment. The Profile is always normalized,
    so it gives the probabilities of each character at each position.
    
    Ignoring chars: you can ignore characters in the alignment by not putting
    the char in the CharOrder. If you ignore all characters at a particular
    position, an error will be raised, because the profile can't be normalized.

    Splitting degenerates: you can split degenerate characters over the 
    non-degenerate characters they code for. For example: R = A or G. So,
    an R at a position counts for 0.5 A and 0.5 G.
   
    Example:
    seq1    TCAG    weight: 0.5
    seq2    TAR-    weight: 0.25
    seq3    YAG-    weight: 0.25
    Profile(aln,alphabet=DnaAlphabet,char_order="TACG",weights=w,
    split_degenerates=True)
    Profile:
       T      A      C      G
    [[ 0.875  0.     0.125  0.   ]
     [ 0.     0.5    0.5    0.   ]
     [ 0.     0.625  0.     0.375]
     [ 0.     0.     0.     1.   ]]
    """

    if alphabet is None:
        alphabet = aln.values()[0].Alphabet
    if char_order is None:
        char_order = list(alphabet)
    if weights is None:
        weights = dict.fromkeys(aln.keys(),1/len(aln))
    
    char_meaning = CharMeaningProfile(alphabet, char_order,\
        split_degenerates)

    profiles = []
    for k,v in aln.items():
        new_p = char_meaning.Data.take(fromstring(str(v).upper(),\
            dtype=uint8), axis=0) * weights[k]

        profiles.append(new_p)
    s = reduce(add,profiles)
    
    result = Profile(s,alphabet, char_order)
    try:
        result.normalizePositions()
    except:
        raise ValueError,\
        "Probably one of the rows in your profile adds up to zero,\n "+\
        "because you are ignoring all of the characters in the "+\
        "corresponding\n column in the alignment"
    return result


def compare_classifications(one, other):
    """Return fraction of two classifications that is identical

    one, other -- classification object or other strings

    The two strings should have the same length, otherwise
    an error is raised.
    """
    if not len(one) == len(other):
        raise ValueError("Input strings should have the same length")
    return len(flatnonzero(array(one, dtype='c') ==\
        array(other, dtype='c')))/len(one)

def consistency_score(cl_profile):
    """Calculate the consistency score of a profile

    cl_profile -- Profile object

    consistency_score is defined as the average rowMax over all positions
    """
    rm = average(cl_profile.rowMax())
    return rm


def consensus_profile(cl_profile, cutoff=0.5):
    """Return consensus profile

    If two characters occur with equal frequencies the first in the profile
        is added.
    """
    co = cl_profile.CharOrder
    result = []
    for r in cl_profile.Data:
        larger = r>cutoff
        if larger.any():
            result.append(co[argmax(larger)])
        else:
            result.append(None)
    return result

def argmax_cl(cl_profile):
    """Return string of structural elements with max frequency

    cl_profile -- Profile object

    If two characters occur with equal frequency the first is chosen
    """
    l = [cl_profile.CharOrder[i] for i in argmax(cl_profile.Data, axis=1)]
    return ''.join(l)

if __name__ == "__main__":

    aln = Alignment(['SSSLLLSSS','SBSLLLSBS','OSSLLLSSO','OSSLLLSSO'])
    print aln
    p = AlnToProfile(aln, alphabet="SLBO", char_order="SLBO")
    print 'CL PROFILE'
    print p
    print 'ROW MAX'
    print p.rowMax()
    print 'AVERAGE ROW MAX'
    print average(p.rowMax())

    print 'CONSISTENCY', consistency_score(p)
    print 'CONSENSUS', consensus_profile(p)
    print 'CONSENSUS 0.4', consensus_profile(p, 0.4)
    print 'ARGMAX', argmax_cl(p)

    print compare_classifications('SLBO','SLSL')

