#!/usr/bin/env python
# classifier.py

"""
Author: Sandra Smit (Sandra.Smit@colorado.edu)

Revision History:
File created on 17 Oct 2006.
10/17/06 Sandra Smit: grouped several pieces of code in this file.

"""
from __future__ import division
from numpy import array, searchsorted
from cogent.util.misc import flatten
from cogent.struct.rna2d import classify
from rnalib.core.classification import Classification

class ClassificationError(Exception):
    """Raises Classification Error"""
    pass

# ==========================================================
# Helper functions
# ==========================================================

def get_pairing_info(pairs):
    """Returns list of paired positions (sorted) and a dict of partners

    pairs: Pairs object, e.g. Pairs([(3,7),(4,6)])

    Output:
    [3,4,6,7] & {3:7,4:6,6:4,7:3}
    """
    if pairs.hasConflicts():
        raise ClassificationError(\
            "Ambiguous pairing information: Pairs object contains conflicts")    
    p = pairs.directed() # remove Nones and duplicates
    paired_pos = flatten(p)
    paired_pos.sort()
    sym_pairs = p.symmetric()
    partners = dict(sym_pairs)
    return paired_pos, partners

def is_stem(pos, paired_pos, partners):
    """Returns True if position is paired

    pos: sequence position (integer)
    paired_pos: sorted list of paired positions [3,4,6,7]
    partners: dictionary of pairings {3:7, 4:6, 6:4, 7:3} 
    """
    if pos in partners:
        return True
    return False

def is_end(pos, paired_pos, partners):
    """Returns True if position is on the 3' or 5' end.
    pos: sequence position (integer)
    paired_pos: sorted list of paired positions [3,4,6,7]
    partners: dictionary of pairings {3:7, 4:6, 6:4, 7:3}
    """
    if pos < paired_pos[0] or pos > paired_pos[-1]:
        return True
    return False

def is_loop(pos, paired_pos, partners):
    """Returns True if position is in pure hairpin loop.
    
    Hairpin loop means that no other base in the loop is interacting
    with another base. In other words: the last paired base before the loop
    should be pairing with the first paired base after the loop.
    
    pos: sequence position (integer)
    paired_pos: sorted list of paired positions [3,4,6,7]
    partners: dictionary of pairings {3:7, 4:6, 6:4, 7:3}
    """
    if pos in partners:
        return False
    if is_end(pos, paired_pos, partners):
        return False
    idx = array(paired_pos).searchsorted(pos)
    if partners[paired_pos[idx-1]] == paired_pos[idx]:
        return True
    return False

def is_bulge(pos, paired_pos, partners):
    """Returns True if position is in bulge.
    
    A bulge is a collection of bases that interrupt a stem, either on one
    side of the helix or on both sides. The base pair before the bulge and
    the base pair after the bulge should be adjacent. Both paired_pos and
    partners are needed. For example:
    paired_pos: [3,4,7,8,12,13,14,15]
    partners: {3:15,4:14,7:13, 8:12} <- and the opposite pairs
    position 5 is in a bulge, because position 4 pairs with 14 and position 7
    pairs with 13 and position 13 and 14 are adjacent in the paired_pos list.

    pos: sequence position (integer)
    paired_pos: sorted list of paired positions [3,4,6,7]
    partners: dictionary of pairings {3:7, 4:6, 6:4, 7:3}
    """
    if pos in partners:
        return False
    if is_end(pos, paired_pos, partners):
        return False
    idx = array(paired_pos).searchsorted(pos)
    before = paired_pos[idx-1]
    after = paired_pos[idx]
    if partners[before] == after:
        return False
    if abs(paired_pos.index(partners[before]) -\
        paired_pos.index(partners[after])) == 1:
        return True
    return False

def is_flexible(pos, paired_pos, partners):
    """Returns True if position is a flexible base
   
    A base is categorized as flexible when all positions before the base
    pair with other positions before the base and when all positions after
    the base pair with positions after the base. 
        
    pos: sequence position (integer)
    paired_pos: sorted list of paired positions [3,4,6,7]
    partners: dictionary of pairings {3:7, 4:6, 6:4, 7:3}
    """
    if pos in partners:
        return False
    if is_end(pos, paired_pos, partners):
        return False
    idx = array(paired_pos).searchsorted(pos)
    if partners[paired_pos[idx]] < pos:
        return False
    for i in paired_pos[:idx]:
        if i > pos or partners[i] > pos:
            return False
    for i in paired_pos[idx:]:
        if i < pos or partners[i] < pos:
            return False
    return True

# ==========================================================
# Real classifiers
# ==========================================================

def paired_unpaired(pairs, length):
    """Returns Classification object, P=paired, U=unpaired
    
    pairs: Pairs object.
    length: length of the classification.

    This method will take any Pairs object. It doesn't care about conflicts,
    duplicates, or pseudoknots.
    """
    paired = dict.fromkeys(flatten(pairs.paired()))
    cl = []
    for pos in range(length):
        if pos in paired:
            cl.append('P')
        else:
            cl.append('U')
    return Classification(''.join(cl))

def stem_loop_other(pairs, length, force_slow=False):
    """Returns classification string (same length as seq) with S, L, and O.

    Use when you just want to distinguish between paired regions, hairpins,
    and other positions.

    seq: Sequence object or string or anything that len works on.
    pairs: list of tuples or Pairs object. e.g. [(3,7),(4,6)]
    """
    if pairs.hasConflicts():
        raise ClassificationError(\
            "Ambiguous pairing information: Pairs object contains conflicts")
    if not pairs.hasPseudoknots() and not force_slow: # quick way
        v = pairs.toVienna(length)
        c = classify(v)
        c = c.replace('B','O')
        c = c.replace('J','O')
        c = c.replace('E','O')
        c = c.replace('F','O')
        return Classification(c)
    else: # slow
        paired_pos, partners = get_pairing_info(pairs)
        c = []
        for pos in range(length):
            if is_stem(pos, paired_pos, partners):
                c.append('S')
            elif is_loop(pos, paired_pos, partners):
                c.append('L')
            else:
                c.append('O')
        return Classification(''.join(c))

SLO = stem_loop_other

def stem_loop_bulge_other(pairs, length, force_slow=False):
    """Returns classification string (same length as seq) with S, L, B, and O.

    Use when you just want to distinguish between paired regions, hairpins,
    bulges, and other positions.

    seq: Sequence object or string or anything that len works on.
    pairs: list of tuples or Pairs object. e.g. [(3,7),(4,6)]
    force_slow: boolean, forces the algorithm to take the slow route,
        implemented for testing

    This method will raise an error when the Pairs object has conflicts. 
    If no pseudoknots are present, the fast way will be used, otherwise
    the slow classification will be used. 
    """
    if pairs.hasConflicts():
        raise ClassificationError(\
            "Ambiguous pairing information: Pairs object contains conflicts")
    if not pairs.hasPseudoknots() and not force_slow: # quick way
        v = pairs.toVienna(length)
        c = classify(v)
        c = c.replace('J','O')
        c = c.replace('E','O')
        c = c.replace('F','O')
        return Classification(c)
    else: # slow
        paired_pos, partners = get_pairing_info(pairs)
        c = []
        for pos in range(length):
            if is_stem(pos, paired_pos, partners):
                c.append('S')
            elif is_end(pos, paired_pos, partners):
                c.append('O')
            elif is_loop(pos, paired_pos, partners):
                c.append('L')
            elif is_bulge(pos, paired_pos, partners):
                c.append('B')
            else:
                c.append('O')
        return Classification(''.join(c))

SLBO = stem_loop_bulge_other

def six_original(pairs, length):
    """Return six way classification string.

    pairs: Pairs object
    length: length of the classification string.

    This is the original six way classification. It only works on fully nested
    structures. An error will be raised when Pairs has either conflicts or
    pseudoknots.
    """
    if pairs.hasConflicts():
        raise ClassificationError(\
            "Ambiguous pairing information: Pairs object contains conflicts")
    elif pairs.hasPseudoknots():
        raise ClassificationError(\
            "The original six way classification on fully "+\
            "nested structures can't handle pseudoknots")
    else:
        v = pairs.toVienna(length)
        return Classification(classify(v))

def six_general(pairs, length, force_slow=False):
    """Return six way classification string. SLBOEF

    pairs: Pairs object
    length: length of the classification string.

    This is the general six way classification. Multi-loop junctions and
    pseudoknotted regions are grouped in the class 'other' (O). 
    When possible (no pseudoknots), the fast way will be chosen and the Js 
    will be replaces by Os. Otherwise, the slow route is chosen. 
    """
    if pairs.hasConflicts():
        raise ClassificationError(\
            "Ambiguous pairing information: Pairs object contains conflicts")
    if not pairs.hasPseudoknots() and not force_slow:
        v = pairs.toVienna(length)
        c = classify(v)
        c = c.replace('J','O')
        return Classification(c)
    else:
        paired_pos, partners = get_pairing_info(pairs)
        c = []
        for pos in range(length):
            if is_stem(pos, paired_pos, partners):
                c.append('S')
            elif is_end(pos, paired_pos, partners):
                c.append('E')
            elif is_loop(pos, paired_pos, partners):
                c.append('L')
            elif is_bulge(pos, paired_pos, partners):
                c.append('B')
            elif is_flexible(pos, paired_pos, partners):
                c.append('F')
            else:
                c.append('O')
        return Classification(''.join(c))



if __name__ == "__main__":
    pass
