#!/usr/bin/env python

from __future__ import division
from random import sample
from cogent.struct.knots import *

def constraint_A(pairs, seq_len, frac=0.5):
    """Fix fraction unpaired positions
    """
    if pairs.hasConflicts():
        raise ValueError("Can't handle conflicts")
    if pairs.hasPseudoknots():
        raise ValueError("Can't handle pseudoknots")
    result = ['.']*seq_len
    prs = PairedRegionsFromPairs(pairs)
    paired = dict.fromkeys(prs.pairedPos())
    unpaired = [idx for idx in range(seq_len) if idx not in paired]
    to_sample = int(frac*len(unpaired)) or 1
    for pos in sample(unpaired, to_sample):
        result[pos] = 'x'
    return ''.join(result)

def constraint_B(pairs, seq_len, frac=0.5):
    """Fix fraction of the base pairs
    """
    p = pairs.directed()
    p.sort()
    if p.hasConflicts():
        raise ValueError("Can't handle conflicts")
    if p.hasPseudoknots():
        raise ValueError("Can't handle pseudoknots")
    result = ['.']*seq_len
    to_sample = int(frac*len(p)) or 1
    to_fix = Pairs(sample(p, to_sample)) 
    for x,y in to_fix:
        result[x] = '('
        result[y] = ')'
    return ''.join(result)

def constraint_C(pairs, seq_len, frac=0.5):
    """Fix fraction of paired regions
    """
    p = pairs.directed()
    p.sort()
    if p.hasConflicts():
        raise ValueError("Can't handle conflicts")
    if p.hasPseudoknots():
        raise ValueError("Can't handle pseudoknots")
    prs = PairedRegionsFromPairs(p)
    result = ['.']*seq_len
    to_sample = int(frac*len(prs)) or 1
    to_fix = PairedRegions(sample(prs, to_sample)).toPairs()
    for x,y in to_fix:
        result[x] = '('
        result[y] = ')'
    return ''.join(result)

def constraint_D(pairs, seq_len, frac=0.5):
    """Fix fraction of unpaired regions
    """
    if pairs.hasConflicts():
        raise ValueError("Can't handle conflicts")
    if pairs.hasPseudoknots():
        raise ValueError("Can't handle pseudoknots")
    result = ['.']*seq_len
    prs = PairedRegionsFromPairs(pairs)
    paired = dict.fromkeys(prs.pairedPos())
    unpaired = []
    curr = []
    for idx in range(seq_len):
        if idx in paired:
            if curr:
                unpaired.append(curr)
                curr = []
        else:
            curr.append(idx)
    if curr:
        unpaired.append(curr)
    to_sample = int(frac*len(unpaired)) or 1
    for region in sample(unpaired, to_sample):
        for pos in region:
            result[pos] = 'x'
    return ''.join(result)

def constraint_E(pairs, seq_len, frac=0.5):
    """Fix pairing direction "<" or ">" for fraction of base pairs
    """
    if pairs.hasConflicts():
        raise ValueError("Can't handle conflicts")
    if pairs.hasPseudoknots():
        raise ValueError("Can't handle pseudoknots")
    prs = PairedRegionsFromPairs(pairs)
    paired = prs.pairedPos()
    upstream = dict.fromkeys(prs.upstream())
    downstream = dict.fromkeys(prs.downstream())

    result = ['.']*seq_len
    to_sample = int(frac*len(paired)) or 1
    to_fix = Pairs(sample(paired, to_sample)) 
    for pos in to_fix:
        if pos in upstream:
            result[pos] = '<'
        elif pos in downstream:
            result[pos] = '>'
        else:
            raise ValueError("Paired pos not in upstream or downstream")
    return ''.join(result)

def constraint_F(pairs, seq_len, frac=0.5):
    """Fix fraction of paired and unpaired regions"""

    if pairs.hasConflicts():
        raise ValueError("Can't handle conflicts")
    if pairs.hasPseudoknots():
        raise ValueError("Can't handle pseudoknots")
    result = ['.']*seq_len
    prs = PairedRegionsFromPairs(pairs)
    paired = dict.fromkeys(prs.pairedPos())
    unpaired = []
    curr = []
    for idx in range(seq_len):
        if idx in paired:
            if curr:
                unpaired.append(curr)
                curr = []
        else:
            curr.append(idx)
    if curr:
        unpaired.append(curr)

    to_sample_unp = int(frac*len(unpaired)) or 1
    for region in sample(unpaired, to_sample_unp):
        for pos in region:
            result[pos] = 'x'
    to_sample_paired = int(frac*len(prs)) or 1
    to_fix = PairedRegions(sample(prs, to_sample_paired)).toPairs()
    for x,y in to_fix:
        result[x] = '('
        result[y] = ')'
    return ''.join(result)


def constraint_unp(seq_len, frac=0.5):
    """Fix stretch of unpaired bases
    """
    result = ['.']*seq_len
    num_unp_pos = int(frac*seq_len) or 1
    start_point_range = range(0,seq_len-num_unp_pos)
    start_point = choice(start_point_range)
    for i in range(num_unp_pos):
        result[start_point+i] = 'x'
    return ''.join(result)

