#!/usr/bin/env python
# pairs_util.py
"""Functions to adjust Pairs from gapped to ungapped or from ungapped to
gapped. Works on strings or Sequence objects, on list of tuples or Pairs
objects.

Author: Sandra Smit (Sandra.Smit@colorado.edu)

6/15/06 Sandra Smit: added comments and made some changes. Added base
    pair comparison functions
6/22/06 Shandy Wikman: written adjust_base (added and changed
    by Sandra on 8/7/06)
8/7/06 Sandra Smit: added compare_random_to_correct.
10/23/06 Sandra Smit: implemented new general gapped_to_ungapped,
    and gapped_to_ungapped. Changed adjust_pairs_from_mapping
    to handle conflicts, pseudoknots, duplications, and Nones.
10/25/06 Sandra Smit: adjusted Shandy's implementation of Garder & Giegerich
    2004 metrics to match their outcomes. Compared results to the output
    of their script (compare_ct.pm).
10/31/06 Sandra Smit: added adjust_base_structures to adjust several structure
    to some offset.
12/11/06 Sandra Smit: removed error in cc_formula when TP <= 0. Instead, just
    returns 0. Thanks to Shandy for finding this. Added test to all_metrics
    to check whether reference and/or prediction is empty.
1/11/07 Sandra Smit: added vienna_mask function to get a Pairs object from
    a ViennaStructure and a pairing mask (used in tmRNA and SRP database).
1/16/07 Sandra Smit: added pairs_to_mask to create a pairing mask from a 
    Pairs object.
"""
from __future__ import division
from copy import copy
from string import strip
from numpy import array, sqrt, searchsorted, flatnonzero, take, sum
from cogent.struct.rna2d import Pairs
from cogent.parse.fasta import MinimalFastaParser
from cogent.struct.knots import *

class PairsAdjustmentError(Exception):
    pass

# ==================================================================
# Adjustment functions for Pairs objects
# ==================================================================

def adjust_base(pairs, offset):
    """Returns new Pairs with values shifted by offset

    pairs: Pairs object or list of tuples
    offset: integer

    Adjusts the base of a pairs object or a list of pairs according to
        the given offset.
    There's no validation in here! It is possible negative values are
        returned -> user responsibility.
    This method treats all pairs as equal. It'll return a pairs object
        of exactly the same length as the input, including pairs containing
        None, and duplicates.

    Example: adjust_base(Pairs([(2,8),(4,None)]), 2) --> [(4,10),(6,None)]
    """
    if not isinstance(offset, int):
        raise PairsAdjustmentError("adjust_base: offset should be integer")
    result = Pairs()
    for x, y in pairs:
        if x is not None:
            new_x = x + offset
        else:
            new_x = x
        if y is not None:
            new_y = y + offset
        else:
            new_y = y
        result.append((new_x, new_y))
    assert len(result) == len(pairs)
    return result

def adjust_base_structures(structures, offset):
    """Adjusts the base of all structures by offset

    structures: list of Pairs objects
    offset: integer
    """
    result = []
    for struct in structures:
        result.append(adjust_base(struct, offset))
    return result

def adjust_pairs_from_mapping(pairs, mapping):
    """Returns new Pairs object with numbers adjusted according to map

    pairs: list of tuples or Pairs object
    mapping: dictionary containing mapping of positions from
        one state to the other (e.g. ungapped to gapped)
         For example:
        {0: 0, 1: 1, 2: 3, 3: 4, 4: 6, 5: 7, 6: 9, 7: 10, 8: 12}

    When the Pairs object corresponds to an ungapped sequence and
        you want to insert gaps, use a mapping from ungapped to gapped.
    When the Pairs object corresponds to a gapped sequence and you
        want to degap it, use a mapping from gapped to ungapped.
    """
    result = Pairs()
    for x,y in pairs:
        if x is None:
            new_x = None
        elif x not in mapping:
            continue
        else:
            new_x = mapping[x]
        if y is None:
            new_y = None
        elif y not in mapping:
            continue
        else:
            new_y = mapping[y]
        result.append((new_x, new_y))

    return result

def delete_gaps_from_pairs(pairs, gap_list):
    """Returns Pairs object with pairs adjusted to gap_list

    pairs: list of tuples or Pairs object
    gap_list: list or array of gapped positions that should be removed
        from the pairs object

    Base pairs of which one of the partners or both of them are in 
        the gap list are removed. If both of them are not in the gap_list, the
        numbering is adjusted according to the gap_list.
    When at least one of the two pair members is in the gap_list, the
        pair will be removed. The rest of the structure will be left
        intact. Pairs containing None, duplicates, pseudoknots, and 
        conflicts will be maintained and adjusted according to the gap_list.
    """
    if not gap_list:
        result = Pairs()
        result.extend(pairs)
        return result

    g = array(gap_list)
    result = Pairs()
    for up, down in pairs:
        if up in g or down in g:
            continue
        else:
            if up is not None:
                new_up = up - g.searchsorted(up)
            else:
                new_up = up
            if down is not None:
                new_down = down - g.searchsorted(down)
            else:
                new_down = down
            result.append((new_up, new_down))
    return result

def insert_gaps_in_pairs(pairs, gap_list):
    """Adjusts numbering in pairs according to the gap list.

    pairs: Pairs object
    gap_list: list of integers, gap positions in a sequence

    The main assumptionis that all positions in pairs correspond to
    ungapped positions. If this is not true, the result will be meaningless.
    """
    if not gap_list:
        new = Pairs()
        new.extend(pairs)
        return new

    ungapped = []
    for idx in range(max(gap_list)+2):
        if idx not in gap_list:
            ungapped.append(idx)
    new = Pairs()
    for x,y in pairs:
        if x is not None:
            try:
                new_x = ungapped[x]
            except IndexError:
                new_x = ungapped[-1] + (x-len(ungapped)+1)
        else:
            new_x = x
        if y is not None:
            try:
                new_y = ungapped[y]
            except IndexError:
                new_y = ungapped[-1] + (y-len(ungapped)+1)
        else:
            new_y = y
        new.append((new_x, new_y))
    return new


def get_gap_symbol(seq):
    """Return gap symbol.

    seq: Sequence object or plain string.

    Should be able to handle cogent.core Sequence and ModelSequence object.
    If the input sequence doesn't have a MolType, '-' will be returned
    as default.
    """
    try:
        gap = seq.MolType.Alphabet.Gap
    except AttributeError:
        gap = '-'
    return gap

def get_gap_list(gapped_seq, gap_symbol=None):
    """Return list of gapped positions.

    gapped_seq: string of sequence object. Should be able to handle
        old_cogent.base.sequence object, cogent.core Sequence and
        ModelSequence object, or plain strings.
    gap_symbol: gap symbol. Will be used for plain strings.
    """
    try: 
        gap_list = gapped_seq.gapList() #should work for RnaSequence
    except AttributeError:
        try:
            gap_list = flatnonzero(gapped_seq.gaps())
        except AttributeError:
            gap_list = flatnonzero(array(gapped_seq,'c') == gap_symbol)
    try: # if gap_list is array, convert it to list
        gap_list = gap_list.tolist()
    except AttributeError: #already a list
        pass
    return gap_list

def degap_model_seq(seq):
    """Returns ungapped copy of self, not changing alphabet.
    
    This function should actually be a method of ModelSequence. Right
        now the ungapped method is broken, so this is a temporary 
        replacement.
    """
    if seq.Alphabet.Gap is None:
        return seq.copy()
    d = take(seq._data, flatnonzero(seq.nongaps()))
    return seq.__class__(d, Alphabet=seq.Alphabet, Name=seq.Name, \
        Info=seq.Info)

def degap_seq(gapped_seq, gap_symbol=None):
    """Return ungapped copy of sequence.

    Should be able to handle
        old_cogent.base.sequence object, cogent.core Sequence and
        ModelSequence object, or plain strings.
    """
    # degap the sequence
    try: #should work for old and new RnaSequence
        ungapped_seq = gapped_seq.degap()
    except AttributeError:
        try:
            ungapped_seq = degap_model_seq(gapped_seq)
        except AttributeError:
            ungapped_symbols = take(array(list(gapped_seq)),\
                flatnonzero((array(list(gapped_seq)) != gap_symbol)))
            ungapped_seq = ''.join(ungapped_symbols)
    return ungapped_seq

def gapped_to_ungapped(gapped_seq, gapped_pairs):
    """Returns ungapped sequence and corresponding Pairs object
    
    gapped_seq: string of characters (can handle Sequence, ModelSequence,
        str, or old_cogent Sequence objects).
    gapped_pairs: Pairs object, e.g. [(3,7),(4,6)]. The Pairs object should
        correspond to the gapped sequence version.

    The gap_symbol will be extracted from the sequence object. In case
    the gapped_seq is a simple str, a '-' will be used as default.
    """
    gap_symbol = get_gap_symbol(gapped_seq)
    gap_list = get_gap_list(gapped_seq, gap_symbol)
    ungapped_seq = degap_seq(gapped_seq, gap_symbol) 
    ungapped_pairs = delete_gaps_from_pairs(gapped_pairs, gap_list)
    return ungapped_seq, ungapped_pairs

def ungapped_to_gapped(gapped_seq, ungapped_pairs):
    """Returns gapped sequence (same obj) and corresponding Pairs object
    
    gapped_seq: string of characters (can handle Sequence, ModelSequence,
        str, or old_cogent Sequence objects).
    ungapped_pairs: Pairs object, e.g. [(3,7),(4,6)]. The Pairs object should
        correspond to the ungapped sequence version.

    The gap_symbol will be extracted from the sequence object. In case
    the gapped_seq is a simple str, a '-' will be used as default.
    """
    gap_symbol = get_gap_symbol(gapped_seq)
    gap_list = get_gap_list(gapped_seq, gap_symbol)
    gapped_pairs = insert_gaps_in_pairs(ungapped_pairs, gap_list)
    return gapped_seq, gapped_pairs


def insert_gaps_in_seq(sequence, gap_list, gap_symbol='-'):
    
    new_seq = []
    seq_idx = 0
    for idx in range(len(sequence)+len(gap_list)):
        if idx in gap_list:
            new_seq.append(gap_symbol)
        else:
            new_seq.append(sequence[seq_idx])
            seq_idx += 1
    return ''.join(new_seq)

# ==================================================================
# Distance/similarity measures and logical operations
# Pairs comparisons
# ==================================================================

def pairs_intersection(one, other):
    """Returns Pairs object with pairs common to one and other

    one: list of tuples or Pairs object
    other: list of tuples or Pairs object

    one and other should map onto a sequence of the same length.
    """
    pairs1 = frozenset(Pairs(one).directed()) #removes duplicates
    pairs2 = frozenset(Pairs(other).directed())
    return Pairs(pairs1&pairs2)

def pairs_union(one, other):
    """Returns the intersection of one and other

    one: list of tuples or Pairs object
    other: list of tuples or Pairs object

    one and other should map onto a sequence of the same length.
    """
    pairs1 = frozenset(Pairs(one).directed()) #removes duplicates
    pairs2 = frozenset(Pairs(other).directed())
    return Pairs(pairs1 | pairs2)

def compare_pairs(one, other):
    """Returns size of intersection divided by size of union between two Pairs

    Use as a similarity measure for comparing secondary structures.
    Returns the number of base pairs common to both structures divided by
    the number of base pairs that is in one or the other structure:
    (A AND B)/(A OR B) (intersection/union)

    one: list of tuples or Pairs object
    other: list of tuples or Pairs object
    """
    if one.hasConflicts() or other.hasConflicts():
        raise ValueError("Can't handle conflicts in the structure""")
    if not one and not other:
        return 1.0
    pairs1 = frozenset(Pairs(one).directed()) #removes duplicates
    pairs2 = frozenset(Pairs(other).directed())
    return len(pairs1 & pairs2)/len(pairs1|pairs2)

def compare_random_to_correct(one, other):
    """Returns fraction of bp in one that is in other (correct)
    
    one: list of tuples or Pairs object
    other: list of tuples or Pairs object

    Note: the second structure is the one compared against (the correct
        structure)
    """
    if not one and not other:
        return 1.0
    if not one or not other:
        return 0.0
    pairs1 = frozenset(Pairs(one).directed()) #removes duplicates
    pairs2 = frozenset(Pairs(other).directed())
    return len(pairs1 & pairs2)/len(pairs1)

def compare_pairs_mapping(one, other, one_to_other):
    """Returns intersection/union given a mapping from the first pairs to second

    Use in case the numbering of the two Pairs object don't correspond.
    Sort of aligning two ungapped sequences and comparing their Pairs
    object via a mapping. 
    
    one: list of tuples or Pairs object
    other: list of tuples or Pairs object
    one_to_other: mapping of positions in first pairs object to positions
        in second pairs object.
    For example:
    # pos in first seq, base, pos in second seq
    #1 U 0
    #2 C 1
    #3 G 2
    #4 A 3
    #  A 4
    #5 C 5
    #6 C 6
    #7 U 
    #8 G 7

    mapping = {1:0, 2:1, 3:2, 4:3, 5:5, 6:6, 7:None, 8:7}
    """
    if not one and not other:
        return 1.0
    just_in_first = 0
    just_in_second = 0
    in_both = 0
    pairs1 = Pairs(one).directed() #removes duplicates
    pairs2 = Pairs(other).directed()
    for x,y in pairs1:
        other_match = (one_to_other[x],one_to_other[y])
        if other_match in pairs2:
            in_both += 1
            pairs2.remove(other_match)
        else:
            just_in_first += 1
    just_in_second += len(pairs2)

    return in_both/(just_in_first + in_both + just_in_second)

def accuracy(ref, predicted):
    """Returns accuracy of a structure prediction wrt reference

    ref, predicted -- Pairs object or something that can be turned into one

    returns dictionary with following keys (& values)
    REF -- number of base pairs in the reference structure
    PRED -- number of base pairs in the predicted structure
    TP -- number of correctly predicted base pairs
    FN -- number of base pairs in reference and not in predicted
    FP -- number of base pairs in predicted and not in reference
    SEN -- tp/(tp+fn)
    PPV -- tp/(tp+fp)
    CC -- sqrt(sen*ppv)
    SD -- symmetric difference between ref and pred union((A-B)(B-A))
    OVERLAP -- tp/union(ref,pred)
    DIFF -- sd/union(ref,pred)
    """
    
    result = {}

    ref_pairs = Pairs(ref)
    pred_pairs = Pairs(predicted)

    if ref_pairs.hasConflicts() or pred_pairs.hasConflicts():
        raise ValueError("Can't handle conflicts in the structure""")
    
    if not ref_pairs and not pred_pairs:
        raise ValueError("Empty prediction and reference")

    ref_set = frozenset(ref_pairs.directed())
    pred_set = frozenset(pred_pairs.directed())
    
    result['REF'] = len(ref_set)
    result['PRED'] = len(pred_set)

    tp_pairs = ref_set.intersection(pred_set)
    fn_pairs = ref_set.difference(pred_set)
    fp_pairs = pred_set.difference(ref_set)
    union_pairs = ref_set.union(pred_set)

    tp = len(tp_pairs)
    fn = len(fn_pairs)
    fp = len(fp_pairs)
    union = len(union_pairs)
    
    result['TP'] = tp
    result['FN'] = fn
    result['FP'] = fp

    assert tp + fp == len(pred_set)
    assert tp + fn == len(ref_set)

    if not ref_set:
        sen = 0.0
    else:
        sen = tp/(tp+fn)
    if not pred_set:
        ppv = 0.0
    else:
        ppv = tp/(tp+fp)
    cc = sqrt(sen*ppv)
    result['SEN'] = sen
    result['PPV'] = ppv
    result['CC'] = cc

    sd = fp+fn
    frac_overlap = tp/union
    frac_diff = sd/union
    result['SD'] = sd
    result['OVERLAP'] = frac_overlap
    result['DIFF'] = frac_diff

    return result

def bp_fractions(structures):
    """Calculates frequency of occurrence of each base pair in the structures

    structures -- list of Pairs objects or list of objects that can be 
        turned into a pairs object.

    Returns a dictionary of base pair frequencies. This dictionary is 
        can be seen as a base pair profile.
    Input structures are made into Pairs objects and the directed version
        is used (all pairs (up, down), and no unpaired bases or duplicates).
        The method doesn't check for bp conflicts (A,B),(A,C).
    """
    num = len(structures)
    result = {}
    for struct in structures:
        for pair in Pairs(struct).directed():
            if pair not in result:
                result[pair] = 0
            result[pair] += 1
    for k in result:
        result[k] /= num
    return result

def bp_consensus(bp_fracs, cutoff=0.5):
    """Returns a consensus of bps that occur in more than cutoff in bp_fracs

    bp_fracs -- dict of {(up,down): frequency} frequency between 0 and 1

    Sorts the base pairs from high to low occurrence, returns list of 
        non-conflicting base pairs that occur with at least the cutoff
        frequency in the list.
    This method will not exclude pseudoknots. It will exclude conflicting
        base pairs (so the returned structure will be conflict-free, but 
        not necessarily pseudoknot-free).
    """
    freq_and_bp = [(v,k) for k,v in bp_fracs.items()]
    freq_and_bp.sort(reverse=True) # from high to low frequency
    result = Pairs()
    seen = {}
    for frac, (up, down) in freq_and_bp:
        if frac >= cutoff and up not in seen and down not in seen:
            result.append((up, down))
            seen[up] = True
            seen[down] = True
    result = result.directed()
    result.sort()
    return result


# ===========================================================
# Gardner & Giegerich 2004 metrics
# ===========================================================
ACCEPTED = dict.fromkeys(map(tuple,["GC","CG","AU","UA","GU","UG"]))

def check_structures(ref, predicted):
    """Raise ValueError if one of the two structures contains conflicts"""
    if ref.hasConflicts():
        raise ValueError("Reference structure contains conflicts")
    if predicted.hasConflicts():
        raise ValueError("Predicted structure contains conflicts")

def get_all_pairs(sequences, min_dist=4):
    """Return number of possible base pairs in the sequece
    
    sequences: list of Sequence objects or strings
    min_dist: integer, minimum distance between two members of a 
        base pair. Default is 4 (i.e. minimum of 3 unpaired bases in a loop)

    The number of pairs is defined as all possible GC, AU, and GU pairs, 
        respecting the minimum distance between the two members of a
        base pair.
    This method returns the average number of possible base pairs over
        all provided sequences.
    """
    if min_dist < 1:
        raise ValueError("Minimum distance should be >= 1")
    if not sequences:
        return 0.0
    tn_counts = []
    for seq in sequences:
        seq_str = str(seq).upper()
        seq_count = 0
        #print 'xrange', range(len(seq)-min_dist)
        for x in range(len(seq)-min_dist):
            for y in range(x+min_dist,len(seq)):
                if (seq_str[x],seq_str[y]) in ACCEPTED:
                    #print x,y, seq_str[x], seq_str[y], 'Y'
                    seq_count += 1
                else:
                    pass
                    #print x,y, seq_str[x], seq_str[y], 'N'
        tn_counts.append(seq_count)
    return sum(tn_counts)/len(tn_counts)

def get_counts(ref, predicted, split_fp=False, sequences=None, min_dist=4):
    """Return TP, TN, FPcont, FPconf FPcomp, FN counts"""

    result = dict.fromkeys(['TP','TN','FN','FP',\
        'FP_INCONS','FP_CONTRA','FP_COMP'],0)

    ref_set = frozenset(Pairs(ref).directed())
    pred_set = frozenset(Pairs(predicted).directed())

    ref_dict = dict(ref.symmetric())
    pred_dict = dict(predicted.symmetric())

    tp_pairs = ref_set.intersection(pred_set)
    fn_pairs = ref_set.difference(pred_set)
    fp_pairs = pred_set.difference(ref_set)
    result['TP'] = len(tp_pairs)
    result['FN'] = len(fn_pairs)
    result['FP'] = len(fp_pairs)
    if split_fp:
        fp_incons = []
        fp_contra = []
        fp_comp = []
        for x,y in fp_pairs:
            if x in ref_dict or y in ref_dict:
                #print "Conflicting: %d - %d"%(x,y)
                fp_incons.append((x,y))
            else:
                five_prime = x
                three_prime = y
                contr_found = False
                for idx in range(x,y+1):
                    if idx in ref_dict and\
                        (ref_dict[idx] < five_prime or\
                            ref_dict[idx] > three_prime):
                        #print "Contradicting: %d - %d"%(x,y)
                        contr_found = True
                        fp_contra.append((x,y))
                        break
                if not contr_found:
                    #print "Comatible: %d - %d"%(x,y)
                    fp_comp.append((x,y))


        result['FP_INCONS'] = len(fp_incons)
        result['FP_CONTRA'] = len(fp_contra)
        result['FP_COMP'] = len(fp_comp)
        assert result['FP_INCONS'] + result['FP_CONTRA'] + result['FP_COMP'] ==\
            result['FP']
    if sequences:
        num_possible_pairs = get_all_pairs(sequences, min_dist)
        result['TN'] = num_possible_pairs - result['TP'] -\
            result['FP_INCONS'] - result['FP_CONTRA']

    return result

def extract_seqs(seqs):
    """Return list of sequences as strings.

    seqs could either be:
    -- a long string in fasta format:
    ">seq1\nACGUAGC\n>seq2\nGGUAGCG"
    -- a list of lines in fasta format:
    [">seq1","ACGUAGC",">seq2","GGUAGCG"]
    -- a list of sequences (strings or objects):
    ['ACGUAGC','GGUAGCG']
    """
    if isinstance(seqs, str): #assume fasta string
        result = [v for (l,v) in list(MinimalFastaParser(seqs.split('\n')))]
    elif isinstance(seqs, list):
        seq_strings = map(strip,map(str, seqs))
        if seq_strings[0].startswith('>'): #list of fasta lines
            result = [v for l,v in list(MinimalFastaParser(seq_strings))]
        else:
            result = seq_strings
    else:
        raise Exception
    result = [s.replace('T','U') for s in result]
    return result

def sensitivity_formula(counts):
    """Return sensitivity
    
    counts: dict of counts, containing at least TP and FN
    """
    tp = counts['TP']
    fn = counts['FN']

    if not tp and not fn:
        return 0.0
    
    sensitivity = tp/(tp + fn)
    return sensitivity

def selectivity_formula(counts):
    """Return selectivity
    
    counts: dict of counts, containing at least TP, FP, and FP_COMP
    """
    tp = counts['TP']
    fp = counts['FP']
    fp_comp = counts['FP_COMP']
    
    if not tp and fp==fp_comp:
        return 0.0

    selectivity = tp/(tp + (fp - fp_comp))
    return selectivity

def ac_formula(counts):
    """Return approximate correlation
    
    counts: dict of counts, containing at least TP, FP, and FP_COMP
    """
    sens = sensitivity_formula(counts)
    sel = selectivity_formula(counts)
    return (sens+sel)/2

def cc_formula(counts):
    """Return correlation coefficient
    
    counts: dict of counts, containing at least TP, TN, FN, FP, and FP_COMP
    """
    tp = counts['TP']
    tn = counts['TN']
    fp = counts['FP']
    fn = counts['FN']
    comp = counts['FP_COMP']
    
    sens = sensitivity_formula(counts)
    sel = selectivity_formula(counts)

    N = tp+ (fp-comp) + fn + tn
    cc = 0.0
    if tp >0:
        cc = (N*sens*sel-tp)/sqrt((N*sens-tp)*(N*sel-tp))
    return cc

def mcc_formula(counts):
    """Return correlation coefficient
    
    counts: dict of counts, containing at least TP, TN, FN, FP, and FP_COMP
    """
    tp = counts['TP']
    tn = counts['TN']
    fp = counts['FP']
    fn = counts['FN']
    comp = counts['FP_COMP']
    
    mcc_quotient = (tp+fp-comp)*(tp+fn)*(tn+fp-comp)*(tn+fn)
    if mcc_quotient > 0:
        mcc = (tp*tn-(fp-comp)*fn)/sqrt(mcc_quotient)
    else:
        raise ValueError("mcc_quotient <= 0: %.2f"%(mcc_quotient))

    return mcc

def sensitivity(ref, predicted):
    """Return sensitivity of the predicted structure

    ref: Pairs object -> reference structure (true structure)
    predicted: Pairs object -> predicted structure

    Formula: sensitivity = tp/(tp + fn)
    tp = True positives
    fn = False negatives
    """
    check_structures(ref, predicted)
    if not ref and not predicted:
        return 1.0
    elif not predicted:
        return 0.0

    counts = get_counts(ref, predicted)
    return sensitivity_formula(counts)

def selectivity(ref,predicted):
    """Return selectivity of the predicted structure
    
    ref: Pairs object -> reference structure (true structure)
    predicted: Pairs object -> predicted structure

    Formula: selectivity = tp/(tp+fp-fp_comp)
    tp = True positives
    fp = False positives
    fp_comp = compatible fp pairs
    """
    check_structures(ref, predicted)
    if not ref and not predicted:
        return 1.0
    elif not predicted:
        return 0.0

    counts = get_counts(ref, predicted, split_fp=True)
    return selectivity_formula(counts)
    
def selectivity_simple(ref, predicted):
    """Return selectivity without subtracting compatible false positives
    
    ref: Pairs object -> reference structure (true structure)
    predicted: Pairs object -> predicted structure

    Formula: selectivity = tp/(tp+fp)
    tp = True positives
    fp = False positives

    Not considering compatible false positives.
    As implemented in Dowell 2004
    """
    check_structures(ref, predicted)
    if not ref and not predicted:
        return 1.0
    elif not predicted:
        return 0.0

    counts = get_counts(ref, predicted)
    tp = counts['TP']
    fp = counts['FP']
    
    if not tp:  #and fp==fp_comp:
        return 0.0

    selectivity = tp/(tp + fp)
    return selectivity

def approximate_correlation(ref, predicted, seqs):
    """Return the approximate correlation between sensitivity and selectivity

    ref: Pairs object -> reference structure (true structure)
    predicted: Pairs object -> predicted structure

    For the specific case of RNA structure comparisons, Matthews
        correlation coefficient can be approximated by the arithmetic-mean
        or geometrix-mean of sensitivity and selectivity
    
    Formula: ac = (sensitivity+selectivity)/2
    """
    check_structures(ref, predicted)
    counts = get_counts(ref, predicted, split_fp=True)
    return ac_formula(counts)

def correlation_coefficient(ref, predicted, seqs, min_dist=4):
    """Return correlation coefficient to relate sensitivity and selectivity

    Implementation copied from compare_ct.pm
    Always same result as MCC?
    """
    check_structures(ref, predicted)
    sequences = extract_seqs(seqs)
    counts = get_counts(ref, predicted, sequences=sequences, split_fp=True,\
        min_dist=min_dist)
    return cc_formula(counts)
    
def mcc(ref, predicted, seqs, min_dist=4):
    """Return the Matthews correlation coefficient

    ref: Pairs object -> reference structure (true structure)
    predicted: Pairs object -> predicted structure
    seqs: list of sequences, necessary to compute the number of true
        negatives. See documentation of extract_seqs function for 
        accepted formats.
    min_dist: minimum distance required between two members of a base pair.
        Needed to calculate the number of true negatives.
    """
    check_structures(ref, predicted)
    if not ref and not predicted:
        return 1.0
    elif not predicted:
        return 0.0
    elif not seqs:
        raise ValueError, 'No sequence provided!'

    sequences = extract_seqs(seqs)
    counts = get_counts(ref, predicted, sequences=sequences, split_fp=True,\
        min_dist=min_dist)
    return mcc_formula(counts)

def all_metrics(ref, predicted, seqs, min_dist=4):
    """Return dictionary containing the values of five metrics

    ref: Pairs object -> reference structure (true structure)
    predicted: Pairs object -> predicted structure
    seqs: list of sequences, necessary to compute the number of true
        negatives. See documentation of extract_seqs function for 
        accepted formats.
    min_dist: minimum distance required between two members of a base pair.
        Needed to calculate the number of true negatives.

    the metrics returned are:
        sensitivity, selectivity, approximate correlation,
        correlation coefficient, and Matthews correlation coefficient
    """
    check_structures(ref, predicted)
    
    result = {}
    if not ref and not predicted: # set all to 1.0
        for i in ['SENSITIVITY','SELECTIVITY','AC','CC','MCC']:
            result[i] = 1.0
        return result
    elif not predicted: # set all to 0.0
        for i in ['SENSITIVITY','SELECTIVITY','AC','CC','MCC']:
            result[i] = 0.0
        return result
    elif not seqs:
        raise ValueError, 'No sequence provided!'

    sequences = extract_seqs(seqs)
    counts = get_counts(ref, predicted, sequences=sequences, split_fp=True,\
        min_dist=min_dist)

    result['SENSITIVITY'] = sensitivity_formula(counts)
    result['SELECTIVITY'] = selectivity_formula(counts)
    result['AC'] = ac_formula(counts)
    result['CC'] = cc_formula(counts)
    result['MCC'] = mcc_formula(counts)
    return result

# =============================================================================
# Other convenience functions
# =============================================================================

def vienna_mask(vienna, mask):
    """Return Pairs object from a dot-bracket structure and pairing mask.

    vienna -- ViennaStructure object or string that contains dots and
        brackets, e.g. '..(((....)).)..'
    mask -- string, same length as vienna, indicating which brackets
        pair with each other. e.g. '.11..222...11..222.' for the
        pseudoknotted structure '.((..(((...))..))).'
    """
    if len(vienna) != len(mask):
        raise ValueError("Lengths of structure and mask don't match")
    result = Pairs()
    queue = {}
    for idx, (struct, mask_symbol) in enumerate(zip(vienna,mask)):
        if struct == '.':
            if mask_symbol != '.':
                raise ValueError(\
                    "Structure has '.' at pos %s, mask hasn't"%(idx))
            else:
                continue
        elif struct == '(':
            if mask_symbol == '.':
                raise ValueError(\
                    "Mask has '.' at pos %s, structure hasn't"%(idx))
            else:
                if mask_symbol not in queue:
                    queue[mask_symbol] = []
                queue[mask_symbol].append(idx)
        elif struct == ')':
            if mask_symbol == '.':
                raise ValueError(\
                    "Mask has '.' at pos %s, structure hasn't"%(idx))
            else:
                if mask_symbol not in queue:
                    raise ValueError(\
                    "Closing something that hasn't been openend")
                else:
                    try:
                        result.append((queue[mask_symbol].pop(), idx))
                    except IndexError:
                        raise IndexError(\
                            "pop from empty list: mask %s"%(mask_symbol))
        else:
            raise ValueError("Unknown symbol found in structure: %s"%(struct))
    for k, v in queue.items():
        if v:
            raise ValueError("Unpaired positions left: %s"%(str(v)))
    result.sort()
    return result

MASK_SYMBOLS = list(\
        "1234567890ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz")

def _get_temp_symbol(partial_vienna, partial_mask):
    """Return symbol for which # opening and # closing brackets is the same"""
    temp_sym = None
    counts = {}
    for bracket, m_sym in zip(partial_vienna, partial_mask):
        if m_sym == '.':
            continue
        if m_sym not in counts:
            counts[m_sym] = [0,0]
        if bracket == '(':
            counts[m_sym][0] += 1
        if bracket == ')':
            counts[m_sym][1] += 1
    for s, v in counts.items():
        if v[0] == v[1]:
            temp_sym = s
            break
    return temp_sym

def pairs_to_mask(pairs, seq_len, recycle=False):
    """Returns vienna structure and pairing mask (as strings) from Pairs object.

    pairs -- Pairs object
    seq_len -- int, length of sequence that pairs corresponds to
    recycle -- boolean, indicating whether mask symbols should be recycled
        when possible.

    Without recycling a new mask symbol is picked whenever the current pair
    doesn't fit within the previous pair, e.g. previous = (1,10), current =
    (12,14). When recycle is True, a mask symbol that has already been used
    will be reused when possible. Recycling is possible if for a symbol
    the number of opening and closing brackets is the same in the part of
    the structure before the current pair under consideration.

    Recycling can be useful for long structures with many different helices.
    Pairing masks created without recycling might be more human friendly
    (easier to read).
    """
    if pairs.hasConflicts():
        raise ValueError("List of base pairs contains conflicts")
    pairs = pairs.directed()
    pairs.sort()
    lmask = ['.']*seq_len # new mask in list form
    lvienna = ['.']*seq_len # new vienna string in list form
    poss = list(\
        "1234567890ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz") 
    poss.reverse() # all possible mask symbols
    sym = poss.pop() # take the first mask symbol
    last = None # previous base pair
    for x,y in pairs:
        lvienna[x] = '('
        lvienna[y] = ')'
        if last and not (x > last[0] and y < last[1]):
            temp_sym = None
            if recycle:
                temp_sym = _get_temp_symbol(lvienna[:x], lmask[:x])
            if temp_sym is not None:
                sym = temp_sym
            else:
                try:
                    sym = poss.pop()
                except IndexError:
                    raise ValueError("No symbols left")
        lmask[x] = sym
        lmask[y] = sym
        last = (x,y)
    mask = ''.join(lmask)
    vienna = ''.join(lvienna)
    return vienna, mask


def pairs_to_mask_regions(pairs, seq_len, recycle=False):

    if pairs.hasConflicts():
        raise ValueError("List of base pairs contains conflicts")
    lmask = ['.']*seq_len # new mask in list form
    lvienna = ['.']*seq_len # new vienna string in list form
    poss = list(\
        "1234567890ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz") 
    poss.reverse()

    prs = PairedRegionsFromPairs(pairs)
    symbols = {} # assignment of symbols to each region
    
    id_to_pr = prs.byId()
    #print id_to_pr.keys()
    cm = ConflictMatrix(prs)
    nc = cm.nonConflicting()
    if nc:
        curr_sym = poss.pop()
        for pr_id in cm.nonConflicting():
            symbols[pr_id] = curr_sym
            prs.remove(id_to_pr[pr_id])
    #print 'symbols', symbols
    for cl_ids in cm.conflictCliques():
        print 'CLIQUE', cl_ids

        cl = PairedRegions([id_to_pr[i] for i in cl_ids])
        cl_poss = copy(poss)
        print cl_poss
        cl_to_pr = cl.byId()
        cm = ConflictMatrix(cl)
        if len(cl) <= len(cl_poss):
            #print 'length short enough'
            for pr_id in cl_ids:
                symbols[pr_id] = cl_poss.pop()
        else:
            #print 'max conflicts'
            while cl:
                cl_to_pr = cl.byId()
                cl_cm = ConflictMatrix(cl)
                #print cl_cm.Matrix
                mc_id = find_max_conflicts(cl_to_pr.keys(), cl_cm, cl_to_pr)
                #print 'mc_id',  mc_id
                symbols[mc_id] = cl_poss.pop()
                #print 'used for %s: %s'%(mc_id, symbols[mc_id])
                cl.remove(cl_to_pr[mc_id])
                cl_cm = ConflictMatrix(cl)
                if cl_cm.nonConflicting():
                    curr_sym = cl_poss.pop()
                    for pr_id in cl_cm.nonConflicting():
                        #print 'pr', pr_id
                        symbols[pr_id] = curr_sym
                        #print 'used for %s: %s'%(pr_id, symbols[pr_id])
                        cl.remove(cl_to_pr[pr_id])
    #print symbols
    for pr_id, sym in symbols.items():
        #print pr_id, sym
        for x,y in id_to_pr[pr_id].Pairs:
            #print x,y
            lmask[x] = sym
            lmask[y] = sym
            lvienna[x] = '('
            lvienna[y] = ')'

    return ''.join(lvienna), ''.join(lmask)

    
def paired_blocks(pairs):
    """Return stretches of nested base pairs as list of lists

    pairs = Pairs object, no conflicts allowed

    Result is a list of stretches of perfectly nested base pairs, which means
    [[(n,m),(n+1, m-1),(n+2,m-2), ...],[(i,j),(i+1,j-1),...]]
    Base pairs are sorted before the stretches are picked out, so the result
    will be in order.
    """
    p = Pairs(pairs)
    if p.hasConflicts():
        raise ValueError("paired_blocks can't handle base pair conflicts")
    clean_pairs = p.directed()
    clean_pairs.sort()
    blocks = []
    curr_block = []
    for pair in clean_pairs:
        if not curr_block:
            curr_block.append(pair)
        else:
            x,y = curr_block[-1]
            if pair == (x+1, y-1):
                curr_block.append(pair)
            else:
                blocks.append(curr_block)
                curr_block = [pair]
    if curr_block: # last block
        blocks.append(curr_block)
    
    return blocks

def get_temp2(bl, used):
    order = "1234567890ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz"
    for sym in order:
        if sym not in used:
            return None
        else:
            available = sym
            for used_bl in used[sym]:
                if used_bl[0][1] < bl[0][0] or\
                    (used_bl[-1][0] < bl[0][0] and bl[0][1] < used_bl[-1][1]):
                    pass
                else:
                    available = None
                    break
            #for used_bl in used[sym]:
            #    if used_bl[0][1] > bl[0][0] or used_bl[-1]:
            #        available = None
            if available:
                return available
        return None

def pairs_to_mask2(pairs, seq_len, recycle=False):

    if pairs.hasConflicts():
        raise ValueError("List of base pairs contains conflicts")
    p = pairs.directed()
    p.sort()
    lmask = ['.']*seq_len # new mask in list form
    lvienna = ['.']*seq_len # new vienna string in list form

    poss = list(\
        "1234567890ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz")
    poss.reverse() # all possible mask symbols
    sym = poss.pop() # take the first mask symbol
    used = {}
    last = None # previous block
    for bl in paired_blocks(p):
        if not last:
            pass
        else:
            #figure out symbol
            #print bl
            x, y = bl[0]
            if last and (x > last[-1][0] and y < last[-1][1]):
                pass
            elif last and not (x > last[-1][0] and y < last[-1][1]):
                temp_sym = None
                if recycle:
                    temp_sym = get_temp2(bl, used)
                if temp_sym is None:
                    try: 
                        sym = poss.pop()
                    except IndexError:
                        raise ValueError("No symbols left")
        for i,j in bl:
            lvienna[i] = '('
            lvienna[j] = ')'
            lmask[i] = sym
            lmask[j] = sym
            last = bl
            if sym not in used:
                used[sym] = []
            used[sym].append((bl))
    mask = ''.join(lmask)
    vienna = ''.join(lvienna)
    #print last
    #print used
    return vienna, mask

def left_right(list_to_left, list_to_right):
    left = None
    for s in list_to_left:
        if s != '.':
            left = s
    list_to_right.reverse()
    right = None
    for s in list_to_right:
        if s != '.':
            right = s
    return left, right

def pairs_to_mask3(pairs, seq_len, recycle=False):
    """Every block different"""
    if pairs.hasConflicts():
        raise ValueError("List of base pairs contains conflicts")
    p = pairs.directed()
    p.sort()
    lmask = ['.']*seq_len # new mask in list form
    lvienna = ['.']*seq_len # new vienna string in list form

    poss = list(\
        "1234567890ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz"*10)
    poss.reverse() # all possible mask symbols
    sym = poss.pop() # take the first mask symbol
    used = {}
    last = None # previous block
    for bl in paired_blocks(p):
        try:
            sym = poss.pop()
        except IndexError:
            raise ValueError("No symbols left")
        ok = False
        left = lmask[:bl[0][0]]
        right = lmask[bl[0][1]:]
        lll, rrr = left_right(left, right)
        print lll, rrr
        while ok is False:
            if sym == lll or sym == rrr:
                try:
                    sym = poss.pop()
                except IndexError:
                    raise ValueError("No symbols left")
            else:
                ok = True
        print sym
        for i,j in bl:
            lvienna[i] = '('
            lvienna[j] = ')'
            lmask[i] = sym
            lmask[j] = sym
            last = bl
            if sym not in used:
                used[sym] = []
            used[sym].append((bl))
    mask = ''.join(lmask)
    vienna = ''.join(lvienna)
    #print last
    #print used
    print mask
    return vienna, mask




if __name__ == "__main__":
    pass
