#!/usr/bin/env python
# pdbrna.py

"""
Author: Sandra Smit (Sandra.Smit@colorado.edu)

Revision History:
File created on 27 Dec 2006.

"""
from __future__ import division
import sys
from commands import getoutput
from operator import itemgetter
from numpy import add
from cogent.struct.rna2d import Pairs
from cogent.parse.rnaview import RnaviewParser, is_canonical, BasePairs,\
    RnaViewParseError, in_chain
from cogent.app.rnaview import RnaView
from rnalib.parse.pdb_format import PdbParser, is_insertion_free,\
    fix_insertions

# =============================================================================
# FUNCTIONS TO GET/VERIFY RNAVIEW AND PDB OUTPUT
# =============================================================================

# Special dictionaries for the combination of PDB and RNAView
# As standard residues, RNA UCAG, and DNA AGC are accepted
# RNA T or DNA T should be translated to U
STD_RES = {'RNA': {'A':'A', 'G':'G', 'U':'U', 'C':'C'},
    'DNA': {'A':'A','G':'G','C':'C'},
    'AA': {'ALA':'A','CYS':'C','ASP':'D','GLU':'E', 'PHE':'F', 'GLY':'G',
    'HIS':'H','ILE':'I','LYS':'K','LEU':'L','MET':'M','ASN':'N','PRO':'P',
    'GLN':'Q','ARG':'R','SER':'S','THR':'T','SEC':'U','VAL':'V','TRP':'W',
    'TYR':'Y', 'UNK':'X'}}

GEN_CHANGES = {'RNA': {'T':'U'},\
    'DNA': {'T':'U'}}

def insertion_free_file(pdb_filename):
    """Return file name of insertion free version of original PDB file
    
    pdb_filename -- string, file name of the PDB file
    """
    f = list(open(pdb_filename))
    if not is_insertion_free(f):
        new_filename = pdb_filename.rsplit('.',1)[0] + 'ni' + '.pdb'
        out = open(new_filename,'w')
        out.write(fix_insertions(f))
        out.close()
    else:
        new_filename = pdb_filename
    return new_filename

def standard_out_ok(lines):
    """Return True if all lines in the standard output are expected.

    lines -- list of lines or anything that behaves like it. Standard 
        output of RNAView

    The expected lines are:
    =======================
    Processing a single PDB file

    PDB data file name: 1B23.pdb
    RNA/DNA chain_ID:  R  from residue    1 to   76
    Finding all the base pairs...
      The total base pairs =  28 (from   74 bases);

    Time used: 0.09 seconds

    JOB finished! Time used: 0.09 seconds
    """
    # ignore everthing after "The total base pairs"
    ignore = False
    for line in lines:
        line = line.strip()
        if not line or ignore:
            continue
        if line.startswith('Processing') or line.startswith('PDB data'):
            continue
        elif 'complicated structure' in line or '=> weird' in line:
            continue
        elif line.startswith('The best NMR model') or\
            line.startswith('The best representative'):
            continue
        elif line.startswith('RNA/DNA chain_ID'):
            continue
        elif line.startswith('Finding all'):
            continue
        elif line.startswith('The total base pairs'):
            ignore = True
            continue
        elif line.startswith('Time used') or line.startswith('JOB finished!'):
            continue
        else:
            print "WARNING: %s"%(line)
            return False
    return True

def selected_model(pdb_info):
    """Return the Model object that was used by RNAView.
    
    pdb_info -- output of PdbParser
    """
    # determine used model
    models = pdb_info['MODELS']
    if models is None:
        return None
    if len(models) > 1:
        try:
            model_id = pdb_info['REMARK']['210']['BEST_REPRESENTATIVE']
        except KeyError:
            model_id = '1'
        if model_id is None:
            model_id = '1'
        model = models[model_id]
    else: #take the first and only model
        model = models.values()[0]
    return model

def verify_base(base, pdb_model):
    """Return None if base matches the PDB Model, a string (error) if not"""
    res = pdb_model[base.ChainId][base.ResId] # residue
    if base.ResName in STD_RES:
        if b.ResName != res.ResName:
            return "RNAView Base %s doesn't match PDB residue %s"\
                %(str(base), res.ResName)
    elif base.ResName == 'P' or base.ResName == 'T':
        if res.StandardRes != 'U':
            return "RNAView Base is '%s', PDB residue is '%s'"\
                %(base.ResName, res.StandardRes)
    elif base.ResName == 'I':
        if res.StandardRes != 'A':
            return "RNAView Base is '%s', PDB residue is '%s'"\
                %(base.ResName, res.StandardRes)
    else:
        if base.ResName.upper() != res.StandardRes:
            return "RNAView Base %s doesn't match PDB residue %s"\
                %(str(base), res.StandardRes)
    return None

def rnaview_matches_pdb(rnaview_info, pdb_info):
    """Return True if base identities and number of bases match.
    
    rnaview_info -- output from RnaviewParser
    pdb_info -- output from PdbParser
    """
    model = selected_model(pdb_info)
    if model is None:
        raise RnaViewParseError(\
            "Can't verify base identities if model is not set")
    else:
        all_bases = []
        for bp in rnaview_info['BP']:
            all_bases.append(bp.Up)
            all_bases.append(bp.Down)
        for multi in rnaview_info['BM']:
            all_bases.extend(multi)
        for base in all_bases: 
            v = verify_base(base, model)
            if v is not None:
                raise RnaViewParseError(str(v))
    return True

def mcann_matches_pdb(mcann_info, pdb_info):
    """Return True if base identities and number of bases match.
    
    mcann_info -- output from McAnnotateParser
    pdb_info -- output from PdbParser
    """
    model = selected_model(pdb_info)
    if model is None:
        raise RnaViewParseError(\
            "Can't verify base identities if model is not set")
    else:
        all_bases = []
        for bp in mcann_info:
            all_bases.append(bp.Up)
            all_bases.append(bp.Down)
        for base in all_bases: 
            v = verify_base(base, model)
            if v is not None:
                raise RnaViewParseError(str(v))
    return True


def existing_files(pdb_filename):
    """Return filname of used pdb file and RNAView output
    """
    
    used_pdb = None
    rnaview_output = None
    #if RNAView output file already present, use it
    if pdb_filename.endswith('.pdb'):
        short = pdb_filename[:-4]
    else:
        raise ValueError("pdb_filename should end in '.pdb'")
    cmd = "find %s"%(short+'*'+'.pdb.out')
    matching  = getoutput(cmd).split('\n')
    #print matching
    if matching[0].startswith('find:'): #file not found
        pass
    elif len(matching) == 1 and matching[0]:
        rnaview_output = matching[0]
    else:
        pass
    if rnaview_output is not None:
        used_pdb = rnaview_output[:-4]
    return used_pdb, rnaview_output


def pdbrna(pdb_filename, chains=None, set_standard=True,\
    standard_res=STD_RES, general_changes=GEN_CHANGES, specific_changes=None):
    """Return result of RnaView run, verified against PDB file.

    pdb_filename: filename of PDB file, string
    chains: should be string of chains

    Procedure:
    Open the PDB file, fix insertions if necessary (RnaView doens't handle
        these well).
    Run RnaView, store the output, and clean up.
    Parse the RnaView results with PDB verification.
    Return the parser results.
    """
   
    used_pdb, rnaview_output = existing_files(pdb_filename)

    if used_pdb is None or rnaview_output is None:
        # fix insertions in pdb file, because rnaview doesn't handle
        # them correctly
        used_pdb = insertion_free_file(pdb_filename)

        # run RNAView, check standard out, store relevant output, clean up
        if chains is not None:
            params = {'-c': "\""+chains+"\""}
            #print params['-c']
            #params = {'-c': chains}
            r = RnaView(params=params)
        else:
            r = RnaView()
        #print r.BaseCommand
        res = r(used_pdb)
        std_out = res['StdOut'].readlines()
        if not standard_out_ok(std_out):
            raise ValueError("Standard out not ok")
        bp_file = res['base_pairs'].readlines()
        res.cleanUp()
    else:
        bp_file = open(rnaview_output)

    # parse PDB file, parse RNAView output
    pdb_info = PdbParser(open(used_pdb), strict=True,\
        set_standard=set_standard, standard_res=standard_res,\
        specific_changes=specific_changes, general_changes=general_changes)
    rnaview_info = RnaviewParser(bp_file)
    # verify parsed data
    rnaview_matches_pdb(rnaview_info, pdb_info)
    #print rnaview_info['NP']['NUM_BASES']
    return pdb_info, rnaview_info

def bps_model(pdb_filename, chains=None, set_standard=True,\
    standard_res=STD_RES, general_changes=GEN_CHANGES, specific_changes=None):
    """Return tuple of BasePairs object and Model object.
    """
    pdb_info, rnaview_info =\
        pdbrna(pdb_filename, chains=chains, set_standard=set_standard,\
        standard_res=standard_res, general_changes=general_changes,\
        specific_changes=specific_changes)
    bps = rnaview_info['BP']
    if chains:
        bps = bps.select(in_chain(chains))
    model = selected_model(pdb_info)
    return bps, model

# =============================================================================
# FUNCTIONS FOR COMPOSITION OF SETS OF BASE PAIRS
# =============================================================================

def get_all_interactions(bps, model=None, chains=None, conflicts_to_add=None):
    """Return BasePairs objects containing all interactions"""
    return bps

def get_all_canonical(bps, model=None, chains=None, conflicts_to_add=None):
    """Return BasePairs object containing all canonical interactions.
    
    Canonical are standard interactions or standard GU pairs. Conflicts
    in this list of base pairs are NOT removed.
    """
    return bps.select(is_canonical)

def get_ncc(bps, model=None, chains=None, conflicts_to_add=None):
    """Return non-conflicting canonical base pairs.

    conflicts_to_add -- dictionary containing the IDs of the conflicting
        pairs to add to the non-conflicting base pairs. Format:
        {(up_chain_id, up_res_id, down_chain_id, down_res_id): None}
    """
    all_can = bps.select(is_canonical)
    if not all_can.hasConflicts():
        return all_can
    else:
        if conflicts_to_add is None:
            raise ValueError("No conflict information provided")
        ncc = non_conflicting_pairs(all_can)
        cc = conflicting_pairs(all_can)
        for bp in cc:
            if (bp.Up.ChainId, bp.Up.ResId,bp.Down.ChainId, bp.Down.ResId)\
                in conflicts_to_add:
                ncc.append(bp)
        if ncc.hasConflicts():
            raise ValueError("Panic! Still conflicts present")
        return ncc


# =============================================================================
# CONVENIENCE FUNCTIONS
# =============================================================================

def get_seq_info(pdb_model, requested_chains):
    """Return sequence information as list of residues.

    pdb_model -- Model object
    requensted_chains -- list of Chain IDs, e.g. "AB"

    Returns list of information for each residue in the requested chains:
    [[ChainId, ResId, ResName, StandardRes, ResType, SeqPos]]
    
    Sequence positions are adjusted to make all the chains into one sequence.
    """
    for c in requested_chains:
        if c not in pdb_model:
            raise ValueError("Chain %s doesn't exist in Model"%(c))
    seq_info = pdb_model.seq_info(requested_chains)

    # adjust overall numbering
    seq_info_full = []
    lengths = add.accumulate([0]+[len(pdb_model[chain]) for chain in\
        requested_chains])
    for info in seq_info:
        seq_info_full.append(info[:5] +\
            [info[5] +\
            lengths[requested_chains.index(info[0])]])
    return seq_info_full

def last_rna_dna(seq_info):
    """Return per chain the seq pos of the last dna/rna residue
    
    seq_info -- list of residue information. 
        [[ChainId, ResId, ResName, StandardRes, ResType, SeqPos]]
        All strings except for SeqPos (int).

    Return value is a dict of {ChainId: SeqPos of last RNA/DNA residue}
    """
    last_rna = {}
    for info in seq_info:
        chain = info[0]
        if chain not in last_rna:
            last_rna[chain] = -1
        if (info[4] == 'RNA' or info[4] == 'DNA') and\
            info[5] > last_rna[chain]:
            last_rna[chain] = info[5]
    return last_rna

def rna_dna_only(seq_info):
    """Return a list of only RNA/DNA residues with renewed numbering.
    
    seq_info -- list of residue information
    """
    
    last_rna = last_rna_dna(seq_info)
    good_pos = []
    idx = 0
    for info in seq_info:
        #print info
        chain = info[0]
        if (info[4] == 'SA' or info[4] == 'AA'):
            #if len(info) == 7: #paired
            #    raise ValueError("paired SA or AA found")
            if info[5] < last_rna[chain]:
                raise ValueError("SA or AA in middle of chain")
            else:
                continue
        if info[4] is None:
            #if len(info) == 7:
            #    raise ValueError("Paired residue with unknown ResType")
            if info[5] < last_rna[chain]:
                raise ValueError("Corrupt residue in middle of chain")
            else:
                continue
        if (info[4] == 'RNA' or info[4] == 'DNA'):
            if info[3] is not None and len(info[3]) == 1:
                info[5] = idx
                good_pos.append(info)
                idx += 1
            else:
                raise ValueError("Something is wrong with this residue")
    return good_pos

def get_seq(seq_info):
    """Return sequence of StandardRes.

    Assumes only RNA/DNA residues with U, C, A, or G as Standard Residue.
    WARNING: this function doesn't check the above assumption.
    """
    seq = ''.join([i[3] for i in seq_info])
    return seq

def get_pairs(bps, seq_info):
    """Return Pairs object corresponding to seq_info.

    bps -- BasePairs object
    seq_info --

    Only includes base pairs for which both upstream and downstream base
    are in the requested chains (in seq_info).
    """
    lookup = dict([((i[0],i[1]),i[5]) for i in seq_info])
    chains = dict.fromkeys([i[0] for i in seq_info])
    pairs = Pairs()
    for i in bps:
        if i.Up.ChainId in chains and i.Down.ChainId in chains:
            try:
                up_info = (i.Up.ChainId, i.Up.ResId)
                down_info = (i.Down.ChainId, i.Down.ResId)
                pairs.append((lookup[up_info],lookup[down_info]))
            except KeyError:
                raise ValueError("Residue not in list")
        else:
            continue
    pairs = pairs.directed()
    pairs.sort()
    return pairs

def rna_dna_residues(by_res_id):
    # check & clean all info
    last_rna = {}
    for info in by_res_id.values():
        chain = info[0]
        if chain not in last_rna:
            last_rna[chain] = -1
        if (info[4] == 'RNA' or info[4] == 'DNA') and\
            info[5] > last_rna[chain]:
            last_rna[chain] = info[5]

    good_pos = []
    for info in by_res_id.values():
        chain = info[0]
        if (info[4] == 'SA' or info[4] == 'AA'):
            if len(info) == 7: #paired
                raise ValueError("paired SA or AA found")
            if info[5] < last_rna[chain]:
                raise ValueError("SA or AA in middle of chain")
            else:
                continue
        if info[4] is None:
            if len(info) == 7:
                raise ValueError("Paired residue with unknown ResType")
            if info[5] < last_rna[chain]:
                raise ValueError("Corrupt residue in middle of chain")
            else:
                continue
        if (info[4] == 'RNA' or info[4] == 'DNA'):
            if info[3] is not None and len(info[3]) == 1:
                good_pos.append(info)
            else:
                raise ValueError("Something is wrong with this residue")
    return good_pos


def seq_pairs(bps, pdb_model, requested_chains=None):
    """Return tuple of RnaSequence and Pairs object
    
    bps -- BasePairs object
    pdb_model -- Model object (model that is used by RNAView)
    requested_chains -- list of chain IDs, e.g. "AD" or ['A',' ','C']
    """
    # it's only useful to execute this if there are no bp conflicts
    has_conflict, conflict = bps.hasConflicts(return_conflict=True)
    if has_conflict:
        raise ValueError(\
            "Conflict in list of base pairs (%s), can't return structure"\
            %(conflict))
    
    #get seq info
    if pdb_model is None:
        raise RnaViewObjectError(\
            "Need PDB Model to extract sequence information")
    if requested_chains is None:
        requested_chains = bps.PresentChains
    seq_info = get_seq_info(pdb_model, requested_chains)
    good_pos = rna_dna_only(seq_info)
    seq = get_seq(good_pos)
    pairs = get_pairs(bps, good_pos)
    return seq, pairs

def bps_by_base(bps):
    """Return dict of {(ChainId, ResId):[BasePair]}"""
    seen = {}
    for bp in bps:
        up_info = (bp.Up.ChainId, bp.Up.ResId)
        down_info = (bp.Down.ChainId, bp.Down.ResId)
        if up_info not in seen:
            seen[up_info] = [bp]
        else:
            seen[up_info].append(bp)
        if down_info not in seen:
            seen[down_info] = [bp]
        else:
            seen[down_info].append(bp)
    return seen


def conflicts(bps):
    """Yield list of conflicting base pairs"""

    per_base = bps_by_base(bps)
    for k, v in per_base.items():
        if len(v) > 1:
            yield v

def non_conflicting_pairs(bps):
    """Return BasePairs object of non-conflicting base pairs"""
    per_base = bps_by_base(bps)
    result = BasePairs()
    for bp in bps:
        up_info = (bp.Up.ChainId, bp.Up.ResId)
        down_info = (bp.Down.ChainId, bp.Down.ResId)
        if len(per_base[up_info]) == 1 and len(per_base[down_info]) == 1:
            result.append(bp)
    return result

def conflicting_pairs(bps):
    """Return BasePairs object of conflicting base pairs"""
    per_base = bps_by_base(bps)
    result = BasePairs()
    for bp in bps:
        up_info = (bp.Up.ChainId, bp.Up.ResId)
        down_info = (bp.Down.ChainId, bp.Down.ResId)
        if len(per_base[up_info]) > 1 or len(per_base[down_info]) > 1:
            result.append(bp)
    return result



if __name__ == "__main__":

    from sys import argv
    pdb_filename = argv[1]
    chains = None
    if len(argv) == 3:
        chains = argv[2]
    #pdb_out, rnaview_out = pdbrna(pdb_filename)
    #print rnaview_out['BP']
    bps, model = bps_model(pdb_filename, chains=chains)
    sys.exit()
    sel = bps.select(is_canonical)
    if chains is not None:
        seq_info = get_seq_info(model, chains)
        good = rna_dna_only(seq_info)
        print get_seq(good)
        print get_pairs(sel, good)
    else:
        for cl in sel.cliques():
            print 'Present chains: %s'%(cl.PresentChains)
            #print seq_pairs(cl, model)
            seq_info = get_seq_info(model, cl.PresentChains)
            #for i in seq_info:
            #    print i
            #print last_rna_dna(seq_info)
            good = rna_dna_only(seq_info)
            #for i in good:
            #    print i
            print get_seq(good)
            print get_pairs(cl, good)
            """
            if cl.hasConflicts():
                print "CONFLICT"
                for conflict in conflicts(cl):
                    print map(str,conflict)
                print "CONFLICTING PAIRS"
                for i in conflicting_pairs(cl):
                    print i
            else:
                seq, pairs = seq_pairs(cl, model)
                print len(seq)
                print seq
                print pairs
            """
