#!/usr/bin/env python
"""
Functionality limited to RNA sequences, with symbols in the RNA alphabet
(as specified in the RNA MolType object).

It handles differnt objects:
RnaSequence
ModelSequence
strings
It always uses the '-' as a gap character
"""

from __future__ import division
import pickle
from copy import copy
from random import sample, choice, randint, shuffle
from cogent.struct.rna2d import Vienna, ViennaStructure, Pairs
from cogent.util.misc import flatten
from rnalib.struct.pairs_util import ungapped_to_gapped, paired_blocks
from rnalib.core.classification import Classification
#from rnalib.struct.vienna_suppl import mfe_structs_from_seqs,\
#    subopt_structs_from_seqs
from cogent.parse.record_finder import LineGrouper
from cogent.app.vienna_package import RNAfold, RNAsubopt
from cogent.struct.knots import PairedRegionsFromPairs, PairedRegions,\
    PairedRegion

def is_info_or_empty(line):
    if (not line) or line.isspace():
        return True
    if line.startswith('INFO:'):
        return True
    return False
    
two_at_the_time = LineGrouper(2, ignore=is_info_or_empty)

def just_seqs_parser(lines, struct_constructor=Vienna):
    return [(seq, struct_constructor(struct))\
        for seq, struct in two_at_the_time(lines)]

def mfe_structs_from_seqs(seqs, struct_constructor=Vienna, noLP=False):
    """Returns a list of MFE Vienna structures calculated for the sequences.

    seqs: list of sequences (no labels!), expects sequences as simple strings
    """
    r = RNAfold()
    if noLP is True:
        r.Parameters['-noLP'].on()
    #print r.BaseCommand
    res = r(seqs)
    output = res['StdOut'].readlines()
    structures = [struct for (seq, struct) in\
        just_seqs_parser(output, struct_constructor)]
    res.cleanUp()
    #print structures
    return structures

def subopt_structs_from_seqs(seqs, struct_constructor=Vienna,\
    energy_range=1, noLP=False, sample_ensemble=None):
    """Returns a list of suboptimal structures calculated by RNAsubopt

    seqs: list of sequences (no labels)
    """
    structures = []
    r = RNAsubopt()
    if noLP is True:
        r.Parameters['-noLP'].on()
    if sample_ensemble is not None:
        r.Parameters['-p'].on(sample_ensemble)
    r.Parameters['-e'].on(energy_range)
    #print r.BaseCommand
    res = r(seqs)
    for line in res['StdOut'].readlines():
        try:
            s = Vienna(line.strip())
            #print 'struct found:', line.strip()
            structures.append(s)
        except ValueError:
            #print 'NOT', line.strip()
            pass
    res.cleanUp()
    return structures


def mfe_constrained(data, struct_constructor=Vienna, noLP=False):
    """Return list of ViennaStructures -- mfe for sequences under constraint

    data -- list of tuples [(seq, constraint)]
    """
    r = RNAfold()
    r.Parameters['-C'].on()
    if noLP is True:
        r.Parameters['-noLP'].on()
    input_data = []
    for seq, con in data:
        input_data.append(str(seq))
        input_data.append(str(con))
    res = r(input_data)
    output = res['StdOut'].readlines()
    structures = [struct for (seq, struct) in\
        just_seqs_parser(output, struct_constructor)]
    res.cleanUp()
    return structures

def subopt_constrained(data, struct_constructor=Vienna,\
    energy_range=1, noLP=False, sample_ensemble=None):
    """Returns a list of suboptimal structures calculated by RNAsubopt

    seqs: list of sequences (no labels)
    """
    structures = []
    r = RNAsubopt()
    r.Parameters['-C'].on()
    if noLP is True:
        r.Parameters['-noLP'].on()
    if sample_ensemble is not None:
        r.Parameters['-p'].on(sample_ensemble)
    r.Parameters['-e'].on(energy_range)
    input_data = []
    for seq, con in data:
        input_data.append(str(seq))
        if '|' in con:
            raise ValueError(\
                "Constraints of type '|' not allowed by RNAsubopt")
        input_data.append(str(con))
    #print r.BaseCommand
    res = r(input_data)
    for line in res['StdOut'].readlines():
        try:
            s = Vienna(line.strip())
            #print 'struct found:', line.strip()
            structures.append(s)
        except ValueError:
            #print 'NOT', line.strip()
            pass
    res.cleanUp()
    return structures


# =============================================================================
# MFE STRUCTURES
# =============================================================================

# Should return MFE structures using RNAfold
# Should handle Alignment and DenseAlignment or RnaSequence or ModelSequence
# All DARCS/Cogent objects

def mfe_structs(seqs, sample_size=None, noLP=False, degap=False):
    """Return list of ViennaStructure objects -- MFE for every seq in seqs.

    Calculates the MFE structures for seqs in the alignment

    seqs -- list of sequences, (can handle Sequence, ModelSequence,
        str, or old_cogent Sequence objects)
    sample_size -- number of sequences that should randomly be selected\
        from the aln
    noLP -- boolean, specifies whether lone pairs are allowed (see RNAfold
        documentation)

    Takes the sequences as is, no degapping or anything.
    """
    if not seqs:
        return []
    str_seqs = map(str, seqs) # convert seqs into strings
    if sample_size is not None:
        try:
            seqs = sample(seqs, sample_size)
        except ValueError: #sample larger than population
            seqs = seqs
    
    if degap:
        degapped_seqs = [seq.replace('-','') for seq in str_seqs]
        structures = mfe_structs_from_seqs(degapped_seqs, noLP=noLP)
        if len(degapped_seqs) != len(structures):
            raise ValueError("Lengths should be the same")
        result = []
        for gapped_seq, gapped_struct in zip(str_seqs, structures):
            p = gapped_struct.toPairs()
            new_seq, gapped_p = ungapped_to_gapped(gapped_seq,p)
            gapped_v = gapped_p.toVienna(gapped_seq)
            if len(gapped_v) != len(gapped_seq):
                raise ValueError("Lengths should be the same")
            result.append(gapped_v)
    else:
        result = mfe_structs_from_seqs(str_seqs, noLP=noLP)
    
    return result

def mfe_structs_aln(aln, sample_size=None, noLP=False, degap=False):
    """Return list of ViennaStructure objects -- MFE for every seq in aln.
    
    Calculates the MFE structures for seqs in the alignment
   
    aln -- Alignment object or DenseAlignment object
    degap: boolean, indicating whether sequences should be degapped 
        before the structure is calculated.
    See mfe_struct for more documentation.
    """
    return mfe_structs(aln.values(), sample_size, noLP, degap=degap)



#OLD===========================================================
def mfe_structs_degap(seqs, sample_size=None, noLP=False):
    """Calculates the MFE structures for seqs in the alignment

    select: Number of sequences that should randomly be selected from the aln

    First degaps the sequence, gets the structures and inserts the 
    original gaps in the structures.
    """
    if sample_size is not None:
        try:
            seqs = sample(seqs, sample_size)
        except ValueError: #sample larger than population
            seqs = seqs

    degapped_seqs = [seq.replace('-','') for seq in seqs]
    structures = mfe_structs_from_seqs(degapped_seqs, noLP=noLP)
    assert len(degapped_seqs) == len(structures)
    result = []
    for gapped_seq, gapped_struct in zip(seqs, structures):
        p = gapped_struct.toPairs()
        new_seq, gapped_p = ungapped_to_gapped(gapped_seq,p)
        gapped_v = gapped_p.toVienna(gapped_seq)
        assert len(gapped_v) == len(gapped_seq)
        result.append(gapped_v)
    return result

def mfe_structs_seqsim_aln(aln, sample_size=None, noLP=False, degap=False):
    """Calculates the MFE structures for seqs in the alignment

    degap: boolean, indicating whether sequences should be degapped 
        before the structure is calculated.
    See mfe_struct for more documentation.
    """
    if degap:
        return mfe_structs_degap(seqsim_aln_to_seqs(aln), sample_size, noLP)
    return mfe_structs(seqsim_aln_to_seqs(aln), sample_size, noLP)

def subopt_structs_seqsim_aln(aln, sample_size=None, energy_range=1,\
    noLP=False, degap=False):
    """Calculates the MFE structures for seqs in the alignment

    select: Number of sequences that should randomly be selected from the aln
    """
    if degap:
        return subopt_structs(seqsim_aln_to_seqs(aln), sample_size,\
        energy_range, noLP)
    return subopt_structs(seqsim_aln_to_seqs(aln), sample_size, energy_range,\
        noLP)


# =============================================================================
# SUBOPT STRUCTURES
# =============================================================================

def subopt_structs(seqs, sample_size=None, degap=False, energy_range=1,\
    noLP=False, sample_ensemble=None):
    """Calculates the suboptimal structures within energy range

    seqs: list of sequences
    sample_size: number of sequences that should be selected from the 
        full list.
    energy_range: as defined in RNAsubopt
    noLP: boolean, indicating whether lone pairs are allowed, see RNAsubopt
    """
    if not seqs:
        return []
    if sample_size is not None:
        try:
            seqs = sample(seqs, sample_size)
        except ValueError: #sample larger than population
            seqs = seqs
    str_seqs = map(str, seqs) # convert seqs into strings
   
    if degap:
        structures = []
        for seq in str_seqs:
            degapped = seq.replace('-','')
            subopt = subopt_structs_from_seqs([degapped],\
                energy_range=energy_range, noLP=noLP,\
                sample_ensemble=sample_ensemble)
            for struct in subopt:
                p = struct.toPairs()
                new_seq, gapped_p = ungapped_to_gapped(seq,p)
                gapped_v = gapped_p.toVienna(seq)
                if len(gapped_v) != len(seq):
                    raise ValueError("Lengths should be equal")
                structures.append(gapped_v)
    else:
        structures = subopt_structs_from_seqs(str_seqs,\
            energy_range=energy_range, noLP=noLP,\
            sample_ensemble=sample_ensemble)
    
    return structures

def subopt_structs_aln(aln, sample_size=None, energy_range=1,\
    noLP=False, degap=False, sample_ensemble=None):
    """Calculates the MFE structures for seqs in the alignment

    select: Number of sequences that should randomly be selected from the aln
    """
    return subopt_structs(aln.values(), sample_size=sample_size, degap=degap,\
        energy_range=energy_range, noLP=noLP, sample_ensemble=sample_ensemble)



# =============================================================================
# MANIPULATE EXISTING STRUCTURES -- SINGLE STRUCTURE
# =============================================================================

# ADD STUFF

def add_pair(pairs, seq_len, number):
    """Return Pairs object containing the original pairs plus added pairs

    pairs -- Pairs object
    seq_len -- int, indicating length of the corresponding seq (used to
        determine the maximum position that can be selected)
    number -- number of base pairs to add
    """
    result = copy(pairs)
    for i in range(1,number+1):
        paired = dict.fromkeys(flatten(result))
        unpaired = [pos for pos in range(seq_len) if pos not in paired]
        found = False
        attempts = 0
        while found is False and attempts <= 100:
            attempts += 1
            up, down = choice(unpaired), choice(unpaired) 
            if down < up:
               up, down = down, up
            if down - up < 4:
                continue
            result.append((up, down))
            try:
                v = result.toVienna(seq_len)
                found = True
            except Exception, e:
                assert result[-1] == (up,down)
                result.pop()
                continue
    return result

def add_bp(pairs, seq_len, number=1, max_dist=None):
    """Return Pairs object containing the original pairs plus added pairs

    pairs -- Pairs object
    seq_len -- int, indicating length of the corresponding seq (used to
        determine the maximum position that can be selected)
    number -- number of base pairs to add
    """
    result = copy(pairs)
    for i in range(1,number+1):
        paired = dict.fromkeys(flatten(result))
        unpaired = [pos for pos in range(seq_len) if pos not in paired]
        if len(unpaired) < 2:
            break
        for x in range(100):
            one, two = sample(unpaired,2)
            if one < two:
                up, down = one, two
            else:
                up, down = two, one
            if down-up < 4:
                continue
            if max_dist is not None and down - up > max_dist:
                continue
            result.append((up, down))
            break
    return result

def add_pr(pairs, seq_len, number=1, min_len=1, max_len=10):
    """Return Pairs object
    
    Add paired region to pairs
    """
    result = copy(pairs).directed()
    for i in range(1,number+1):
        paired = dict.fromkeys(flatten(result))
        unpaired = []
        curr = []
        for idx in range(seq_len):
            if idx in paired:
                if curr:
                    unpaired.append(curr)
                    curr = []
            else:
                curr.append(idx)
        if curr:
            unpaired.append(curr)

        one = None
        two = None
        for x in range(100):
            region_len = choice(range(min_len,max_len+1))
            #print region_len
            fit_regions = [r for r in unpaired if len(r) >= region_len]
            if len(fit_regions) < 2:
                continue
            #print fit_regions
            one, two = sample(fit_regions,2)
            #print one, two
            break
        if not one:
            continue
        if one[0] < two[0]:
            up, down = one, two
        else:
            up, down = two, one
        #print 'up, down', up, down

        if len(up) == region_len:
            start_opts = [up[0]]
        else:
            start_opts = up[:len(up)-region_len+1]
        #print start_opts
        start = choice(start_opts)
        if len(down) == region_len:
            end_opts = [down[-1]]
        else:
            end_opts = down[region_len-1:]
        end = choice(end_opts)
        result.extend(PairedRegion(start, end, region_len).Pairs)
    return result

def extend_helices(pairs, seq_len, number):
    """Return Pairs object with extensions on the existing helices.
    
    pairs -- Pairs object
    seq_len -- int, length of the matching sequence, used to determine
        the maximum index of a paired position
    number -- number of base pairs to be added

    Method uses blocks of perfectly nested base pairs [(n,m),(n+1,m-1),...]
    """
    result = copy(pairs)
    num_added = 0
    while num_added < number:
        bp_blocks = PairedRegionsFromPairs(result)
        #bp_blocks = paired_blocks(result)
        paired = bp_blocks.pairedPos() #dict.fromkeys(flatten(result))
        potential = []
        for bl in bp_blocks:
            up, down = bl.Start-1,bl.End+1
            if up < down and up >= 0 and down < seq_len:
                if up not in paired and down not in paired:
                    potential.append((up,down))
            if bl.Pairs[-1] != bl.Pairs[0]:
                up, down = bl.Pairs[-1][0]+1, bl.Pairs[-1][1]-1
                if up < down and up >= 0 and down < seq_len: 
                    if up not in paired and down not in paired and up < down:
                        potential.append((up,down))
        #print potential
        if not potential:
            break
        chosen = choice(potential)
        result.append(chosen)
        num_added += 1
    result.sort()
    return result

# REMOVE STUFF

def remove_pair(pairs, number):
    """Return Pairs object containing original pairs minus 'number' of pairs.

    pairs -- Pairs object
    number -- int, number of pairs to remove randomly
    """
    result = Pairs(copy(pairs))
    to_remove = sample(pairs, number)
    for item in to_remove:
        result.remove(item)
    return result

def degrade_helices(pairs, number):
    """Return Pairs object with number of bps removed from the helix flanks

    pairs -- Pairs object
    number -- int, number of base pairs that has to be removed
    """
    result = copy(pairs)
    num_removed = 0
    while num_removed < number:
        bp_blocks = paired_blocks(result)
        potential = []
        for bl in bp_blocks:
            if len(bl) > 1:
                potential.append(bl[0])
                potential.append(bl[-1])
            else:
                potential.append(bl[0])
        if not potential:
            break
        chosen = choice(potential)
        result.remove(chosen)
        num_removed += 1
    return result

def remove_blocks(pairs, number):
    """Return Pairs object with one or more base pair blocks removed.
    
    pairs -- Pairs object
    number -- number of base pair blocks to be removed.

    A base pair block is a stretch of perfectly nested base pairs:
        [(n,m),(n+1,m-1),...]
    """
    result = copy(pairs).directed()
    result.sort()
    num_removed = 0
    while num_removed < number:
        bp_blocks = paired_blocks(result)
        
        if not bp_blocks:
            break
        chosen = choice(bp_blocks)
        for i in chosen:
            result.remove(i)
        num_removed += 1
    return result

def add_remove(pairs, seq_len, number):
    """Return Pairs object with several base pairs removed and added
    """
    new_pairs = remove_pair(pairs, randint(0,number))
    result = add_pair(pairs, seq_len, randint(0,number))
    return result

# =============================================================================
# MANIPULATE EXISTING STRUCTURES -- MULTIPLE STRUCTURES
# =============================================================================

def shorter_structures(pairs, num_structs, num_bps, whole_range=False):
    """Return list of Pairs objects that are all shorter than the original.
    
    pairs -- original Pairs object
    num_structs -- the number of structure you want back (of a particular
        length). If you always remove 2 bps, and you want 10 struct, you get
        10 structs back. If you want the whole range, you get 10 structs with
        1 bp removed, and 10 structs with 2 bps removed.
    num_bps -- (max) number of base pairs to be removed
    whole_range -- boolean, indicating whether you want a series of up to 
        num_bps removed. If False, you get back num_structs with num_bps
        removed. If True you get back num_bps * num_structs with 1,..., num_bps
        removed.
    """
    result = []
    if whole_range:
        begin = 1
    else:
        begin = num_bps
    for i in range(num_structs):
        for j in range(begin, num_bps+1):
            result.append(remove_pair(pairs, j))
    return result

def longer_structures(pairs, seq_len, num_structs, num_bps, whole_range=False):
    """Return list of Pairs objects that are all longer than the original.
    
    pairs -- original Pairs object
    seq_len -- length of the matching sequence, used to determine the max
        index of a paired position.
    num_structs -- the number of structure you want back (of a particular
        length). If you always remove 2 bps, and you want 10 struct, you get
        10 structs back. If you want the whole range, you get 10 structs with
        1 bp removed, and 10 structs with 2 bps removed.
    num_bps -- (max) number of base pairs to be removed
    whole_range -- boolean, indicating whether you want a series of up to 
        num_bps removed. If False, you get back num_structs with num_bps
        removed. If True you get back num_bps * num_structs with 1,..., num_bps
        removed.
    """
    result = []
    if whole_range:
        begin = 1
    else:
        begin = num_bps
    for i in range(num_structs):
        for j in range(begin, num_bps+1):
            result.append(add_pair(pairs, seq_len=seq_len, number=j))
    return result


def keep_percentage(pairs, fraction=0.5, regions=False, number=10):
    """
    number = number of structures (Pairs objects) to return
    """
    result = []
    if regions == False:
        all = len(pairs)
        for x in range(number):
            to_choose = int(fraction*all)
            result.append(Pairs(sample(pairs, to_choose)))
    else:
        prs = PairedRegionsFromPairs(pairs)
        all = len(prs)
        for x in range(number):
            to_choose = int(fraction*all)
            result.append(PairedRegions(sample(prs, to_choose)).toPairs())
    return result


def xxx_degrade_structure(pairs):
    """Yield multiple Pairs objects with several base pairs removed"""
    for i in range(1,11):
        for j in range(1,11):
            new_pairs = copy(pairs)
            to_remove = sample(pairs, i)
            for item in to_remove:
                new_pairs.remove(item)
            yield Pairs(new_pairs)

def xxx_remove_some(pairs, seq_len, number):
    """Return list of Pairs objects with several base pairs removed"""
    #print 'remove', number
    structures = []
    for i in range(1, number+1):
        for j in range(1, number+1):
            new_p = remove_pair(pairs, i)
            new_v = new_p.toVienna(seq_len)
            structures.append(new_v)
    return structures

def xx_add_some(pairs, seq_len, number):
    """Return list of Pairs objects with several base pairs added"""
    #print 'add', number
    structures = []
    for i in range(1,number+1):
        for j in range(1,number+1):
            new_p = add_pair(pairs, seq_len, i)
            new_v = new_p.toVienna(seq_len)
            structures.append(new_v)
    return structures

def xxx_add_remove_some(pairs, seq_len, number):
    """Return list of Pairs objects with several base pairs removed and added.
    """
    structures = []
    for i in range(1,number+1):
        for j in range(1,number+1):
            new_p = add_remove(pairs, seq_len, i)
            new_v = new_p.toVienna(seq_len)
            structures.append(new_v)
    return structures

# =============================================================================
# MANIPULATE EXISTING CLASSIFICATION -- SINGLE
# =============================================================================

def permute_shuffle(cl):
    """Return new Classification object, random shuffle of the original.
    
    cl -- Classification object or simple string
    """
    x = list(str(cl))
    shuffle(x)
    return Classification(''.join(x))

def permute_replace(cl, ss='S', num=1, repl_sym='X'):
    """Return new Classification object, some positions replaces by repl_sym.

    cl -- Classification object or simple string
    ss -- structural category to be replaces by repl_sym
    num -- number of symbols to be replaces
    repl_sym -- symbol to replace the original with, default 'X'
    """
    x = list(str(cl))
    count = 0
    while count < num:
        idx = choice(range(len(x)))
        if x[idx] != ss:
            continue
        else:
            x[idx] = repl_sym
            count += 1
    return Classification(''.join(x))

def permute_add_random(cl, ss='S', num=1):
    """Return new Classification object, num positions of category ss added.
    
    cl -- Classification object or simple string
    ss -- symbol to be added to the string, only positions that didn't have
        this symbol can be replaces by it.
    num -- number of symbols of type ss to be added
    """
    x = list(str(cl))
    count = 0
    while count < num:
        idx = choice(range(len(x)))
        if x[idx] == ss:
            continue
        else:
            x[idx] = ss
            count += 1
    return Classification(''.join(x))

# =============================================================================
# OTHER HELPER FUNCTIONS
# =============================================================================

def get_unique_vienna(vienna_structures):
    """Return list of unique ViennaStructures (without energies).

    vienna_structures -- list of ViennaStructure objects or dot-bracket
        strings.
    """
    result = []
    seen = {}
    for struct in vienna_structures:
        if isinstance(struct, ViennaStructure):
            #Remove Energy from structure, so that pickle doesn't mess it up
            struct_str = str(struct).split()[0]
        else: #assume dot-bracket string
            struct_str = struct
        #s = ViennaStructure(str(s).split()[0])
        if struct_str not in seen:
            seen[struct_str] = ViennaStructure(struct_str)
        else:
            pass
    return seen.values()

def get_unique_pairs(multiple_pairs):
    """Return list of unique Pairs objects.

    multiple_pairs -- list of Pairs objects.
    """
    seen = {}
    for pairs in multiple_pairs:
        pairs.sort()
        t = tuple(pairs)
        if t not in seen:
            seen[t] = Pairs(pairs)
        else:
            pass
    return seen.values()


def label_structures(structures):
    """Changes list of structures into dictionary with unique labels

    structures: list of ViennaStructure objects

    Label format = struct_[int], starting at 1
    """
    result = {}
    for idx, struct in enumerate(structures):
        label = 'struct_' + str(idx+1)
        result[label] = struct
    return result


if __name__ == "__main__":
    
    from sys import argv
    prefix = argv[1]
    aln = pickle.load(open(prefix+'_aln'))
    cs = pickle.load(open(prefix+'_cs'))
    labels = pickle.load(open(prefix+'_lab'))
    #structs = pickle.load(open(prefix+'_struct'))

    #seq_len = map(len,aln.values())[0]
    #for seq in aln.values():
    #    assert len(seq) == seq_len

    seq_len = len(aln.Seqs[0])
    print seq_len

    cs_pairs = cs.toPairs()
   
    structures = []
    
    structures.extend(add_some(cs_pairs, seq_len, 12))
    structures.extend(remove_some(cs_pairs, seq_len, 12))
    structures.extend(add_remove_some(cs_pairs, seq_len, 10))
    
    structures.extend(mfe_structs_seqsim_aln(aln, sample_size=50, noLP=False))
    structures.extend(mfe_structs_seqsim_aln(aln, sample_size=50,\
        degap=True, noLP=False))
    structures.extend(mfe_structs_seqsim_aln(aln, sample_size=50,\
        degap=True, noLP=True))
    
    structures.extend(subopt_structs_seqsim_aln(aln,\
        sample_size=50, energy_range=2, noLP=False, degap=False))
    structures.extend(subopt_structs_seqsim_aln(aln,\
        sample_size=50, energy_range=2, noLP=False, degap=True))
    structures.extend(subopt_structs_seqsim_aln(aln,\
        sample_size=50, energy_range=2, noLP=True, degap=True))
    
    
    for s in structures:
        assert len(s) == seq_len
    
    structures = get_unique_structures(structures)
    struct_dict = label_structures(structures)
    #print struct_dict
    print len(struct_dict)
    pickle.dump(struct_dict, open(prefix+'_struct','w'))
    
    classifications = to_classifications(struct_dict)
    #print classifications
    pickle.dump(classifications, open(prefix+'_classi','w'))
    
    
    
