#!/usr/bin/env python
# test_structgen.py

"""
Author: Sandra Smit (Sandra.Smit@colorado.edu)

Revision History:
File created on 02 May 2007.

"""
from __future__ import division
from cogent.util.unit_test import TestCase, main
from cogent.core.sequence import RnaSequence, ModelSequence
from cogent.core.moltype import RNA
from cogent.core.alignment import Alignment, DenseAlignment
from cogent.struct.rna2d import ViennaStructure, Vienna, Pairs
from structgen import mfe_structs, mfe_structs_aln, subopt_structs,\
    subopt_structs_aln, add_pair, remove_pair, get_unique_vienna,\
    get_unique_pairs, shorter_structures, longer_structures,\
    degrade_helices, extend_helices, remove_blocks, add_pr,\
    mfe_constrained, subopt_constrained

class MFETests(TestCase):

    def test_mfe_structs_rnaseq(self):
        f = mfe_structs
        seq1 = RnaSequence('AGAUCGCAUGCUAGCUAGCUAGCUA')
        seq2 = RnaSequence('UUUUCCCAUGCGGGGAAAA')
        exp1 = '.........(((((....)))))..'
        exp_energy1 = -6.90
        exp_struct1 = ViennaStructure(exp1, Energy=exp_energy1)
        exp2 = '(((((((....))))))).'
        exp_energy2 = -7.40
        exp_struct2 = ViennaStructure(exp2, Energy=exp_energy2)
        self.assertEqual(f([seq1]), [exp_struct1])
        self.assertEqual(f([seq1,seq2]), [exp_struct1,exp_struct2])

    def test_mfe_structs_modelseq(self):
        f = mfe_structs
        seq1 = ModelSequence('AGAUCGCAUGCUAGCUAGCUAGCUA', Alphabet=RNA.Alphabet)
        seq2 = ModelSequence('UUUUCCCAUGCGGGGAAAA', Alphabet=RNA.Alphabet)
        exp1 = '.........(((((....)))))..'
        exp_energy1 = -6.90
        exp_struct1 = ViennaStructure(exp1, Energy=exp_energy1)
        exp2 = '(((((((....))))))).'
        exp_energy2 = -7.40
        exp_struct2 = ViennaStructure(exp2, Energy=exp_energy2)
        self.assertEqual(f([seq1]), [exp_struct1])
        self.assertEqual(f([seq1,seq2]), [exp_struct1,exp_struct2])

    def test_mfe_structs_rnaseq_degap(self):
        f = mfe_structs
        seq1 = RnaSequence('UGC-UGCUAG--UACGUAAGG-AG--GAACUU')
        exp1 = '..............(....)............' #folded while gapped
        exp_energy1 = -0.40
        exp_struct1 = ViennaStructure(exp1, Energy=exp_energy1)
        # folded while ungapped and regapped afterwards
        exp2 = '..(.(.((..........))..))........'
        exp_energy2 = -1.60
        exp_struct2 = ViennaStructure(exp2, Energy=exp_energy2)
        self.assertEqual(f([seq1], degap=False), [exp_struct1])
        self.assertEqual(f([seq1], degap=True), [exp_struct2])

    def test_mfe_aln(self):
        f = mfe_structs_aln
        seq1 = RnaSequence('AGAUCGCAUGCUAGCUAGCUAGCUA')
        seq2 = RnaSequence('UUUUCCCAUGCGGGGAAAAAUCGUA')
        aln = Alignment([seq1,seq2], MolType=RNA)
        exp1 = '.........(((((....)))))..'
        exp_energy1 = -6.90
        exp_struct1 = ViennaStructure(exp1, Energy=exp_energy1)
        exp2 = '(((((((.....)))))))......'
        exp_energy2 = -7.80
        exp_struct2 = ViennaStructure(exp2, Energy=exp_energy2)
        #self.assertEqual(f([seq1]), [exp_struct1])
        self.assertEqual(f(aln), [exp_struct1,exp_struct2])

    def test_mfe_aln_dense(self):
        f = mfe_structs_aln
        seq1 = ModelSequence('AGAUCGCAUGCUAGCUAGCUAGCUA', Alphabet=RNA.Alphabet)
        seq2 = ModelSequence('UUUUCCCAUGCGGGGAAAAAUCGUA', Alphabet=RNA.Alphabet)
        aln = DenseAlignment([seq1,seq2], Alphabet=RNA.Alphabet, MolType=RNA)
        exp1 = '.........(((((....)))))..'
        exp_energy1 = -6.90
        exp_struct1 = ViennaStructure(exp1, Energy=exp_energy1)
        exp2 = '(((((((.....)))))))......'
        exp_energy2 = -7.80
        exp_struct2 = ViennaStructure(exp2, Energy=exp_energy2)
        self.assertEqual(f(aln), [exp_struct1,exp_struct2])

    def test_mfe_constrained(self):
        seq = 'ACGUAGCUACGUAGCUGCUAGCUAGCUAGCUAGCGCUAGCGAC'
        con = '...xxxx.......xxxx.........xxxxx...........'
        exp = '(((......)))........((((((........))))))... (-11.3)'
        obs = mfe_constrained([(seq, con)])
        self.assertEqual(len(obs), 1)
        self.assertEqual(str(obs[0]), exp)
        self.assertEqual(obs[0].Energy, -11.3)


class SuboptTests(TestCase):

    def test_subopt_structs_basic(self):
        f = subopt_structs
        seq1 = RnaSequence('AGAUCGCAUGCUAGCUAGCUAGCUA')
        seq2 = RnaSequence('UUUUCCCAUGCGGGGAAAA')
        exp1 = Vienna('.........(((((....)))))..  -6.90')
        exp2 = Vienna('(((((((.....)))))))  -7.00')
        exp3 = Vienna('(((((((....))))))).  -7.40')
        exp4 = Vienna('.((((((.....)))))).  -7.10')
        exp5 = Vienna('.((((((....))))))..  -6.70')
        self.assertEqual(f([seq1]), [exp1])
        self.assertEqual(f([seq1], degap=True), [exp1])
        self.assertEqual(f([seq1,seq2]), [exp1, exp2,exp3,exp4,exp5])
        #self.assertEqual(f([seq1,seq2]), [exp_struct1,exp_struct2])

    def test_subopt_structs_sample(self):
        f = subopt_structs
        seq1 = RnaSequence('AGAUCGCAUGCUAGCUAGCUAGCUA')
        seq2 = RnaSequence('UUUUCCCAUGCGGGGAAAA')
        res = f([seq1], sample_ensemble=5)
        self.assertEqual(len(res), 5)
        res = f([seq1,seq2], sample_ensemble=5)
        self.assertEqual(len(res), 10)
        lengths = map(len,res)
        self.assertEqual(lengths.count(25), 5)
        self.assertEqual(lengths.count(19), 5)

    def test_subopt_structs_degap(self):
        f = subopt_structs
        seq1 = RnaSequence('ACGU---GG-ACGU--C-C--CGCG')
        exp1 = Vienna('(((........)))...........  -0.40')
        exp2 = Vienna('((((......))))...........  -0.90')
        exp3 = Vienna('.(((......)))............  -0.30')
        exp4 = Vienna('.........................   0.00')
        exp_gapped1 = Vienna('.(((...((.........)..))))  -6.00')
        exp_gapped2 = Vienna('..((...((.........)..))).  -5.10')
        self.assertEqual(f([seq1]), [exp1,exp2,exp3,exp4])
        self.assertEqual(f([seq1], degap=True), [exp_gapped1,exp_gapped2])

        res = f([seq1], sample_ensemble=5)
        lengths = map(len, res)
        self.assertEqual(len(res), 5)
        self.assertEqual(lengths[0], 25)

        res = f([seq1], sample_ensemble=5, degap=True)
        lengths = map(len, res)
        self.assertEqual(len(res), 5)
        self.assertEqual(lengths[0], 25)

    def test_subopt_structs_aln(self):
        f = subopt_structs_aln
        seq1 = RnaSequence('ACGU---GG-ACGU--C-C--CGCG')
        seq2 = RnaSequence('AAGCUUUGCUAGCGCUAGCAUCGCG')
        aln = Alignment([seq1,seq2], MolType=RNA)

        exp1 = map(Vienna, ['(((........)))...........  -0.40',\
            '((((......))))...........  -0.90',\
            '.(((......)))............  -0.30',\
            '.........................   0.00'])
        exp2 = map(Vienna,['..((..(((((....)))))..)).  -8.80'])
        exp3 = map(Vienna,['.(((...((.........)..))))  -6.00',\
            '..((...((.........)..))).  -5.10'])
        self.assertEqual(f(aln, degap=False), exp1+exp2)
        self.assertEqual(f(aln, degap=True), exp3+exp2)

    def test_subopt_structs_dense_aln(self):
        f = subopt_structs_aln
        alpha = RNA.Alphabets['DegenGapped']
        seq1 = ModelSequence('ACGU---GG-ACGU--C-C--CGCG', Alphabet=alpha)
        seq2 = ModelSequence('AAGCUUUGCUAGCGCUAGCAUCGCG', Alphabet=alpha)
        aln = DenseAlignment([seq1,seq2], Alphabet=alpha, MolType=RNA)

        exp1 = map(Vienna, ['(((........)))...........  -0.40',\
            '((((......))))...........  -0.90',\
            '.(((......)))............  -0.30',\
            '.........................   0.00'])
        exp2 = map(Vienna,['..((..(((((....)))))..)).  -8.80'])
        exp3 = map(Vienna,['.(((...((.........)..))))  -6.00',\
            '..((...((.........)..))).  -5.10'])
        self.assertEqual(f(aln, degap=False), exp1+exp2)
        self.assertEqual(f(aln, degap=True), exp3+exp2)

    def test_subopt_constrained(self):
        seq = 'ACGUAGCUACGUAGCUGCUAGCUAGCUAGCUAGCGCUAGCGAC'
        con = '...xxxx.......xxxx.........xxxxx...........'
        exp1 = '(((......)))........((((((........))))))... (-11.3)'
        exp2 = '.((......)).........((((((........))))))... (-10.0)'
        exp3 = '....................((((((........))))))... (-10.0)'
        obs = subopt_constrained([(seq, con)], energy_range=2)
        self.assertEqual(len(obs), 3)
        for sol in obs:
            self.failUnless(str(sol) in [exp1, exp2, exp3])
        con = '...xxxx...|...xxxx.........xxxxx...........'
        self.assertRaises(ValueError, subopt_constrained, [(seq, con)])

class ManipulationSingleTests(TestCase):

    def test_add_pair(self):
        p = Pairs([(3,10),(4,9)])
        obs = add_pair(p, 20, 2)
        self.assertEqual(len(obs), len(p)+2)

        # will make certain number attempts and add as many and possible
        # if nothing can be added anymore you get the structure as is
        p = Pairs([(0,7),(1,6),(2,5),(3,4)])
        obs = add_pair(p,9,2)
        self.assertEqual(obs, p)

    def test_remove_pair(self):
        p = Pairs([(0,7),(1,6),(2,5),(3,4)])
        self.assertEqual(len(remove_pair(p,1)),3)
        self.assertEqual(len(remove_pair(p,2)),2)
        self.assertEqual(len(remove_pair(p,4)),0)

    def test_degrade_helices(self):
        p = Pairs([(1,10),(2,9),(3,8),(4,7)])
        exp1 = Pairs([(2,9),(3,8),(4,7)])
        exp2 = Pairs([(1,10),(2,9),(3,8)])
        obs = degrade_helices(p, 1)
        self.failUnless(obs == exp1 or obs == exp2)

    def test_extend_helices(self):
        p = Pairs([(2,9),(3,8),(12,18)])
        exp1 = Pairs([(2,9),(3,8),(4,7)])
        exp2 = Pairs([(1,10),(2,9),(3,8)])
        obs = extend_helices(p, seq_len=25, number=5)
        print obs

    def test_remove_blocks(self):
        p = Pairs([(2,9),(3,8),(12,18),(13,17),(16,14),(5,25)])
        obs = remove_blocks(p, number=1)
        print obs

    def test_add_pr(self):
        p = Pairs([(1,10),(2,9), (12,18)])
        print add_pr(p, 30)
        print add_pr(p, number=2, seq_len=30)
        print add_pr(p, number=4, seq_len=30)
        print add_pr(p, number=8, seq_len=30)



class ManipulateMultipleTests(TestCase):

    def est_shorter_structures(self):
        p1 = Pairs([(1,10),(2,9),(3,8),(6,15),(7,14),(18,25),(20,23)])
        res = shorter_structures(p1, 5, 2, whole_range=True)
        print res
        print len(res)
        print map(len,res)

    def est_longer_structures(self):
        p1 = Pairs([(1,10),(2,9),(18,25),(20,23)])
        res = longer_structures(p1, 50, 10, 2, whole_range=True)
        print res
        print len(res)
        print map(len,res)


class HelperFunctionTests(TestCase):

    def test_get_unique_vienna_str(self):
        f = get_unique_vienna
        v1 = '(((...)))'
        v2 = '((((...))))'
        v3 = '(((...)))'
        v4 = '((...)).((...))..'
        structs = [v1,v2,v3,v4]
        exp = [v1,v2,v4]
        obs = f(structs)
        self.assertEqualItems(obs, exp)
        self.failUnless(isinstance(obs[0], ViennaStructure))

    def test_get_unique_vienna_object(self):
        f = get_unique_vienna
        v1 = Vienna('(((...))) -23.5')
        v2 = Vienna('((((...)))) -8.3')
        v3 = Vienna('(((...))) -0.17') # energy ignored anyway
        v4 = Vienna('((...)).((...))..')
        structs = [v1,v2,v3,v4]
        exp = [v1,v2,v4]
        obs = f(structs)
        self.assertEqualItems(obs, exp)
        self.failUnless(isinstance(obs[0], ViennaStructure))

    def test_get_unique_pairs(self):
        f = get_unique_pairs
        p1 = Pairs([(1,10),(2,9),(3,8)])
        p2 = Pairs([(3,8),(1,10),(2,9)])
        p3 = [(10,20),(11,19),(14,16)]
        p4 = [(2,8),(6,20)]
        mp = [p1,p2,p3,p4]
        exp = [p1,p3,p4]
        obs = f(mp)
        self.assertEqualItems(obs, exp)
        for res in obs:
            self.failUnless(isinstance(res, Pairs))


if __name__ == "__main__":
    main()
