#!/usr/bin/env python
# aln_util.py

"""
Author: Sandra Smit (Sandra.Smit@colorado.edu)

Revision History:
File created on 26 Oct 2006.
10/27/06 Sandra Smit: written alignment comparison functions.
5/22/07 Sandra Smit: added take_sub_aln
"""

from __future__ import division
from random import choice
from numpy import sum
from cogent.core.alignment import Alignment, DenseAlignment
from cogent.core.sequence import RnaSequence, ModelSequence
from cogent.core.moltype import RNA
from cogent.parse.fasta import MinimalFastaParser

# ===========================================================
# ALIGNMENT COMPARISON
# ===========================================================

def fraction_same(first_set, second_set):
    """Return fraction of aligned positions in both sets"""
    both = frozenset(first_set) & frozenset(second_set)
    return 2*len(both)/(len(first_set)+len(second_set))

def bp_measure(first_set, second_set):
    """Return similarity between sets as (A and B)/(A or B)"""
    first_set = frozenset(first_set)
    second_set = frozenset(second_set)
    return len(first_set&second_set)/len(first_set|second_set)

def aligned_positions(seq1, seq2):
    """Return set of aligned positions in ungapped coordinates
    """
    assert len(seq1) == len(seq2) # because aligned
    gap1 = seq1.MolType.Gaps
    gap2 = seq2.MolType.Gaps
    g_to_u1 = seq1.gapMaps()[1]
    g_to_u2 = seq2.gapMaps()[1]
    result = []
    for idx, (s1, s2) in enumerate(zip(seq1, seq2)):
        if s1 not in gap1 and s2 not in gap2:
            result.append((g_to_u1[idx], g_to_u2[idx]))
    result_set = frozenset(result)
    assert len(result) == len(result_set) # no duplicates
    return result

def alignment_similarity(aln1, aln2, set_comp_f=fraction_same):
    """Return similarity between the alignments.

    aln1, aln2: dictionaries of {seq_id: seq} seq should be real Sequence
        object with a MolType

    This should be used for two alignments, containing identical sequences
    gapped in different ways. For example, aligning a set of sequences with
    program A versus aligning the same set with program B. How similar are
    the two alignments?
    
    Both alignments should contain same sequences (ungapped) under the
    same keys.
    Right now it doesn't work for real Alignment objects.

    For each pairwise sequence comparison, by default, the fraction
    of positions that is aligned in both alignments is calculated as
    2*(aln_pos_in_both)/(aln_pos_in_1+aln_pos_in_2)
    The measure for the whole alignment is the average over all pairwise
    comparisons.
    """
    assert len(aln1) == len(aln2) # same length
    aln1_keys = aln1.keys()
    aln2_keys = aln2.keys()
    aln1_keys.sort()
    aln2_keys.sort()
    assert aln1_keys == aln2_keys # same keys
    # same ungapped sequences in both alignments
    for k in aln1_keys:
        assert aln1[k].degap() == aln2[k].degap()

    all = []
    # calculate alignment similarity
    for i in range(len(aln1_keys)):
        for j in range(i+1, len(aln1_keys)):
            i_key, j_key = aln1_keys[i], aln1_keys[j]
            first_aln_pos = aligned_positions(aln1[i_key], aln1[j_key])
            second_aln_pos = aligned_positions(aln2[i_key], aln2[j_key])
            all.append(set_comp_f(first_aln_pos, second_aln_pos))
    avg_all = sum(all)/len(all)
    return avg_all

# ===========================================================
# END -- ALIGNMENT COMPARISON
# ===========================================================

def take_sub_aln(aln, size=100):
    """Return new RNA Alignment containing a random subset of the sequences.
    
    aln -- Alignment object, should contain RnaSequences
    size -- int, number of sequences in the new alignment
    """
    if size < 1:
        raise ValueError("Size of subalignment has to be at least one")
    if size >= len(aln.Seqs):
        return aln
    all = range(len(aln.Seqs))
    selected = []
    for i in range(size):
        x = choice(all)
        selected.append(x)
        all.remove(x)
    selected.sort()
    sel_seqs = []
    # use NamedSeqs to keep names in subalignment
    for idx, (name, seq) in enumerate(list(aln.NamedSeqs.items())):
        if idx in selected:
            if isinstance(aln, Alignment):
                sel_seqs.append(RnaSequence(str(seq), Name=name))
            if isinstance(aln, DenseAlignment):
                sel_seqs.append(ModelSequence(str(seq), Name=name,\
                Alphabet=RNA.Alphabets.DegenGapped))
    if isinstance(aln, Alignment):
        return aln.__class__(sel_seqs, MolType=aln.MolType)
    if isinstance(aln, DenseAlignment):
        return aln.__class__(sel_seqs, MolType=aln.MolType,\
            Alphabet=aln.Alphabet)


if __name__ == "__main__":
    pass
