#!/usr/bin/env python

from __future__ import division
from numpy import asarray

def unbiased_var(x, axis=None, dtype=None, out=None):
    """Return the variance of array x along the specified axis.

    x -- numpy array or something that can be made into one
    axis -- integer, axis along which the variance is computed. The default
        is to compute the variance of the flattened array
    dtype -- type, type to use in computing the variance. For arrays of integer
        type the default is float32, for arrays of float types it is the same 
        as the array type.
    out -- ndarray, alternative output array in which to place the result.
        It must have the same shape as the expected output but the type
        will be cast if necessary. 

    Variance is a measure of the spread of a distribution. The variance is
    computed for the flattened array by default, otherwise over the specified
    axis. 

    The variance is the average of the squared deviations from the mean, 
    i.e. var = mean((x - x.mean())**2). The computed variance is UNBIASED
    (in contrast to the numpy var() function), i.e. the mean is computed
    by dividing by the number of elements minus one (N-1), rather than by N.
    """
    if axis is None:
        # numeric.asarray(self.compressed()).var()
        return unbiased_var(asarray(x).flatten(), axis=0)
    if axis >= len(x.shape):
        raise ValueError("axis(=%s) out of bounds"%(axis))
    a = x.swapaxes(axis,0)
    a = a - a.mean(axis=0)
    a *= a
    # a /= (a.count(axis=0)-1)
    if len(a) == 1:
        return 0.0
    a /= (len(a)-1)
    return a.swapaxes(0,axis).sum(axis)

def unbiased_std(x, axis=None, dtype=None, out=None):
    """Return the standard deviation of array x along the specifies axis.

    x -- numpy array or something that can be made into one
    axis -- integer, axis along which the variance is computed. The default
        is to compute the variance of the flattened array
    dtype -- type, type to use in computing the variance. For arrays of integer
        type the default is float32, for arrays of float types it is the same 
        as the array type.
    out -- ndarray, alternative output array in which to place the result.
        It must have the same shape as the expected output but the type
        will be cast if necessary. 

    The standard deviation is a measure of the spread of a distribution. The
    standard deviation is computed for the flattened array by default,
    otherwise over the specified axis.

    The standard deviation is the square root of the average of the
    squared deviations from the mean, i.e. var = sqrt(mean((x - x.mean())**2)).
    The computed standard deviation is UNBIASED (in contrast to the numpy
    std() function), i.e. the mean is computed by dividing by the number
    of elements minus one (N-1), rather than by N. 
    """
    return (unbiased_var(x, axis, dtype))**0.5
