#!/usr/bin/env python
# test_aln_util.py

"""
Author: Sandra Smit (Sandra.Smit@colorado.edu)

Revision History:
File created on 27 Oct 2006.
10/27/06 Sandra Smit: written tests for alignment comparison.
5/22/07 Sandra Smit: added tests for take_sub_aln
"""
from __future__ import division
from cogent.util.unit_test import TestCase, main
from cogent.struct.rna2d import Pairs
from cogent.core.sequence import RnaSequence, ModelSequence
from cogent.core.moltype import RNA
from cogent.core.alignment import Alignment, DenseAlignment
from rnalib.util.aln_util import fraction_same, bp_measure, aligned_positions,\
    alignment_similarity, take_sub_aln

class AlnCmpTest(TestCase):
    """Tests for alignment comparison"""

    def test_measures(self):
        """fraction_same and bp_measure: should give expected answers"""
        """
        CAG----
        CAGUCAA
        aligned = (0,0),(1,1),(2,2)
        ----CAG
        CAGUCAA
        aligned = (0,4),(1,5),(2,6)
        --C-GU-
        AGCAGUU
        aligned = (0,2),(1,4),(2,5)
        --C-G-U
        AGCAGUU
        aligned = (0,2),(1,4),(2,6)
        """
        p1 = [(0,0),(1,1),(2,2)]
        p2 = [(0,4),(1,5),(2,6)]
        p3 = [(0,2),(1,4),(2,5)]
        p4 = [(0,2),(1,4),(2,6)]

        self.assertEqual(fraction_same(p1,p2), 0)
        self.assertEqual(bp_measure(p1,p2), 0)
        self.assertFloatEqual(fraction_same(p3,p4), 2/3)
        self.assertFloatEqual(bp_measure(p3,p4), .5)

    def test_aligned_positions(self):
        """aligned_positions: valid input"""
        r1 = RnaSequence('UUCAG-CAU')
        r2 = RnaSequence('U-CUGGCUU')
        exp = [(0,0),(2,1),(3,2),(4,3),(5,5),(6,6),(7,7)]
        obs = aligned_positions(r1, r2)
        self.assertEqual(obs, exp)

        r1 = RnaSequence('--GCCUACGA')
        r2 = RnaSequence('AAG--UU-G-')
        exp = [(0,2),(3,3),(4,4),(6,5)]
        obs = aligned_positions(r1, r2)
        self.assertEqual(obs, exp)

    def test_aligned_positions_empty(self):
        """aligned_positions: invalid input or empty seq"""
        r1 = RnaSequence('')
        r2 = RnaSequence('')
        exp = []
        obs = aligned_positions(r1, r2)
        self.assertEqual(obs, exp)

        r1 = RnaSequence('UCA--GG')
        r2 = RnaSequence('UCG')
        self.assertRaises(AssertionError, aligned_positions, r1, r2)

    def test_aln_similarity(self):
        """alignment_similarity: should return expected answer"""
        aln1 = {'a': RnaSequence('-UC--GA'), 'b': RnaSequence('U-CCAG-'),\
            'c': RnaSequence('AUGGUG-')}
        aln2 = {'a': RnaSequence('UC-GA--'), 'b': RnaSequence('UC-CA-G'),\
            'c': RnaSequence('AUGG-UG')}

        self.assertFloatEqual(alignment_similarity(aln1, aln2,\
            set_comp_f=fraction_same), 0.3333333)
        self.assertFloatEqual(alignment_similarity(aln1, aln2,\
            set_comp_f=bp_measure), 0.233333)

class AlnModificationTests(TestCase):

    def setUp(self):
        """setUp for AlnModificationTests"""
        self.alpha = RNA.Alphabets.DegenGapped
        self.r1 = RnaSequence('ACGUACGU', Name='r1')
        self.r2 = RnaSequence('A-UUACG-', Name='r2')
        self.r3 = RnaSequence('-CUAA-GU', Name='r3')
        self.r4 = RnaSequence('AA---CGU', Name='r4')
        self.rna_seqs = [self.r1, self.r2, self.r3, self.r4]
        self.aln = Alignment(self.rna_seqs, MolType=RNA)
        
        self.m1 = ModelSequence('ACGUACGU', Name='m1', Alphabet=self.alpha)
        self.m2 = ModelSequence('A-UUACG-', Name='m2', Alphabet=self.alpha)
        self.m3 = ModelSequence('-CUAA-GU', Name='m3', Alphabet=self.alpha)
        self.m4 = ModelSequence('AA---CGU', Name='m4', Alphabet=self.alpha)
        self.model_seqs = [self.m1, self.m2, self.m3, self.m4]
        self.dense_aln = DenseAlignment(self.model_seqs,\
            MolType=RNA, Alphabet=self.alpha)

        """
        self.x1 = ModelSequence('ACGUACGU', Alphabet=self.alpha)
        self.x2 = ModelSequence('A-UUACG-', Alphabet=self.alpha)
        self.x3 = ModelSequence('-CUAA-GU', Alphabet=self.alpha)
        self.x_seqs = {'x1':self.x1, 'x2':self.x2, 'x3':self.x3}
        self.dense_aln2 = DenseAlignment(self.x_seqs,\
            MolType=RNA, Alphabet=self.alpha)
        #print self.dense_aln2
        print self.dense_aln2.Seqs
        print self.dense_aln2.NamedSeqs
        """

    def test_take_sub_aln_normal(self):
        """take_sub_aln: should handle Alignment object"""
        # keeps names at the moment
        sub_aln = take_sub_aln(self.aln, size=2)
        self.assertEqual(len(sub_aln.Seqs), 2)
        self.failUnless(isinstance(sub_aln, Alignment))
        for l in sub_aln.NamedSeqs:
            self.failUnless(l in self.aln.NamedSeqs.keys())

        sub_aln2 = take_sub_aln(self.aln, size=100)
        self.assertEqual(sub_aln2, self.aln)
        self.assertEqual(sub_aln2.Seqs, self.aln.Seqs)
        
        self.assertRaises(ValueError, take_sub_aln, self.aln, size=0)

    def test_take_sub_aln_dense(self):
        """take_sub_aln: should handle DenseAlignment object"""
        sub_aln = take_sub_aln(self.dense_aln, size=2)
        self.assertEqual(len(sub_aln.Seqs), 2)
        self.failUnless(isinstance(sub_aln, DenseAlignment))
        for l in sub_aln.NamedSeqs:
            self.failUnless(l in self.dense_aln.NamedSeqs.keys())

        sub_aln2 = take_sub_aln(self.dense_aln, size=100)
        self.assertEqual(sub_aln2, self.dense_aln)
        self.assertEqual(sub_aln2.Seqs, self.dense_aln.Seqs)
        
        self.assertRaises(ValueError, take_sub_aln, self.dense_aln, size=0)

if __name__ == "__main__":
    main()

