#!/usr/bin/env python
"""Script to create fasta alignment from DCSE formatted file. 
Discard structure information

On the meaning of N and o:
A problem arises because the symbol "N", used by authors when publishing
or submitting sequences, can have two different meanings:

A residue could not be properly identified on a sequencing gel. This can
be due to template heterogeneity (e.g. in the case of reverse transcriptase
sequencing) or to ambiguity in the polymerase reaction. In this case the
number of unidentified nucleotides, although not their identity, is known.

The sequence was only partially determined and after alignment with a
complete sequence of a related sequence, the undetermined areas were
padded with N's. In this case both the number and the identity of the
nucleotides is unknown. Unfortunately, most authors do not explicitly
mention which case applies. We assume, somewhat arbitrarily, that case
1 (known number of nucleotides, unidentifiable on a sequencing gel)
applies if a single N, or a row of 5 N's at most, is found intercalated
between known nucleotides. Rows of more than 5 N's are treated as
unsequenced areas of unknown length in a partially sequenced RNA. Two
different symbols are used to distinguish the two cases, as follows.
N unidentified nucleotide, length of unidentified area probably known.
o unidentified nucleotide, length of unidentified area unknown.
In this case we intercalate a number of "o" symbols matching the number
nucleotides in the most closely related species.

N is left untouched by the script. 'o' is replaced by ?
Only sequences that contain symbols in the RnaAlphabet are returned

If files contain duplicate keys, this is reported at the top of the
output file. These comments have to be removed manually.

Revision history
8/16/06 Sandra Smit: written script
"""

from string import strip
from cogent.core.moltype import RNA
from Numeric import array, arange, take, put, nonzero, alltrue

def dcse_to_seq_dict(infile, split_on=' '*6):
    """Returns dict of labels and unparsed sequence strings"""
    res = {}
    for line in infile:
        if line.startswith('R'):
            continue
        if line.rstrip(): #ignore blank lines
            rdb_seq, ann = line.rstrip().rsplit(split_on)
            rdb_seq = rdb_seq.rstrip()
            ann = '_'.join(map(strip, ann.split()))
            if ann.startswith('Helix_numbering'):
                continue
            if ann in res:
                print "Duplicate key found: %s"%(ann)
                if res[ann] == rdb_seq:
                    continue
                else:
                    print "Sequences don't match"
                    continue
                    #raise KeyError, "Duplicate key found: %s"%(ann)
            res[ann] = rdb_seq
            #len(rdb_seq), ann
    return res

def check_seq_lengths(seq_dict):
    l = None
    try:
        for k, v in seq_dict.items():
            if l is None:
                l = len(v)
            else:
                assert len(v) == l
    except AssertionError:
        return False
    return True
        
def check_sequences(seq_dict):
    good_seqs = {}
    for k, v in seq_dict.items():
        seq_array = array(v)
        to_take = arange(1,len(seq_array),2)
        seq_without_struct = take(seq_array, to_take)
        put(seq_without_struct, nonzero(seq_without_struct=='o'),'?')
        #print seq_without_struct
        #print len(seq_array), len(seq_without_struct)

        if alltrue([i[0] in RNA for i in seq_without_struct]):
            good_seqs[k] = seq_without_struct.tostring().upper()
        else:
            raise ValueError("Sequence contains unknown character: %s"\
                %(seq_without_struct))
            #print seq_without_struct
            #continue
    return good_seqs


if __name__ == "__main__":
    
    from sys import argv
    seq_dict = dcse_to_seq_dict(open(argv[1]))
    #print len(seq_dict)
    if not check_seq_lengths(seq_dict):
        print "Sequence lengths don't match, check alignment"
    
    good_seqs = check_sequences(seq_dict)
    if not check_seq_lengths(good_seqs):
        print "Sequence lengths don't match, check alignment"
    #print len(good_seqs)
    for k, v in good_seqs.items():
        print '>'+str(k)
        print v
